unit SpTBXDkPanels;

{==============================================================================
Version 0.4

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License.
You may obtain a copy of the License at http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The initial developer of this code is Robert Lee.

Requirements:
  - Jordan Russell's Toolbar 2000
    http://www.jrsoftware.org
  - Alex Denisov's TBX
    http://g32.org

Wish list for TBX:
  - Fix this bug in TTBXMultiDock: when the TBXMultiDock is created the
    SplitHeight and SplitWidth properties of the docked panels are not updated.
    This causes inconsistences in runtime resizing via SplitHeight/SplitWidth.
    The difference between the Height and the SplitHeight of all the docked
    panels is allways 16, the same size as DSGN_DROPZONESIZE declared
    in TTBXMultiDock.ArrangeToolbars.
    DSGN_DROPZONESIZE is used to leave some space at the bottom of the
    MultiDock at design time to add a drop zone.
    To reproduce it:
    1) Drop a TTBXMultiDock rename it to 'M' and set the Position to dpLeft
    2) Drop 3 TBXDockablePanels rename them to 'DP1', 'DP2' and 'DP3'
    3) Resize the dockable panels at design time
    4) Run this line:
       Caption := Format('%d - %d - %d', [M.Height,
         DP1.Height + DP2.Height + DP3.Height,
         DP1.SplitHeight + DP2.SplitHeight + DP3.SplitHeight]);
    The same problem is present with SplitWidth when the panels are
    horizontally docked.
    To workaround this use the following utility function in Form.OnShow and
    TBXMultiDock.OnResize:

    type
      TTBDockAccess = class(TTBDock);

    procedure UpdateDockSplit(MD: TTBXMultiDock);
    // Fix TBX bug, when the Dock is created SplitHeight and SplitWidth are
    // not updated.
    var
      I: Integer;
      DockList: TList;
      P: TTBXDockablePanel;
    begin
      DockList := TTBDockAccess(MD).DockVisibleList;
      MD.BeginUpdate;
      try
        if MD.Position in [dpLeft, dpRight] then begin
          for I := 0 to DockList.Count - 1 do begin
            P:= DockList[I];
            P.SplitHeight := P.Height;
          end;
        end
        else begin
          for I := 0 to DockList.Count - 1 do begin
            P := DockList[I];
            P.SplitWidth := P.Width;
          end;
        end;
      finally
        MD.EndUpdate;
      end;
    end;
  - Create a TTBXCustomDockablePanel and make TTBXDockablePanel descend from it.
    This will allow to hide undesirable properties of the component.
  - Add a PaintDockPanelCaptionBackground method to the TTBXTheme class: using
    CurrentTheme.PaintDockPanelNCArea draws ALL the NC area with a fixed caption
    height(GetSystemMetrics(SM_CYSMCAPTION)), there's no way to just draw the
    caption background.

Development notes:
  - All the TBX theme changes and adjustments are marked with '[TBXTheme-Change]'.
  - There are two ways of painting the panel caption: using
    CurrentTheme.PaintDockPanelNCArea or CurrentTheme.PaintFloatingBorder.
    PaintDockPanelNCArea paints ALL the NC area!!! and is too small to
    draw, the Height depends on GetSystemMetrics(SM_CYSMCAPTION).
    PaintFloatingBorder can adjust the caption size, but some custom painting
    must be done, because the title bar is painted by the themes.

History:
21 July 2004 - version 0.4
  - New component added, TSpTBXProgressBar, a progress bar with
    TBX themes support and custom painting events.
  - Added Cancel and ModalResult properties for TSpTBXButton.
  - Added CaptionGlow and CaptionGlowColor properties to TSpTBXLabel,
    TSpTBXButton, TSpTBXCheckBox, TSpTBXRadioButton and
    TSpTBXProgressBar. Setting these properties will paint the caption
    outline with the desired color.
  - Added maximize/minimize/restore capability to TSpTBXDockablePanel.
  - Added FixedDockedSize property to TSpTBXDockablePanel, it determines
    if the DockabelPanel is resizable when docked.

12 July 2004 - version 0.3.1
  - Unchanged.

9 July 2004 - version 0.3
  - Fixed incorrect painting on TSpTBXGroupBox when the theme
    service was not available, thanks to Cyril for reporting this.
  - Fixed AVs in SpTBXTabControl when the parent was a Frame,
    thanks to Cyril for fix.
  - Fixed incorrect text wrapping on TSpTBXLabel, thanks
    to Cyril for the fix.
  - Added Underline and UnderlineColor properties to TSpTBXLabel.
  - Added LinkText, LinkTextParams, and LinkFont properties to
    TSpTBXLabel and TSpTBXButton. Setting this properties will
    allow the control to browse the specified LinkText, some of
    the values that the control can execute are:
    - mailto:name@name.com
    - http://www.borland.com
    - ftp://www.borland.com
    - news://news.jrsoftware.org
    - C:\Windows (opens a folder in Windows Explorer, use the
      LinkTextParams property to specify the parameters).
    - Notepad.exe (runs an application, use the LinkTextParams
      property to specify the parameters).
    Note: the controls can execute unicode links and params, for
    example you can open a folder named with unicode text or
    browse for unicode URI web sites, LinkText and LinkTextParams
    are WideStrings.

24 June 2004 - version 0.2
  - Fixed incorrect painting on TSpTBXPanel.
  - Fixed incorrect painting on TSpTBXGroupBox.
  - Fixed TSpTBXRadioButton bug, it was uncheckable.
  - Fixed TSpTBXRadioButton focus selection by keyboard, it now
    checks itself when it gets the focus when using arrow keys.
  - Added ThemeType property to TSpTBXCheckBox, it allows the
    CheckBox to override the active TBX theme.
  - Added ThemeType property to TSpTBXRadioButton, it allows the
    RadioButton to override the active TBX theme.

22 June 2004 - version 0.1
  - Initial release.

==============================================================================}

interface

uses
  Windows, Messages, Classes, SysUtils, Controls, Graphics, ImgList, Forms,
  StdCtrls, ExtCtrls, TB2Item, TB2Dock, TB2Toolbar, TB2Common,
  TBX, TBXDkPanels, TBXThemes, SpTBXItem, SpTBXTabs;

const
  //Buttons have 4 states (normal, hottrack, pushed, disabled)
  ConstStatesCount = 4;

type
  TSpTBXDockablePanel = class;

  TSpTBXDockPosition = (
    dpxLeft,
    dpxTop,
    dpxRight,
    dpxBottom,
    dpxClient
  );

  TSpTBXProgressCaption = (
    pctNone,
    pctDefault,
    pctPercentage,
    pctProgress
  );

  TSpTBXProgressBarChangeEvent = procedure(Sender: TObject; NewPosition: Integer) of object;

  { TSpTBXMultiDock }

  TSpTBXMultiDock = class(TTBXMultiDock)
  private
    FPosition: TSpTBXDockPosition;
    procedure SetPosition(Value: TSpTBXDockPosition);
  public
    procedure ArrangeToolbars; override;
    constructor Create(AOwner: TComponent); override;
  published
    property Position: TSpTBXDockPosition read FPosition write SetPosition default dpxTop;
  end;

  { TSpTBXDockablePanelToolbar }

  TSpTBXDockablePanelToolbar = class(TSpTBXToolbar)
  protected
    function GetRightAlignMargin: Integer; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X: Integer; Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
    function GetParentDockablePanel: TTBXDockablePanel;
  end;

  { TSpTBXDockablePanel }

  TSpTBXDockablePanel = class(TTBXDockablePanel, ITBItems)
  private
    FCaptionRotation: TDPCaptionRotation;
    FFixedDockedSize: Boolean;
    FRestoring: Boolean;
    FPanel: TPanel;
    FShowCaptionWhenDocked: Boolean;
    FShowCaption: Boolean;
    FOnDrawCaptionPanel: TSpTBXDrawEvent;
    procedure DockRequestDock(Sender: TObject; Bar: TTBCustomDockableWindow; var Accept: Boolean);
    procedure DockResize(Sender: TObject);
    function GetCaptionPanelSize: TPoint;
    function GetItems: TTBCustomItem;  // For ITBItems interface
    function GetRootItems: TTBRootItem;
    procedure SetShowCaptionWhenDocked(const Value: Boolean);
    procedure SetCaptionRotation(const Value: TDPCaptionRotation);
    procedure SetShowCaption(const Value: Boolean);
    function GetImages: TCustomImageList;
    procedure SetImages(const Value: TCustomImageList);
    function GetView: TTBToolbarView;
  protected
    FDock: TSpTBXDock;
    FToolbar: TSpTBXDockablePanelToolbar;
    FRestoreSize: TPoint;
    FMaximized: Boolean;
    FMinimized: Boolean;
    procedure ConstrainedResize(var MinWidth, MinHeight, MaxWidth, MaxHeight: Integer); override;
    procedure DockDrawBackground(Sender: TObject; ACanvas: TCanvas; ARect: TRect;
      const PaintStage: TSpTBXPaintStage; var PaintDefault: Boolean); virtual;
    procedure DrawNCArea(const DrawToDC: Boolean; const ADC: HDC; const Clip: HRGN); override;
    procedure DoDrawCaptionPanel(ACanvas: TCanvas; ARect: TRect;
      const PaintStage: TSpTBXPaintStage; var PaintDefault: Boolean); virtual;
    procedure GetChildren(Proc: TGetChildProc; Root: TComponent); override;  // For ITBItems interface
    procedure GetDockPanelInfo(out DockPanelInfo: TTBXDockPanelInfo); override;
    function InternalMaximize(Restore: Boolean): Boolean;
    procedure Resize; override;
    procedure Loaded; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure DrawCaptionPanel(ACanvas: TCanvas; X, Y: Integer);
    function Maximize: Boolean;
    function Minimize(ProportionalSizing: Boolean = False): Boolean;
    function Restore(ProportionalSizing: Boolean = False): Boolean;
    function Minimized: Boolean;
    function Maximized: Boolean;
    function SizeToggle(ToMaximize: Boolean = False; ProportionalSizing: Boolean = False): Boolean;
    property CaptionPanelSize: TPoint read GetCaptionPanelSize;
    property View: TTBToolbarView read GetView;
  published
    property MinClientHeight default 8;
    property MinClientWidth default 8;
    property Images: TCustomImageList read GetImages write SetImages;
    property Items: TTBRootItem read GetRootItems;
    property FixedDockedSize: Boolean read FFixedDockedSize write FFixedDockedSize default False;
    // Override ShowCaptionWhenDocked, the inherited property should allways be False
    property ShowCaption: Boolean read FShowCaption write SetShowCaption default True;
    // Override ShowCaptionWhenDocked, the inherited property should allways be False
    property ShowCaptionWhenDocked: Boolean read FShowCaptionWhenDocked write SetShowCaptionWhenDocked default True;
    // Override CaptionRotation, the inherited property should allways be dpcrAlwaysHorz
    property CaptionRotation: TDPCaptionRotation read FCaptionRotation write SetCaptionRotation default dpcrAlwaysHorz;
    property SupportedDocks default [dkStandardDock, dkMultiDock];
    property OnDrawCaptionPanel: TSpTBXDrawEvent read FOnDrawCaptionPanel write FOnDrawCaptionPanel;
  end;

  { TSpTBXPanel }

  TSpTBXPanel = class(TCustomControl)
  private
    FCaption: WideString;
    procedure ReadCaptionW(Reader: TReader);
    procedure WriteCaptionW(Writer: TWriter);
    procedure SetCaption(const Value: WideString);
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure WMWindowPosChanged(var Message: TWMWindowPosChanged); message WM_WINDOWPOSCHANGED;
  protected
    procedure AdjustClientRect(var Rect: TRect); override;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure DefineProperties(Filer: TFiler); override;
    procedure Paint; override;
    procedure SetName(const Value: TComponentName); override; // Initial value for FCaption
  public
    constructor Create(AOwner: TComponent); override;
    procedure InvalidateBackground(InvalidateChildren: Boolean = True);
  published
    // Don't let the VCL store the WideString, it doesn't write it correctly.
    // Use DefineProperties instead
    property Caption: WideString read FCaption write SetCaption stored False; // Hides the inherited Caption
    property Align;
    property Anchors;
    property BiDiMode;
    property Color;
    property Constraints;
    property UseDockManager;
    property DockSite;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property ParentBiDiMode;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnCanResize;
    property OnClick;
    property OnConstrainedResize;
    property OnContextPopup;
    property OnDockDrop;
    property OnDockOver;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnGetSiteInfo;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDock;
    property OnStartDrag;
    property OnUnDock;
  end;

  { TSpTBXGroupBox }

  TSpTBXGroupBox = class(TSpTBXPanel)
  private
    procedure CMDialogChar(var Message: TCMDialogChar); message CM_DIALOGCHAR;
    procedure CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;
  protected
    procedure AdjustClientRect(var Rect: TRect); override;
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
  end;

  { TSpTBXTextObject }

  TSpTBXTextObject = class(TTBXTextObject)
  private
    FAlignment: TAlignment;
    FCaption: WideString;
    FHint: WideString;
    FCaptionChanged: Boolean;
    FCaptionGlow: Boolean;
    FCaptionGlowColor: TColor;
    FChecked: Boolean;
    FDrawPushedCaption: Boolean;
    FImageChangeLink: TChangeLink;
    FImageIndex: TImageIndex;
    FImages: TCustomImageList;
    FThemeType: TSpTBXThemeType;
    FLinkText: WideString;
    FLinkTextParams: WideString;
    FLinkFont: TFont;
    FOnDraw: TSpTBXDrawEvent;
    FOnDrawCaption: TSpTBXDrawTextEvent;
    FOnDrawHint: TSpTBXDrawHintEvent;
    FOnGetImageIndex: TSpTBXGetImageIndexEvent;
    procedure ImageListChange(Sender: TObject);
    procedure SetCaption(const Value: WideString);
    procedure ReadCaptionW(Reader: TReader);
    procedure WriteCaptionW(Writer: TWriter);
    procedure SetHint(const Value: WideString);
    procedure ReadHintW(Reader: TReader);
    procedure WriteHintW(Writer: TWriter);
    procedure ReadLinkTextW(Reader: TReader);
    procedure WriteLinkTextW(Writer: TWriter);
    procedure ReadLinkTextParamsW(Reader: TReader);
    procedure WriteLinkTextParamsW(Writer: TWriter);
    function GetPushed: Boolean;
    procedure SetAlignment(const Value: TAlignment);
    procedure SetCaptionGlow(const Value: Boolean);
    procedure SetCaptionGlowColor(const Value: TColor);
    procedure SetImageIndex(const Value: TImageIndex);
    procedure SetImages(const Value: TCustomImageList);
    procedure SetLinkFont(const Value: TFont);
    procedure SetThemeType(const Value: TSpTBXThemeType);
    procedure CMHintShow(var Message: TCMHintShow); message CM_HINTSHOW;
    procedure WMSetCursor(var Message: TWMSetCursor); message WM_SETCURSOR;
  protected
    // Painting
    procedure AdjustFont(AFont: TFont); override;
    procedure DoDrawHint(AHintBitmap: TBitmap; var AHint: Widestring; var PaintDefault: Boolean); virtual;
    function DoDrawItem(ACanvas: TCanvas; ARect: TRect; const PaintStage: TSpTBXPaintStage): Boolean; virtual;
    function DoDrawText(ACanvas: TCanvas; var Rect: TRect; Flags: Longint): Integer; override;
    procedure DoGetImageIndex(var AImageList: TCustomImageList; var AImageIndex: integer); virtual;
    function GetFocusRect(R, TextR, GlyphR: TRect): TRect; reintroduce; virtual;
    function GetTextAlignment: TAlignment; override;
    function GetTextMargins: TRect; override;
    procedure Paint; override;

    // Sizing
    function  CanAutoSize(var NewWidth, NewHeight: Integer): Boolean; override;
    procedure DoAdjustHeight(ACanvas: TCanvas; var NewHeight: Integer); override;

    // Mouse
    procedure DoMouseEnter; override;
    procedure DoMouseLeave; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override; // Fix a TBX bug, TTBXTextObject can't recieve focus if it's inside a Panel

    // Component
    procedure DefineProperties(Filer: TFiler); override;
    procedure ExecuteLink; virtual;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetChecked(Value: Boolean); virtual;
    procedure SetName(const Value: TComponentName); override; // Initial value for FCaption

    property Alignment: TAlignment read FAlignment write SetAlignment default taLeftJustify; // Hides the inherited Alignment
    property Autosize default True;
    property Checked: Boolean read FChecked write SetChecked default False;
    property CaptionGlow: Boolean read FCaptionGlow write SetCaptionGlow default False;
    property CaptionGlowColor: TColor read FCaptionGlowColor write SetCaptionGlowColor default clYellow;
    property DrawPushedCaption: Boolean read FDrawPushedCaption write FDrawPushedCaption default False;
    property Images: TCustomImageList read FImages write SetImages;
    property ImageIndex: TImageIndex read FImageIndex write SetImageIndex default -1;
    property LinkText: WideString read FLinkText write FLinkText;
    property LinkTextParams: WideString read FLinkTextParams write FLinkTextParams;
    property LinkFont: TFont read FLinkFont write SetLinkFont;
    property ThemeType: TSpTBXThemeType read FThemeType write SetThemeType default thtTBX;
    property OnDraw: TSpTBXDrawEvent read FOnDraw write FOnDraw;
    property OnDrawCaption: TSpTBXDrawTextEvent read FOnDrawCaption write FOnDrawCaption;
    property OnDrawHint: TSpTBXDrawHintEvent read FOnDrawHint write FOnDrawHint;
    property OnGetImageIndex: TSpTBXGetImageIndexEvent read FOnGetImageIndex write FOnGetImageIndex;
    property SmartFocus default True;
    property TabStop default True;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function CanFocus: Boolean; override;
    procedure Click; override;
    function GetControlsAlignment: TAlignment; override;
    procedure GetSize(out TotalR, TextR, GlyphR: TRect); virtual;
    function GetTextFlags: Cardinal;
    property MouseInControl;
    property Pushed: Boolean read GetPushed;
  published
    // Don't let the VCL store the WideString, it doesn't write it correctly.
    // Use DefineProperties instead
    property Caption: WideString read FCaption write SetCaption stored False; // Hides the inherited Caption
    property Hint: WideString read FHint write SetHint stored False; // Hides the inherited Hint
  end;

  { TSpTBXLabel }

  TSpTBXCustomLabel = class(TSpTBXTextObject)
  private
    FFocusControl: TWinControl;
    FUnderline: Boolean;
    FUnderlineColor: TColor;
    procedure SetFocusControl(const Value: TWinControl);
    procedure SetUnderline(const Value: Boolean);
    procedure SetUnderlineColor(const Value: TColor);
    procedure CMDialogChar(var Message: TCMDialogChar); message CM_DIALOGCHAR;
  protected
    function DoDrawItem(ACanvas: TCanvas; ARect: TRect; const PaintStage: TSpTBXPaintStage): Boolean; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    property FocusControl: TWinControl read FFocusControl write SetFocusControl;
    property Underline: Boolean read FUnderline write SetUnderline default False;
    property UnderlineColor: TColor read FUnderlineColor write SetUnderlineColor default clBtnShadow;
  public
    constructor Create(AOwner: TComponent); override;
    procedure GetSize(out TotalR: TRect; out TextR: TRect; out GlyphR: TRect); override;
  end;

  TSpTBXLabel = class(TSpTBXCustomLabel)
  published
    property Align;
    property Anchors;
    property Autosize;
    property BiDiMode;
    property Color;
    property Constraints;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property Margins;
    property ParentBiDiMode;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowAccelChar;
    property ShowHint;
    property Visible;
    property Wrapping;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
    // TSpTBXCustomLabel properties
    property Alignment;
    property CaptionGlow;
    property CaptionGlowColor;
    property FocusControl;
    property Images;
    property ImageIndex;
    property LinkText;
    property LinkTextParams;
    property LinkFont;
    property Underline;
    property UnderlineColor;
    property OnDraw;
    property OnDrawCaption;
    property OnDrawHint;
    property OnGetImageIndex;
  end;

  { TSpTBXButtonControl }

  TSpTBXButtonControl = class(TSpTBXTextObject)
  private
    procedure CMDialogChar(var Message: TCMDialogChar); message CM_DIALOGCHAR;
    procedure CMDialogKey(var Message: TCMDialogKey); message CM_DIALOGKEY;
  public
    function CanFocus: Boolean; override;
  end;

  { TSpTBXCheckBox }

  TSpTBXCustomCheckBox = class(TSpTBXButtonControl)
  private
    FAllowGrayed: Boolean;
    FState: TCheckBoxState;
    FStateChanged: Boolean;
    procedure SetState(const Value: TCheckBoxState);
  protected
    procedure DoGetImageIndex(var AImageList: TCustomImageList; var AImageIndex: integer); override;
    procedure SetChecked(Value: Boolean); override;
    procedure Toggle; virtual;
    property AllowGrayed: Boolean read FAllowGrayed write FAllowGrayed default False;
    property State: TCheckBoxState read FState write SetState default cbUnchecked;
  public
    procedure Click; override;
    constructor Create(AOwner: TComponent); override;
    procedure GetSize(out TotalR, TextR, GlyphR: TRect); override;
  end;

  TSpTBXCheckBox = class(TSpTBXCustomCheckBox)
  published
    property Align;
    property Anchors;
    property BiDiMode;
    property Color;
    property Constraints;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property Margins;
    property ParentBiDiMode;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowAccelChar;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property Wrapping;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
    // TSpTBXCustomCheckBox properties
    property Alignment;
    property AllowGrayed;
    property CaptionGlow;
    property CaptionGlowColor;
    property Checked;
    property State;
    property ThemeType;
    property OnDraw;
    property OnDrawCaption;
    property OnDrawHint;
    property OnGetImageIndex;
  end;

  { TSpTBXRadioButton }

  TSpTBXCustomRadioButton = class(TSpTBXButtonControl)
  private
    FStateChanged: Boolean;
    procedure CMFocusChanged(var Message: TCMFocusChanged); message CM_FOCUSCHANGED;
  protected
    procedure DoGetImageIndex(var AImageList: TCustomImageList; var AImageIndex: integer); override;
    procedure SetChecked(Value: Boolean); override;
    procedure Toggle; virtual;
    procedure TurnSiblingsOff;
  public
    procedure Click; override;
    procedure GetSize(out TotalR, TextR, GlyphR: TRect); override;
  end;

  TSpTBXRadioButton = class(TSpTBXCustomRadioButton)
  published
    property Align;
    property Anchors;
    property BiDiMode;
    property Color;
    property Constraints;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property Margins;
    property ParentBiDiMode;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowAccelChar;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property Wrapping;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
    // TSpTBXCustomRadioButton properties
    property Alignment;
    property CaptionGlow;
    property CaptionGlowColor;
    property Checked;
    property ThemeType;
    property OnDraw;
    property OnDrawCaption;
    property OnDrawHint;
    property OnGetImageIndex;
  end;

  { TSpTBXButton }

  TSpTBXCustomButton = class(TSpTBXButtonControl)
  private
    FBitmap: TBitmap;
    FBitmapTransparent: Boolean;
    FActive: Boolean;
    FCancel: Boolean;
    FDefault: Boolean;
    FModalResult: TModalResult;
    procedure BitmapChanged(Sender: TObject);
    procedure SetBitmap(const Value: TBitmap);
    procedure SetDefault(const Value: Boolean);
    procedure CMDialogKey(var Message: TCMDialogKey); message CM_DIALOGKEY;
    procedure CMFocusChanged(var Message: TCMFocusChanged); message CM_FOCUSCHANGED;
  protected
    procedure AdjustFont(AFont: TFont); override;
    function BitmapValid: boolean;
    function DoDrawItem(ACanvas: TCanvas; ARect: TRect; const PaintStage: TSpTBXPaintStage): Boolean; override;
    function GetFocusRect(R, TextR, GlyphR: TRect): TRect; override;
    function GetTextMargins: TRect; override;
    property Alignment default taCenter;
    property Bitmap: TBitmap read FBitmap write SetBitmap;
    property BitmapTransparent: Boolean read FBitmapTransparent write FBitmapTransparent default True;
    property Cancel: Boolean read FCancel write FCancel default False;
    property Default: Boolean read FDefault write SetDefault default False;
    property ModalResult: TModalResult read FModalResult write FModalResult default 0;
    property ThemeType default thtWindows;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Click; override;
    function GetSkinStateRect: TRect;
  end;

  TSpTBXButton = class(TSpTBXCustomButton)
  published
    property Align;
    property Anchors;
    property BiDiMode;
    property Color;
    property Constraints;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property Margins;
    property ParentBiDiMode;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowAccelChar;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property Wrapping;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
    // TSpTBXCustomButton properties
    property Alignment;
    property Bitmap;
    property BitmapTransparent;
    property Cancel;
    property CaptionGlow;
    property CaptionGlowColor;
    property Checked;
    property Default;
    property Images;
    property ImageIndex;
    property LinkText;
    property LinkTextParams;
    property LinkFont;
    property ModalResult;
    property ThemeType;
    property OnDraw;
    property OnDrawCaption;
    property OnDrawHint;
    property OnGetImageIndex;
  end;

  TSpTBXCustomProgressBar = class(TSpTBXTextObject)
  private
    FMin: Integer;
    FMax: Integer;
    FPosition: Integer;
    FSmooth: Boolean;
    FVertical: Boolean;
    FCaptionType: TSpTBXProgressCaption;
    FOnProgressChange: TSpTBXProgressBarChangeEvent;
    procedure SetMax(const Value: integer);
    procedure SetMin(const Value: integer);
    procedure SetPosition(Value: integer);
    procedure SetSmooth(const Value: Boolean);
    procedure SetVertical(const Value: Boolean);
    procedure SetCaptionType(const Value: TSpTBXProgressCaption);
  protected
    function DoDrawItem(ACanvas: TCanvas; ARect: TRect; const PaintStage: TSpTBXPaintStage): Boolean; override;
    procedure DoProgressChange; virtual;
    function GetTextMargins: TRect; override;
    property Alignment default taCenter;
    property CaptionGlow default True;
    property CaptionType: TSpTBXProgressCaption read FCaptionType write SetCaptionType default pctPercentage;
    property Color default clBtnFace;
    property Max: Integer read FMax write SetMax default 100;
    property Min: Integer read FMin write SetMin default 0;
    property Position: Integer read FPosition write SetPosition default 0;
    property Smooth: Boolean read FSmooth write SetSmooth default False;
    property Vertical: Boolean read FVertical write SetVertical default False;
    property ThemeType default thtWindows;
    property OnProgressChange: TSpTBXProgressBarChangeEvent read FOnProgressChange write FOnProgressChange;
  public
    constructor Create(AOwner: TComponent); override;
    procedure StepIt(Delta: Integer = 1);
  end;

  TSpTBXProgressBar = class(TSpTBXCustomProgressBar)
  published
    property Align;
    property Anchors;
    property Color;
    property Constraints;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Visible;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
    // TSpTBXCustomProgressBar properties
    property Alignment;
    property CaptionGlow;
    property CaptionGlowColor;
    property CaptionType;
    property Max;
    property Min;
    property Position;
    property Smooth;
    property Vertical;
    property ThemeType;
    property OnDraw;
    property OnDrawCaption;
    property OnDrawHint;
    property OnProgressChange;    
  end;

{ Dockable Panel helpers }
function SpDkPanelInmediateResizableSibling(DP: TTBXDockablePanel): TTBXDockablePanel;
function SpDkPanelResize(DP: TTBXDockablePanel; NewSize: Integer; ProportionalSizing: Boolean = False): Boolean;
function SpDkPanelValidate(Sender: TObject; out DP: TSpTBXDockablePanel): Boolean;
function SpDkPanelSizeToggle(Sender: TObject; ToMaximize: Boolean = False; ProportionalSizing: Boolean = False): Boolean;

{ Painting helpers }
procedure SpDrawXPCaptionPanel(WindowHandle: THandle; ACanvas: TCanvas; AWidth, AHeight: Integer; IsActive, IsFloating: Boolean);
procedure SpDrawXPCheckBoxGlyph(ACanvas: TCanvas; ARect: TRect; Enabled: Boolean; State: TCheckBoxState; MouseInControl, Pushed, Focused: Boolean; ThemeType: TSpTBXThemeType);
procedure SpDrawXPRadioButtonGlyph(ACanvas: TCanvas; ARect: TRect; Enabled: Boolean; Checked, MouseInControl, Pushed, Focused: Boolean; ThemeType: TSpTBXThemeType);
procedure SpDrawXPGroupBox(ACanvas: TCanvas; ARect: TRect; ACaption: WideString; TextFlags: Cardinal; Enabled: Boolean);
procedure SpDrawXPProgressBar(ACanvas: TCanvas; ARect: TRect; Min, Max, Position: Integer; Back, Fore: TBitmap); overload;
function SpDrawXPProgressBar(ACanvas: TCanvas; ARect: TRect; Vertical, Smooth: Boolean; Min, Max, Position: Integer; ThemeType: TSpTBXThemeType): Integer; overload;

implementation

uses
  TB2Consts, TBXUtils, TBXUxThemes, ShellAPI;

const
  HT_TB2k_Border = 2000;

type
  TTBRootItemAccess = class(TTBRootItem);
  TTBCustomItemAccess = class(TTBCustomItem);
  TTBXDockablePanelAccess = class(TTBXDockablePanel);
  TTBDockAccess = class(TTBDock);

var
  SpStockImageList: TImageList;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ Helpers }

function SpDkPanelInmediateResizableSibling(DP: TTBXDockablePanel): TTBXDockablePanel;
var
  I, J: Integer;
  R: TRect;
  Dock: TTBDock;
  DockList: TList;
  DkPanel: TTBXDockablePanel;
begin
  Result := nil;
  DkPanel := nil;
  Dock := DP.CurrentDock;
  if Assigned(Dock) then DockList := TTBDockAccess(Dock).DockVisibleList
  else DockList := nil;
  if not Assigned(DockList) or (DockList.Count <= 1) or (DP.Floating) or
    (Dock.Position in [dpTop, dpBottom]) then Exit;

  // Resize only the inmediate dockable panel sibling
  // Find DP on the DockList
  for I := 0 to DockList.Count - 1 do
    if DockList[I] = DP then begin
      DkPanel := DockList[I];
      Break;
    end;

  // Find the inmediate resizable sibling
  if Assigned(DkPanel) then begin
    // Find the next sibling
    if I + 1 < DockList.Count then
      for J := I + 1 to DockList.Count - 1 do begin
        DkPanel := DockList[J];
        if DkPanel.SplitHeight > DP.SplitHeight then begin
          R := Rect(1, 1, 0, 0);
          TTBXDockablePanelAccess(DkPanel).ConstrainedResize(R.Left, R.Top, R.Right, R.Bottom);
          if R.Top <> R.Bottom then begin
            Result := DkPanel;
            Break;
          end;
        end;
      end;
    // If not found find the prev sibling
    if not Assigned(Result) then
      for J := I - 1 downto 0 do begin
        DkPanel := DockList[J];
        if DkPanel.SplitHeight > DP.SplitHeight then begin
          R := Rect(1, 1, 0, 0);
          TTBXDockablePanelAccess(DkPanel).ConstrainedResize(R.Left, R.Top, R.Right, R.Bottom);
          if R.Top <> R.Bottom then begin
            Result := DkPanel;
            Break;
          end;
        end;
      end;
  end;
end;

function SpDkPanelResize(DP: TTBXDockablePanel; NewSize: Integer;
  ProportionalSizing: Boolean = False): Boolean;
// Resize the dockable panel to NewSize, if ProportionalSizing is true the
// remaining pixels are distributed proportionally, if it's false only the
// inmediate sibling is resized.
// Horizontal resizing is not supported.
var
  I, H, PrevSize: Integer;
  Dock: TTBDock;
  DockList: TList;
  DkPanel: TTBXDockablePanel;
begin
  Result := False;
  Dock := DP.CurrentDock;
  if Assigned(Dock) then DockList := TTBDockAccess(Dock).DockVisibleList
  else DockList := nil;
  if not Assigned(DockList) or (DockList.Count <= 1) or (DP.Floating) or
    (Dock.Position in [dpTop, dpBottom]) then Exit;

  Dock.BeginUpdate;
  try
    // Fix TBX bug, when the Dock is resized SplitHeight is not updated.
    for I := 0 to DockList.Count - 1 do begin
      DkPanel := DockList[I];
      DkPanel.SplitHeight := DkPanel.Height;
    end;

    PrevSize := DP.Height;
    DP.SplitHeight := NewSize;
    // Redistribute the Dock.Height through all the docked panels
    if ProportionalSizing then begin
      // Proportionally distribute the remaining pixels
      H := (Dock.Height - DP.Height) div (DockList.Count - 1);
      for I := 0 to DockList.Count - 1 do begin
        DkPanel := DockList[I];
        if DkPanel <> DP then DkPanel.SplitHeight := H;
      end;
    end
    else begin
      // Resize only the inmediate dockable panel sibling
      // Find DP on the DockList
      DkPanel := SpDkPanelInmediateResizableSibling(DP);
      if Assigned(DkPanel) then
        DkPanel.SplitHeight := DkPanel.Height + (PrevSize - NewSize);
    end;
  finally
    Dock.EndUpdate;
  end;

  Result := DP.SplitHeight = DP.Height;
end;

function SpDkPanelValidate(Sender: TObject; out DP: TSpTBXDockablePanel): Boolean;
var
  I: TTBCustomItem;
  C: TComponent;
  T: TTBCustomToolbar;
begin
  Result := False;
  DP := nil;
  if Assigned(Sender) and (Sender is TTBCustomItem) then begin
    I := Sender as TTBCustomItem;
    C := I.GetParentComponent;
    if Assigned(C) and (C is TTBCustomToolbar) then begin
      T := C as TTBCustomToolbar;
      if Assigned(T.Owner) and (T.Owner is TSpTBXDockablePanel) then begin
        DP := T.Owner as TSpTBXDockablePanel;
        Result := True;
      end;
    end;
  end;
end;

function SpDkPanelSizeToggle(Sender: TObject; ToMaximize: Boolean = False;
  ProportionalSizing: Boolean = False): Boolean;
var
  DP: TSpTBXDockablePanel;
begin
  Result := False;
  if SpDkPanelValidate(Sender, DP) then
    Result := DP.SizeToggle(ToMaximize, ProportionalSizing);
end;

procedure SpDrawXPCaptionPanel(WindowHandle: THandle; ACanvas: TCanvas; AWidth, AHeight: Integer;
  IsActive, IsFloating: Boolean);
var
  R: TRect;
  Theme: string;
  DockPanelInfo: TTBXDockPanelInfo;
  B: TBitmap;
begin
  // There are two ways of painting the panel caption: using
  // CurrentTheme.PaintDockPanelNCArea or CurrentTheme.PaintFloatingBorder.
  // But both paint ALL the NC area!!! and is too small to
  // draw, the Height depends on GetSystemMetrics(SM_CYSMCAPTION).
  // We have to get the title bar and paint it in the canvas.

  Theme := TBXCurrentTheme;

  B := TBitmap.Create;
  try
    B.Width := AWidth;
    B.Height := AHeight + 300;
    FillChar(DockPanelInfo, SizeOf(DockPanelInfo), 0);
    DockPanelInfo.WindowHandle := WindowHandle;
    DockPanelInfo.ViewType := DPVT_NORMAL or DPVT_RESIZABLE;
    DockPanelInfo.IsVertical := True;
    DockPanelInfo.BorderStyle := bsSingle;
    CurrentTheme.GetViewBorder(DockPanelInfo.ViewType, DockPanelInfo.BorderSize);
    DockPanelInfo.ClientWidth := B.Width;
    DockPanelInfo.ClientHeight := B.Height;
    DockPanelInfo.EffectiveColor := ACanvas.Brush.Color;
    DockPanelInfo.ShowCaption := True;

    R := Rect(0, 0, B.Width, B.Height);
    CurrentTheme.PaintDockPanelNCArea(B.Canvas, R, DockPanelInfo);

    StretchBlt(ACanvas.Handle, 0, 0, AWidth, AHeight,
      B.Canvas.Handle, 0, 0, B.Width, GetSystemMetrics(SM_CYSMCAPTION) + 1, SRCCOPY);
  finally
    B.Free;
  end;
end;

procedure SpDrawXPCheckBoxGlyph(ACanvas: TCanvas; ARect: TRect; Enabled: Boolean;
  State: TCheckBoxState; MouseInControl, Pushed, Focused: Boolean;
  ThemeType: TSpTBXThemeType);
const
  EnabledState: array [Boolean] of Integer = (PFS_DISABLED, 0);
  StateFlags: array [TCheckBoxState] of Integer = (0, PFS_CHECKED, PFS_MIXED);
  HotState: array [Boolean] of Integer = (0, PFS_HOT);
  PushedState: array [Boolean] of Integer = (0, PFS_PUSHED);
  FocusedState: array [Boolean] of Integer = (0, PFS_FOCUSED);
var
  Flags: Integer;
begin
  if (ThemeType = thtWindows) and not SpXPThemesAvailable then
    ThemeType := thtNone;
  Flags := 0;

  Case ThemeType of
    thtNone:
      begin
        Case State of
          cbChecked: Flags := DFCS_BUTTONCHECK or DFCS_CHECKED;
          cbGrayed: Flags := DFCS_BUTTON3STATE or DFCS_CHECKED;
          cbUnChecked: Flags := DFCS_BUTTONCHECK;
        end;
        if not Enabled then
          Flags := Flags or DFCS_INACTIVE;
        if Pushed then
          Flags := Flags or DFCS_PUSHED;
        DrawFrameControl(ACanvas.Handle, ARect, DFC_BUTTON, Flags);
      end;
    thtWindows:
      begin
        Case State of
          cbChecked: Flags := CBS_CHECKEDNORMAL;
          cbGrayed: Flags := CBS_MIXEDNORMAL;
          cbUnChecked: Flags := CBS_UNCHECKEDNORMAL;
        end;
        
        if not Enabled then Inc(Flags, 3)
        else
          if Pushed then Inc(Flags, 2)
          else if MouseInControl then Inc(Flags);
        DrawThemeBackground(BUTTON_THEME, ACanvas.Handle, BP_CHECKBOX, Flags, ARect, nil);
      end;
    thtTBX:
      begin
        Flags := EnabledState[Enabled];
        if Enabled then Flags := Flags or StateFlags[State] or
          HotState[MouseInControl] or PushedState[Pushed] or FocusedState[Focused];
        CurrentTheme.PaintFrameControl(ACanvas, ARect, PFC_CHECKBOX, Flags, nil);
      end;
  end;
end;

procedure SpDrawXPRadioButtonGlyph(ACanvas: TCanvas; ARect: TRect; Enabled: Boolean;
  Checked, MouseInControl, Pushed, Focused: Boolean; ThemeType: TSpTBXThemeType);
const
  EnabledState: array [Boolean] of Integer = (PFS_DISABLED, 0);
  StateFlags: array [Boolean] of Integer = (0, PFS_CHECKED);
  HotState: array [Boolean] of Integer = (0, PFS_HOT);
  PushedState: array [Boolean] of Integer = (0, PFS_PUSHED);
  FocusedState: array [Boolean] of Integer = (0, PFS_FOCUSED);
var
  Flags: Integer;
begin
  if (ThemeType = thtWindows) and not SpXPThemesAvailable then
    ThemeType := thtNone;

  Case ThemeType of
    thtNone:
      begin
        Flags := DFCS_BUTTONRADIO;
        if Checked then
          Flags := Flags or DFCS_CHECKED;
        if not Enabled then
          Flags := Flags or DFCS_INACTIVE;
        if Pushed then
          Flags := Flags or DFCS_PUSHED;
        DrawFrameControl(ACanvas.Handle, ARect, DFC_BUTTON, Flags);
      end;
    thtWindows:
      begin
        if Checked then Flags := RBS_CHECKEDNORMAL
        else Flags := RBS_UNCHECKEDNORMAL;

        if not Enabled then Inc(Flags, 3)
        else
          if Pushed then Inc(Flags, 2)
          else if MouseInControl then Inc(Flags);
        DrawThemeBackground(BUTTON_THEME, ACanvas.Handle, BP_RADIOBUTTON, Flags, ARect, nil);
      end;
    thtTBX:
      begin
        Flags := EnabledState[Enabled];
        if Enabled then Flags := Flags or StateFlags[Checked] or
          HotState[MouseInControl] or PushedState[Pushed] or FocusedState[Focused];
        CurrentTheme.PaintFrameControl(ACanvas, ARect, PFC_RADIOBUTTON, Flags, nil);
      end;
  end;
end;

procedure SpDrawXPGroupBox(ACanvas: TCanvas; ARect: TRect; ACaption: WideString;
  TextFlags: Cardinal; Enabled: Boolean);
var
  H, Width, Height: Integer;
  R: TRect;
  ThemeFlags: Longint;
  CaptionRect,
  OuterRect: TRect;
begin
  Width := ARect.Right - ARect.Left;
  Height := ARect.Bottom - ARect.Top;
  if ACaption <> '' then begin
    CaptionRect := Rect(0, 0, 1, 1);
    SpDrawXPText(ACanvas, ACaption, CaptionRect, TextFlags or DT_CALCRECT);
    if (TextFlags and DT_RTLREADING) = 0 then
      OffsetRect(CaptionRect, 8, 0)
    else
      OffsetRect(CaptionRect, Width - 8 - CaptionRect.Right, 0);
  end
  else
    CaptionRect := Rect(0, 0, 0, 0);

  if SpXPThemesAvailable then begin
    OuterRect := ARect;
    OuterRect.Top := (CaptionRect.Bottom - CaptionRect.Top) div 2;
    with CaptionRect do
      ExcludeClipRect(ACanvas.Handle, Left, Top, Right, Bottom);
    if Enabled then
      ThemeFlags := GBS_NORMAL
    else
      ThemeFlags := GBS_DISABLED;

    DrawThemeBackground(BUTTON_THEME, ACanvas.Handle, BP_GROUPBOX, ThemeFlags, OuterRect, nil);

    SelectClipRgn(ACanvas.Handle, 0);
    if ACaption <> '' then
      DrawThemeText(BUTTON_THEME, ACanvas.Handle, BP_GROUPBOX, ThemeFlags, PWideChar(ACaption), -1, TextFlags, 0, CaptionRect);
  end
  else begin
    H := SpDrawXPText(ACanvas, '0', R, TextFlags or DT_CALCRECT);
    R := Rect(0, H div 2 - 1, Width, Height);
    Inc(R.Left);
    Inc(R.Top);

    if ACaption <> '' then begin
      ExcludeClipRect(ACanvas.Handle, CaptionRect.Left, CaptionRect.Top,
        CaptionRect.Right, CaptionRect.Bottom);
      try
        DrawEdge(ACanvas.Handle, R, EDGE_ETCHED, BF_RECT);
      finally
        SelectClipRgn(ACanvas.Handle, 0);
      end;
      SpDrawXPText(ACanvas, ACaption, CaptionRect, TextFlags);
    end
    else
      DrawEdge(ACanvas.Handle, R, EDGE_ETCHED, BF_RECT);
  end;
end;

{
procedure SpDrawXPProgressBar(ACanvas: TCanvas; ARect: TRect; Enabled, Vertical, Smooth: Boolean;
  Min, Max, Position: Integer; ThemeType: TSpTBXThemeType);
var
  ItemInfo: TTBXItemInfo;
  Percent: Integer;
  DeltaR: TRect;
begin
  if Position < Min then Position := Min
  else if Position > Max then Position := Max;
  if (ThemeType = thtWindows) and not SpThemesAvailable then
    ThemeType := thtNone;

  // Get the delta
  if (Max > Min) and (Position > Min) then begin
    DeltaR := ARect;
    Case ThemeType of
      thtNone: InflateRect(DeltaR, -2, -2);
    end;
    Percent := (Position * 100) div (Max - Min);
    if Vertical then
      DeltaR.Bottom := DeltaR.Top + (((DeltaR.Bottom - DeltaR.Top) * Percent) div 100)
    else
      DeltaR.Right := DeltaR.Left + (((DeltaR.Right - DeltaR.Left) * Percent) div 100);
  end
  else
    DeltaR := Rect(0, 0, 0, 0);


  Case ThemeType of
    thtNone:
      begin
        DrawEdge(ACanvas.Handle, ARect, BDR_SUNKENOUTER, BF_RECT);
        if not IsRectEmpty(DeltaR) then begin
          ACanvas.Brush.Color := clHighlight;
          ACanvas.FillRect(DeltaR);
        end;
      end;
    thtTBX:
      begin
        SpFillItemInfo(Enabled, True, True, False, ItemInfo);
        CurrentTheme.PaintButton(ACanvas, ARect, ItemInfo);
        if not IsRectEmpty(DeltaR) then begin
          if Vertical then
            ExcludeClipRect(ACanvas.Handle, DeltaR.Left, DeltaR.Bottom, ARect.Right, ARect.Bottom)
          else
            ExcludeClipRect(ACanvas.Handle, DeltaR.Right, DeltaR.Top, ARect.Right, ARect.Bottom);
          ItemInfo.Pushed := False;
          ItemInfo.HoverKind := hkMouseHover;
          CurrentTheme.PaintButton(ACanvas, DeltaR, ItemInfo);
          SelectClipRgn(ACanvas.Handle, 0);
        end;
      end;
  end;
end;
}

procedure SpDrawXPProgressBar(ACanvas: TCanvas; ARect: TRect;
  Min, Max, Position: Integer; Back, Fore: TBitmap);
var
  Percent, Delta: Integer;
  DeltaR, R: TRect;
begin
  if Position < Min then Position := Min
  else if Position > Max then Position := Max;

  // Get the delta
  if (Max > Min) and (Position > Min) then begin
    Percent := (Position * 100) div (Max - Min);
    DeltaR := ARect;
    R := Rect(0, 0, Back.Width, Back.Height);
    if Back.Height > Back.Width then begin
      Delta := (Back.Height * Percent) div 100;
      DeltaR.Bottom := DeltaR.Top + Delta;
      R.Bottom := R.Top + Delta;
    end
    else begin
      Delta := (Back.Width * Percent) div 100;
      DeltaR.Right := DeltaR.Left + Delta;
      R.Right := R.Left + Delta;
    end;
  end
  else
    Delta := 0;

  ACanvas.Draw(ARect.Left, ARect.Top, Back);
  if Delta > 0 then
    ACanvas.CopyRect(DeltaR, Fore.Canvas, R);
end;

function SpDrawXPProgressBar(ACanvas: TCanvas; ARect: TRect; Vertical, Smooth: Boolean;
  Min, Max, Position: Integer; ThemeType: TSpTBXThemeType): Integer;
const
  PartID: array [Boolean] of Integer = (PP_BAR, PP_BARVERT);
  ChunkID: array [Boolean] of Integer = (PP_CHUNK, PP_CHUNKVERT);
var
  ItemInfo: TTBXItemInfo;
  ChunkPaint: Boolean;
  I: Integer;
  DeltaR, R: TRect;
  B: TBitmap;
begin
  Result := 0;
  ChunkPaint := False;
  if Position < Min then Position := Min
  else if Position > Max then Position := Max;
  if (ThemeType = thtWindows) and not SpXPThemesAvailable then
    ThemeType := thtNone;

  // Get the delta
  if (Max > Min) and (Position > Min) then begin
    DeltaR := ARect;
    Case ThemeType of
      thtWindows:
        if Vertical then InflateRect(DeltaR, -3, -4)
        else InflateRect(DeltaR, -4, -3);
      thtNone: InflateRect(DeltaR, -2, -2);
    end;
    Result := (Position * 100) div (Max - Min);
    if Vertical then
      DeltaR.Top := DeltaR.Bottom - (((DeltaR.Bottom - DeltaR.Top) * Result) div 100)
    else
      DeltaR.Right := DeltaR.Left + (((DeltaR.Right - DeltaR.Left) * Result) div 100);
  end
  else
    DeltaR := Rect(0, 0, 0, 0);

  B := TBitmap.Create;
  try
    Case ThemeType of
      thtWindows:
        begin
          DrawThemeBackground(SP_PROGRESS_THEME, ACanvas.Handle, PartID[Vertical], 0, ARect, nil);
          if not IsRectEmpty(DeltaR) then begin
            // [TBXTheme-Change]: Windows XP progress bar chunks are 8 x 11,
            // but DrawThemeBackground draws 10 x 11 chunks.
            // We must draw the chunks manually.
            if Vertical then begin
              B.Width := DeltaR.Right - DeltaR.Left;
              B.Height := 8;
              R := Rect(0, 2, B.Width, B.Height);
            end
            else begin
              B.Width := 8;
              B.Height := DeltaR.Bottom - DeltaR.Top;
              R := Rect(0, 0, B.Width - 2, B.Height);
            end;
            DrawThemeBackground(SP_PROGRESS_THEME, B.Canvas.Handle, ChunkID[Vertical], 0, R, nil);
            ChunkPaint := True;
          end;
        end;
      thtNone:
        begin
          DrawEdge(ACanvas.Handle, ARect, BDR_SUNKENOUTER, BF_RECT);
          if not IsRectEmpty(DeltaR) then
            if Smooth then begin
              ACanvas.Brush.Color := clHighlight;
              ACanvas.FillRect(DeltaR);
            end
            else begin
              // Chunks are 10 x 13
              if Vertical then begin
                B.Width := DeltaR.Right - DeltaR.Left;
                B.Height := 10;
                R := Rect(0, 2, B.Width, B.Height);
              end
              else begin
                B.Width := 10;
                B.Height := DeltaR.Bottom - DeltaR.Top;
                R := Rect(0, 0, B.Width - 2, B.Height);
              end;
              B.Canvas.Brush.Color := clBtnFace;
              B.Canvas.FillRect(Rect(0, 0, B.Width, B.Height));
              B.Canvas.Brush.Color := clHighlight;
              B.Canvas.FillRect(R);
              ChunkPaint := True;
            end;
        end;
      thtTBX:
        begin
          B.Width := ARect.Right - ARect.Left;
          B.Height := ARect.Bottom - ARect.Top;
          R := Rect(0, 0, B.Width, B.Height);

          SpFillItemInfo(True, True, True, False, ItemInfo);
          CurrentTheme.PaintButton(ACanvas, ARect, ItemInfo);
          if not IsRectEmpty(DeltaR) then begin
            ItemInfo.Pushed := False;
            ItemInfo.HoverKind := hkMouseHover;
            B.Canvas.CopyRect(R, ACanvas, ARect); // B is transparent
            CurrentTheme.PaintButton(B.Canvas, R, ItemInfo);
            if Vertical then
              R.Top := R.Bottom - (DeltaR.Bottom - DeltaR.Top)
            else
              R.Right := DeltaR.Right - DeltaR.Left;
            ACanvas.CopyRect(DeltaR, B.Canvas, R);
          end;
        end;
    end;

    if ChunkPaint then begin
      if Vertical then begin
        ExcludeClipRect(ACanvas.Handle, ARect.Left, ARect.Top, ARect.Right, ARect.Top + 2);
        I := DeltaR.Bottom - B.Height;
        while I > DeltaR.Top - B.Height do begin
          ACanvas.Draw(DeltaR.Left, I, B);
          Dec(I, B.Height);
        end;
      end
      else begin
        ExcludeClipRect(ACanvas.Handle, ARect.Right - 2, ARect.Top, ARect.Right, ARect.Bottom);
        I := DeltaR.Left;
        while I < DeltaR.Right do begin
          ACanvas.Draw(I, DeltaR.Top, B);
          Inc(I, B.Width);
        end;
      end;
      SelectClipRgn(ACanvas.Handle, 0);
    end;
  finally
    B.Free;
  end;
end;

procedure ApplyMargins(var R: TRect; const Margins: TTBXControlMargins); overload;
begin
  with Margins do
  begin
    Inc(R.Left, Left); Inc(R.Top, Top);
    Dec(R.Right, Right); Dec(R.Bottom, Bottom);
  end;
end;

procedure ApplyMargins(var R: TRect; const Margins: TRect); overload;
begin
  with Margins do
  begin
    Inc(R.Left, Left); Inc(R.Top, Top);
    Dec(R.Right, Right); Dec(R.Bottom, Bottom);
  end;
end;

function GetRealAlignment(Control: TControl): TAlignment;
const
  ReverseAlignment: array [TAlignment] of TAlignment = (taRightJustify, taLeftJustify, taCenter);
begin
  Result := Control.GetControlsAlignment;
  if Control.UseRightToLeftAlignment then Result := ReverseAlignment[Result];
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXMultiDock }

procedure TSpTBXMultiDock.ArrangeToolbars;
var
  I: integer;
  P: TTBXDockablePanel;
begin
  if FPosition = dpxClient then begin
    for I := 0 to DockList.Count - 1 do begin
      P := DockList[I];
      P.Width := Width;
    end;
  end;

  inherited;
end;

constructor TSpTBXMultiDock.Create(AOwner: TComponent);
begin
  inherited;
  FPosition := dpxTop;
end;

procedure TSpTBXMultiDock.SetPosition(Value: TSpTBXDockPosition);
begin
  if FPosition <> Value then begin
    if (ControlCount <> 0) then
      raise EInvalidOperation.Create(STBDockCannotChangePosition);
    FPosition := Value;
    Case Value of
      dpxLeft:   inherited Position := dpLeft;
      dpxTop:    inherited Position := dpTop;
      dpxRight:  inherited Position := dpRight;
      dpxBottom: inherited Position := dpBottom;
      dpxClient:
        begin
          inherited Position := dpLeft;
          Align := alClient;
        end;
    end;
  end;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXDockablePanelToolbar }

constructor TSpTBXDockablePanelToolbar.Create(AOwner: TComponent);
begin
  inherited;
  CompoundToolbar := True;
end;

function TSpTBXDockablePanelToolbar.GetParentDockablePanel: TTBXDockablePanel;
var
  P: TWinControl;
begin
  Result := nil;
  P := Parent;
  while Assigned(P) do
    if P is TTBXDockablePanel then begin
      Result := P as TTBXDockablePanel;
      Break;
    end
    else
      P := P.Parent;
end;

function TSpTBXDockablePanelToolbar.GetRightAlignMargin: Integer;
begin
  Result := 3;
end;

procedure TSpTBXDockablePanelToolbar.MouseDown(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  TransparentClick: Boolean;
  DP: TTBXDockablePanel;
begin
  // Move the DockablePanel if the toolbar client area or an item with
  // tbisClicksTransparent itemstyle is clicked (like a TBXLabelItem)
  if not (csDesigning in ComponentState) and (Button = mbLeft) then begin
    //if not(ssDouble in Shift)
    DP := GetParentDockablePanel;
    if Assigned(DP) and DP.IsMovable then begin
      if Assigned(View.Selected) then
        TransparentClick := tbisClicksTransparent in TTBCustomItemAccess(View.Selected.Item).ItemStyle
      else
        TransparentClick := True;
      if TransparentClick then
        if ssDouble in Shift then
          TTBXDockablePanelAccess(DP).DoubleClick
        else begin
          SendMessage(DP.Handle, WM_NCLBUTTONDOWN, HT_TB2k_Border, 0);
          Exit;
        end;
    end;
  end;

  inherited;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXDockablePanel }

constructor TSpTBXDockablePanel.Create(AOwner: TComponent);
begin
  inherited;
  FPanel := TPanel.Create(Self);
  FPanel.Parent := Self;
  FPanel.Align := alTop;
  FPanel.BevelOuter := bvNone;

  FDock := TSpTBXDock.Create(Self);
  FDock.Parent := FPanel;
  FDock.OnRequestDock := DockRequestDock;
  FDock.OnDrawBackground := DockDrawBackground;
  FDock.OnResize := DockResize;

  FToolbar := TSpTBXDockablePanelToolbar.Create(Self);
  FToolbar.Parent := FDock;
  FToolbar.CurrentDock := FDock;
  FToolbar.BorderStyle := bsNone;
  FToolbar.DockMode := dmCannotFloatOrChangeDocks;
  FToolbar.DragHandleStyle := dhNone;
  FToolbar.ParentShowHint := False;
  FToolbar.Stretch := True;
  FToolbar.ShrinkMode := tbsmNone;
  FToolbar.ShowCaption := False;

  FRestoreSize := Point(Height, Width);

  MinClientHeight := 8;
  MinClientWidth := 8;
  // Hide the NC area of the FloatingWindow
  inherited ShowCaption := False;
  FShowCaption := True;
  // Hide the NC area of the DockablePanel
  inherited ShowCaptionWhenDocked := False;
  FShowCaptionWhenDocked := True;
  // CaptionRotation is not supported
  inherited CaptionRotation := dpcrAlwaysHorz;
  FCaptionRotation := dpcrAlwaysHorz;
end;

destructor TSpTBXDockablePanel.Destroy;
begin
  FToolbar.Free;
  FDock.Free;
  FPanel.Free;

  inherited;
end;

procedure TSpTBXDockablePanel.Loaded;
var
  I: Integer;
  C: TControl;
begin
  inherited;

  // The parent of TTBControlItem.Control should be the toolbar, not Self
  // (as setted in GetChildren for dfm streaming).
  for I := 0 to Items.Count - 1 do
    if Items[I] is TTBControlItem then begin
      C := TTBControlItem(Items[I]).Control;
      if Assigned(C) and (C.Parent <> FToolbar) then
        C.Parent := FToolbar;
    end;
end;

procedure TSpTBXDockablePanel.ConstrainedResize(var MinWidth, MinHeight,
  MaxWidth, MaxHeight: Integer);
begin
  inherited ConstrainedResize(MinWidth, MinHeight, MaxWidth, MaxHeight);
  if not FRestoring and not Floating then
    if FFixedDockedSize then begin
      MinHeight := Height;
      MaxHeight := Height;
    end
    else
      if Minimized then begin
        MinHeight := CaptionPanelSize.Y;
        MaxHeight := CaptionPanelSize.Y;
      end
      else
        MinHeight := CaptionPanelSize.Y;
end;

procedure TSpTBXDockablePanel.DockDrawBackground(Sender: TObject;
  ACanvas: TCanvas; ARect: TRect; const PaintStage: TSpTBXPaintStage;
  var PaintDefault: Boolean);
begin
  if PaintStage = pstPrePaint then begin
    PaintDefault := False;

    // OnDrawBackground is triggered by the Dock and by the docked Toolbar.
    // The Toolbar triggers it only if Dock.ThemedBackground is true, which depends
    // on CurrentTheme.PaintDockBackground, this is done in
    // TTBXToolbar.WMEraseBkgnd.

    if Floating then begin
      if ARect.Top = 0 then
        DrawCaptionPanel(ACanvas, -2, -2)  // Dock
      else
        DrawCaptionPanel(ACanvas, -2, -4); // Toolbar
    end
    else
      DrawCaptionPanel(ACanvas, ARect.Left - 2, ARect.Top - 2);
  end;
end;

procedure TSpTBXDockablePanel.DockRequestDock(Sender: TObject;
  Bar: TTBCustomDockableWindow; var Accept: Boolean);
begin
  if Assigned(FToolbar) then
    Accept := Bar = FToolbar;
end;

procedure TSpTBXDockablePanel.DoDrawCaptionPanel(ACanvas: TCanvas;
  ARect: TRect; const PaintStage: TSpTBXPaintStage;
  var PaintDefault: Boolean);
begin
  if Assigned(FOnDrawCaptionPanel) then
    FOnDrawCaptionPanel(Self, ACanvas, ARect, PaintStage, PaintDefault);
end;

procedure TSpTBXDockablePanel.DrawCaptionPanel(ACanvas: TCanvas; X, Y: Integer);
var
  R: TRect;
  PaintDefault: Boolean;
  B: TBitmap;
begin
  if not Assigned(Parent) then Exit;

  B := TBitmap.Create;
  try
    B.Width := CaptionPanelSize.X;
    B.Height := CaptionPanelSize.Y;

    R := Rect(0, 0, B.Width, B.Height);

    PaintDefault := True;
    DoDrawCaptionPanel(B.Canvas, R, pstPrePaint, PaintDefault);
    if PaintDefault then
      if Floating then
        SpDrawXPCaptionPanel(Parent.Handle, B.Canvas, B.Width, B.Height, True, True)
      else
        SpDrawXPCaptionPanel(WindowHandle, B.Canvas, B.Width, B.Height, True, False);

    PaintDefault := True;
    DoDrawCaptionPanel(B.Canvas, R, pstPostPaint, PaintDefault);
    ACanvas.Draw(X, Y, B);
  finally
    B.Free;
  end;
end;

procedure TSpTBXDockablePanel.DrawNCArea(const DrawToDC: Boolean;
  const ADC: HDC; const Clip: HRGN);
var
  DC: HDC;
  R, CR: TRect;
  ACanvas: TCanvas;
  Sz: Integer;
  DockPanelInfo: TTBXDockPanelInfo;
begin
  if not Docked or not HandleAllocated then Exit;

  if not DrawToDC then DC := GetWindowDC(Handle)
  else DC := ADC;

  Assert(DC <> 0, 'TTBXToolWindow.DrawNCArea Error');
  try
    GetDockPanelInfo(DockPanelInfo);
    GetWindowRect(Handle, R);
    OffsetRect(R, -R.Left, -R.Top);
    if not DrawToDC then
    begin
      SelectNCUpdateRgn(Handle, DC, Clip);
      CR := R;
      with DockPanelInfo.BorderSize, CR do
      begin
        InflateRect(CR, -X, -Y);
        if DockPanelInfo.ShowCaption then
        begin
          Sz := CaptionPanelSize.Y;
          if DockPanelInfo.IsVertical then Inc(Top, Sz)
          else Inc(Left, Sz);
        end;
        ExcludeClipRect(DC, Left, Top, Right, Bottom);
      end;
    end;
    ACanvas := TCanvas.Create;
    try
      ACanvas.Handle := DC;
      ACanvas.Brush.Color := EffectiveColor;
      CurrentTheme.PaintDockPanelNCArea(ACanvas, R, DockPanelInfo);

      ExcludeClipRect(ACanvas.Handle, 2, 2, CaptionPanelSize.X - 2, CaptionPanelSize.Y - 2);
      DrawCaptionPanel(ACanvas, 0, 0);
    finally
      ACanvas.Handle := 0;
      ACanvas.Free;
    end;
  finally
    if not DrawToDC then ReleaseDC(Handle, DC);
  end;
end;

function TSpTBXDockablePanel.GetCaptionPanelSize: TPoint;
var
  P: TPoint;
begin
  if Floating then begin
    P := GetFloatingBorderSize;
    Result := Point(Width + 4, FToolbar.Height + 4);
  end
  else
    Result := Point(Width, FToolbar.Height + 2);
end;

procedure TSpTBXDockablePanel.GetChildren(Proc: TGetChildProc;
  Root: TComponent);
var
  I: Integer;
  C: TControl;
begin
  // Needed to fake the DFM streaming system because the owner of the items
  // is the Form and not the Toolbar nor Self.
  // But the parent must be the Toolbar.
  // GetChildren is used to pass the children components of Self to the DFM
  // streaming system.
  // We also need to do the same with the controls of TTBControlItems.
  // More info on the Delphi help or Classes.TWriter.WriteData

  TTBRootItemAccess(Items).GetChildren(Proc, Root);
  for I := 0 to Items.Count - 1 do
    if (Items[I] is TTBControlItem) then
      if Assigned(TTBControlItem(Items[I]).Control) then begin
        C := TTBControlItem(Items[I]).Control;
        if SpFindControl(Self, C) = -1 then Proc(C);
      end;
  inherited;
end;

procedure TSpTBXDockablePanel.GetDockPanelInfo(out DockPanelInfo: TTBXDockPanelInfo);
begin
  FillChar(DockPanelInfo, SizeOf(DockPanelInfo), 0);
  DockPanelInfo.WindowHandle := WindowHandle;
  DockPanelInfo.ViewType := GetViewType;
  DockPanelInfo.IsVertical := True;
  if Assigned(CurrentDock) then begin
    DockPanelInfo.IsVertical := not IsVertCaption;
    DockPanelInfo.AllowDrag := CurrentDock.AllowDrag;
  end;
  DockPanelInfo.BorderStyle := BorderStyle;
  CurrentTheme.GetViewBorder(DockPanelInfo.ViewType, DockPanelInfo.BorderSize);
  DockPanelInfo.ClientWidth := ClientWidth;
  DockPanelInfo.ClientHeight := ClientHeight;
  DockPanelInfo.EffectiveColor := EffectiveColor;
  // Don't show caption or close button
  DockPanelInfo.ShowCaption := False;
end;

function TSpTBXDockablePanel.GetImages: TCustomImageList;
begin
  if Assigned(FToolbar) then
    Result := FToolbar.Images
  else
    Result := nil;
end;

function TSpTBXDockablePanel.GetItems: TTBCustomItem;
begin
  Result := FToolbar.Items;
end;

function TSpTBXDockablePanel.GetRootItems: TTBRootItem;
begin
  Result := FToolbar.Items;
end;

function TSpTBXDockablePanel.GetView: TTBToolbarView;
begin
  Result := FToolbar.View;
end;

procedure TSpTBXDockablePanel.SetCaptionRotation(const Value: TDPCaptionRotation);
begin
  // Do nothing
end;

procedure TSpTBXDockablePanel.SetImages(const Value: TCustomImageList);
begin
  if Assigned(FToolbar) then
    FToolbar.Images := Value;
end;

procedure TSpTBXDockablePanel.SetShowCaption(const Value: Boolean);
begin
  if FShowCaption <> Value then begin
    FShowCaption := Value;
    if Floating then
      FPanel.Visible := Value;
  end;
end;

procedure TSpTBXDockablePanel.SetShowCaptionWhenDocked(const Value: Boolean);
begin
  if FShowCaptionWhenDocked <> Value then begin
    FShowCaptionWhenDocked := Value;
    if not Floating then
      FPanel.Visible := Value;
    FPanel.Visible := Value;
  end;
end;

procedure TSpTBXDockablePanel.Resize;
begin
  inherited;
  // Invalidate NC area
  if HandleAllocated then
    SetWindowPos(Handle, 0, 0, 0, 0, 0, SWP_NOZORDER or SWP_NOSIZE or
      SWP_NOMOVE or SWP_DRAWFRAME or SWP_SHOWWINDOW);
end;

procedure TSpTBXDockablePanel.DockResize(Sender: TObject);
begin
  if FPanel.Height <> FDock.Height then
    FPanel.Height := FDock.Height;
end;

function TSpTBXDockablePanel.InternalMaximize(Restore: Boolean): Boolean;
// Resize the dockable panel to the maximum size, and minimize the rest of
// the panels
// Horizontal resizing is not supported.
var
  I, NewSize: Integer;
  DockList, ResizedList: TList;
  DkPanel: TSpTBXDockablePanel;
begin
  Result := False;
  if Assigned(CurrentDock) then DockList := TTBDockAccess(CurrentDock).DockVisibleList
  else DockList := nil;
  if not Assigned(DockList) or (DockList.Count <= 1) or (Floating) or
    (CurrentDock.Position in [dpTop, dpBottom]) then Exit;

  ResizedList := TList.Create;
  try
    CurrentDock.BeginUpdate;
    try
      // Fix TBX bug, when the Dock is resized SplitHeight is not updated.
      for I := 0 to DockList.Count - 1 do begin
        DkPanel := DockList[I];
        DkPanel.SplitHeight := DkPanel.Height;
      end;

      FRestoreSize.Y := Height;
      NewSize := Height;

      for I := 0 to DockList.Count - 1 do begin
        if (DockList[I] <> Self) and (TControl(DockList[I]) is TSpTBXDockablePanel) then begin
          DkPanel := TControl(DockList[I]) as TSpTBXDockablePanel;
          if Restore then begin
            if DkPanel.Minimized then begin
              dec(NewSize, DkPanel.FRestoreSize.Y - DkPanel.SplitHeight);
              DkPanel.FRestoring := True;
              DkPanel.FMinimized := False;
              DkPanel.SplitHeight := DkPanel.FRestoreSize.Y;
              ResizedList.Add(DkPanel);
            end;
          end
          else
            if not DkPanel.Minimized then begin
              if DkPanel.Maximized then DkPanel.FMaximized := False
              else DkPanel.FRestoreSize.Y := DkPanel.Height;
              inc(NewSize, DkPanel.SplitHeight - DkPanel.CaptionPanelSize.Y);
              DkPanel.FRestoring := True;
              DkPanel.FMinimized := True;
              DkPanel.SplitHeight := DkPanel.CaptionPanelSize.Y;
              ResizedList.Add(DkPanel);
            end;
        end;
      end;

      SplitHeight := NewSize;
      Result := True;
    finally
      CurrentDock.EndUpdate;
    end;

    // Clear the flag
    for I := 0 to ResizedList.Count - 1 do begin
      DkPanel := ResizedList[I];
      DkPanel.FRestoring := False;
    end;
  finally
    ResizedList.Free;
  end;
end;

function TSpTBXDockablePanel.Maximize: Boolean;
begin
  Result := False;
  if not FMaximized then begin
    FRestoring := True;
    FMaximized := True; // Maximized is needed by SpDkPanelMaximize
    try
      FRestoreSize.Y := Height;
      Result := InternalMaximize(False);
      FMaximized := Result;
      if Result then FMinimized := False;
    finally
      FRestoring := False;
    end;
  end;
end;

function TSpTBXDockablePanel.Minimize(ProportionalSizing: Boolean = False): Boolean;
begin
  Result := False;
  if not FMinimized then begin
    FRestoring := True;
    FMinimized := True;
    try
      FRestoreSize.Y := Height;
      Result := SpDkPanelResize(Self, CaptionPanelSize.Y, ProportionalSizing);
      FMinimized := Result;
      if Result then FMaximized := False;
    finally
      FRestoring := False;
    end;
  end;
end;

function TSpTBXDockablePanel.Restore(ProportionalSizing: Boolean = False): Boolean;
begin
  Result := False;
  FRestoring := True;
  try
    if FMaximized then begin
      Result := InternalMaximize(True);
      if Result then FMaximized := False;
    end
    else
      if FMinimized then begin
        Result := SpDkPanelResize(Self, FRestoreSize.Y, ProportionalSizing);
        if Result then FMinimized := False;
      end;
  finally
    FRestoring := False;
  end;
end;

function TSpTBXDockablePanel.SizeToggle(ToMaximize: Boolean = False;
  ProportionalSizing: Boolean = False): Boolean;
begin
  if Minimized or Maximized then
    Result := Restore(ProportionalSizing)
  else
    if ToMaximize then Result := Maximize
    else Result := Minimize(ProportionalSizing);
end;

function TSpTBXDockablePanel.Minimized: Boolean;
begin
  Result := FMinimized;
end;

function TSpTBXDockablePanel.Maximized: Boolean;
begin
  Result := FMaximized;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXPanel }

constructor TSpTBXPanel.Create(AOwner: TComponent);
begin
  inherited;
  ControlStyle := ControlStyle + [csAcceptsControls, csOpaque];
end;

procedure TSpTBXPanel.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  if not (csDesigning in ComponentState) then
    with Params.WindowClass do
      Style := Style and not (CS_HREDRAW or CS_VREDRAW);
end;

procedure TSpTBXPanel.AdjustClientRect(var Rect: TRect);
begin
  inherited AdjustClientRect(Rect);
  InflateRect(Rect, -1, -1);
end;

procedure TSpTBXPanel.DefineProperties(Filer: TFiler);
begin
  inherited;
  // Don't let the VCL store the WideString, it doesn't write it correctly.
  // Use DefineProperties instead, with a new name for the property
  Filer.DefineProperty('CaptionW', ReadCaptionW, WriteCaptionW, FCaption <> '');
end;

procedure TSpTBXPanel.InvalidateBackground(InvalidateChildren: Boolean);
begin
  // Invalidate will not fire WM_ERASEBKGND, because csOpaque is setted
  if HandleAllocated then
    if InvalidateChildren then
       RedrawWindow(Handle, nil, 0, RDW_ERASE or RDW_INVALIDATE or RDW_ALLCHILDREN)
    else
      RedrawWindow(Handle, nil, 0, RDW_ERASE or RDW_INVALIDATE);
end;

procedure TSpTBXPanel.ReadCaptionW(Reader: TReader);
begin
  case Reader.NextValue of
    vaLString, vaString:
      SetCaption(Reader.ReadString);
  else
    SetCaption(Reader.ReadWideString);
  end;
end;

procedure TSpTBXPanel.WriteCaptionW(Writer: TWriter);
begin
  Writer.WriteWideString(FCaption);
end;

procedure TSpTBXPanel.SetCaption(const Value: WideString);
begin
  if FCaption <> Value then begin
    FCaption := Value;
    inherited Caption := Value;
    InvalidateBackground;
  end;
end;

procedure TSpTBXPanel.Paint;
var
  R: TRect;
begin
  R := ClientRect;
  InflateRect(R, -1, -1);
  with R do ExcludeClipRect(Canvas.Handle, Left, Top, Right, Bottom);
  InflateRect(R, 1, 1);
  DitherRect(Canvas.Handle, R, clBtnFace, clBtnShadow);
end;

procedure TSpTBXPanel.SetName(const Value: TComponentName);
var
  ChangeText: Boolean;
begin
  ChangeText := (csSetCaption in ControlStyle) and
    not (csLoading in ComponentState) and (Name = FCaption) and
    ((Owner = nil) or not (Owner is TControl) or
    not (csLoading in TControl(Owner).ComponentState));
  if ChangeText then FCaption := Value;
  inherited;
end;

procedure TSpTBXPanel.WMEraseBkgnd(var Message: TWMEraseBkgnd);
var
  DC: HDC;
  P: TPoint;
  M: TMessage;
begin
  Message.Result := 1;
  DC := Message.DC;
  P := Point(0, 0);
  GetWindowOrgEx(DC, P);
  if (P.X <> 0) and (P.Y <> 0) then begin
    // If it was called by DrawParentBackground by a child control then dispatch
    // the message to the parent, don't call inherited because it will draw
    // to (0, 0)
    SaveDC(DC);
    try
      P := Parent.ScreenToClient(ClientToScreen(P));
      SetWindowOrgEx(DC, P.X, P.Y, nil);
      M.Msg := WM_ERASEBKGND;
      M.WParam := DC;
      M.LParam := DC;
      M.Result := 0;
      Parent.Dispatch(M);
    finally
      RestoreDC(DC, -1);
    end;
  end
  else
    TBX.DrawParentBackground(Self, Message.DC, ClientRect);
  Paint;
end;

procedure TSpTBXPanel.WMWindowPosChanged(var Message: TWMWindowPosChanged);
begin
  InvalidateBackground;
  inherited;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXGroupBox }

constructor TSpTBXGroupBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csSetCaption];
  Width := 185;
  Height := 105;
end;

procedure TSpTBXGroupBox.AdjustClientRect(var Rect: TRect);
var
  R: TRect;
  H: Integer;
begin
  inherited AdjustClientRect(Rect);
  Canvas.Font := Font;
  R := Rect;
  H := SpDrawXPText(Canvas, '0', R, DT_SINGLELINE or DT_CALCRECT);
  Inc(Rect.Top, H);
  InflateRect(Rect, -1, -1);
  if Ctl3d then InflateRect(Rect, -1, -1);
end;

procedure TSpTBXGroupBox.CMDialogChar(var Message: TCMDialogChar);
begin
  with Message do
    if IsAccel(CharCode, Caption) and CanFocus then
    begin
      SelectFirst;
      Result := 1;
    end else
      inherited;
end;

procedure TSpTBXGroupBox.CMTextChanged(var Message: TMessage);
begin
  inherited;
  InvalidateBackground;
  Realign;
end;

procedure TSpTBXGroupBox.Paint;
var
  Flags: Cardinal;
begin
  Canvas.Font := Self.Font;
  Canvas.Brush.Color := Color;

  Flags := DT_SINGLELINE;
  if UseRightToLeftAlignment then
    Flags := DT_SINGLELINE or DT_RTLREADING;

  SpDrawXPGroupBox(Canvas, ClientRect, Caption, Flags, Enabled);
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXTextObject }

constructor TSpTBXTextObject.Create(AOwner: TComponent);
begin
  inherited;

  FImageIndex := -1;
  FImageChangeLink := TChangeLink.Create;
  FImageChangeLink.OnChange := ImageListChange;

  FLinkFont := TFont.Create;
  FLinkFont.Color := clBlue;
  FLinkFont.Style := [fsUnderline];

  FCaptionGlowColor := clYellow;

  Autosize := True;
  TabStop := True;
  SmartFocus := True;
  SpaceAsClick := True;
  FAlignment := taLeftJustify;
  FDrawPushedCaption := False;
  FThemeType := thtTBX;
end;

destructor TSpTBXTextObject.Destroy;
begin
  FImageChangeLink.Free;
  FLinkFont.Free;

  inherited;
end;

procedure TSpTBXTextObject.DefineProperties(Filer: TFiler);
begin
  inherited;
  // Don't let the VCL store the WideString, it doesn't write it correctly.
  // Use DefineProperties instead, with a new name for the property
  Filer.DefineProperty('CaptionW', ReadCaptionW, WriteCaptionW, FCaption <> '');
  Filer.DefineProperty('HintW', ReadHintW, WriteHintW, FHint <> '');
  Filer.DefineProperty('LinkTextW', ReadLinkTextW, WriteLinkTextW, FLinkText <> '');
  Filer.DefineProperty('LinkTextParamsW', ReadLinkTextParamsW, WriteLinkTextParamsW, FLinkTextParams <> '');
end;

procedure TSpTBXTextObject.ReadCaptionW(Reader: TReader);
begin
  case Reader.NextValue of
    vaLString, vaString:
      SetCaption(Reader.ReadString);
  else
    SetCaption(Reader.ReadWideString);
  end;
end;

procedure TSpTBXTextObject.WriteCaptionW(Writer: TWriter);
begin
  Writer.WriteWideString(FCaption);
end;

procedure TSpTBXTextObject.ReadHintW(Reader: TReader);
begin
  case Reader.NextValue of
    vaLString, vaString:
      SetHint(Reader.ReadString);
  else
    SetHint(Reader.ReadWideString);
  end;
end;

procedure TSpTBXTextObject.WriteHintW(Writer: TWriter);
begin
  Writer.WriteWideString(FHint);
end;

procedure TSpTBXTextObject.ReadLinkTextW(Reader: TReader);
begin
  case Reader.NextValue of
    vaLString, vaString:
      FLinkText := Reader.ReadString;
  else
    FLinkText := Reader.ReadWideString;
  end;
end;

procedure TSpTBXTextObject.WriteLinkTextW(Writer: TWriter);
begin
  Writer.WriteWideString(FLinkText);
end;

procedure TSpTBXTextObject.ReadLinkTextParamsW(Reader: TReader);
begin
  case Reader.NextValue of
    vaLString, vaString:
      FLinkTextParams := Reader.ReadString;
  else
    FLinkTextParams := Reader.ReadWideString;
  end;
end;

procedure TSpTBXTextObject.WriteLinkTextParamsW(Writer: TWriter);
begin
  Writer.WriteWideString(FLinkTextParams);
end;

procedure TSpTBXTextObject.AdjustFont(AFont: TFont);
begin
  if (FLinkText <> '') and MouseInControl then
    AFont.Assign(FLinkFont)
  else
    inherited;
end;

function TSpTBXTextObject.CanFocus: Boolean;
begin
  Result := False;
end;

procedure TSpTBXTextObject.Click;
begin
  if not (csLoading in ComponentState) then begin
    Invalidate;
    inherited;
    ExecuteLink;
  end;
end;

function TSpTBXTextObject.CanAutoSize(var NewWidth, NewHeight: Integer): Boolean;
var
  R, R1, R2: TRect;
begin
  if ([csReading, csLoading] * ComponentState = []) and AutoSize then begin
    GetSize(R, R1, R2);
    NewWidth := R.Right - R.Left;
    NewHeight := R.Bottom - R.Top;
    Result := True;
  end
  else
    Result := False;
end;

procedure TSpTBXTextObject.DoAdjustHeight(ACanvas: TCanvas; var NewHeight: Integer);
var
  R, R1, R2: TRect;
begin
  // DoAdjustHeight changes the Height of the control when Autosize is setted.
  // There should be an AdjustBounds method instead, like TCustomLabel.
  // We use this method to change the Width and the Height, NewHeight
  // should be the current Height

  GetSize(R, R1, R2);
  Width := R.Right - R.Left; // set the width
  NewHeight := R.Bottom - R.Top; // the height is setted by AdjustHeight
end;

procedure TSpTBXTextObject.DoDrawHint(AHintBitmap: TBitmap;
  var AHint: Widestring; var PaintDefault: Boolean);
begin
  if Assigned(FOnDrawHint) then FOnDrawHint(Self, AHintBitmap, AHint, PaintDefault);
end;

function TSpTBXTextObject.DoDrawItem(ACanvas: TCanvas; ARect: TRect;
  const PaintStage: TSpTBXPaintStage): Boolean;
begin
  Result := True;
  if Assigned(FOnDraw) then FOnDraw(Self, ACanvas, ARect, PaintStage, Result);
end;

function TSpTBXTextObject.DoDrawText(ACanvas: TCanvas; var Rect: TRect;
  Flags: Integer): Integer;
var
  PaintDefault: Boolean;
  I: integer;
  IL: TCustomImageList;
  R, R1, R2: TRect;
  WS: WideString;
begin
  Result := 0;
  WS := Caption;
  if Flags and DT_CALCRECT = 0 then begin
    ACanvas.Brush.Style := bsClear;
    PaintDefault := True;
    if Assigned(FOnDrawCaption) then
      FOnDrawCaption(Self, ACanvas, ClientRect, WS, Rect, False, pstPrePaint, PaintDefault);
    if PaintDefault then begin
      IL := FImages;
      I := FImageIndex;
      DoGetImageIndex(IL, I);
      Result := SpDrawXPText(ACanvas, Rect, WS, FCaptionGlow, FCaptionGlowColor,
        GetRealAlignment(Self), Flags, IL, I, ghlGlyphLeft, Enabled, DrawPushedCaption and Pushed, R1, R2);
    end;

    PaintDefault := True;
    if Assigned(FOnDrawCaption) then
      FOnDrawCaption(Self, ACanvas, ClientRect, WS, Rect, False, pstPostPaint, PaintDefault);
    if PaintDefault then
      if Focused then begin
        R := ClientRect;
        ApplyMargins(R, Margins);
        SpDrawFocusRect(ACanvas, GetFocusRect(R, R1, R2));
      end;
  end
  else
    Result := SpDrawXPText(ACanvas, WS, Rect, Flags);
end;

procedure TSpTBXTextObject.DoGetImageIndex(var AImageList: TCustomImageList; var AImageIndex: integer);
begin
  if Assigned(FOnGetImageIndex) then FOnGetImageIndex(Self, AImageList, AImageIndex);
end;

procedure TSpTBXTextObject.DoMouseEnter;
begin
  inherited;
  Invalidate;
end;

procedure TSpTBXTextObject.DoMouseLeave;
begin
  inherited;
  Invalidate;
end;

procedure TSpTBXTextObject.ExecuteLink;
var
  FilenameS, ParamsS: String;
begin
  if FLinkText <> '' then begin
    if Win32Platform = VER_PLATFORM_WIN32_WINDOWS then begin
      FilenameS := FLinkText;
      ParamsS := FLinkTextParams;
      ShellExecute(Application.Handle, 'open', PChar(FilenameS), PChar(ParamsS), '', SW_SHOWNORMAL);
    end
    else
      ShellExecuteW(Application.Handle, 'open', PWideChar(FLinkText), PWideChar(FLinkTextParams), '', SW_SHOWNORMAL);
  end;
end;

function TSpTBXTextObject.GetControlsAlignment: TAlignment;
begin
  Result := FAlignment;
end;

function TSpTBXTextObject.GetTextAlignment: TAlignment;
begin
  Result := FAlignment;
end;

function TSpTBXTextObject.GetTextFlags: Cardinal;
const
  Alignments: array [TAlignment] of Integer = (DT_LEFT, DT_RIGHT, DT_CENTER);
  WordWraps: array [TTextWrapping] of Integer = (DT_SINGLELINE,
    DT_SINGLELINE or DT_END_ELLIPSIS,
    DT_SINGLELINE or DT_PATH_ELLIPSIS, DT_WORDBREAK);
  ShowAccelChars: array [Boolean] of Integer = (DT_NOPREFIX, 0);
begin
  Result := DT_EXPANDTABS or WordWraps[Wrapping] or
    Alignments[GetRealAlignment(Self)] or ShowAccelChars[ShowAccelChar];
  Result := DrawTextBiDiModeFlags(Result);
end;

function TSpTBXTextObject.GetFocusRect(R, TextR, GlyphR: TRect): TRect;
begin
  if Caption = '' then
    Result := Rect(0, 0, 0, 0)
  else begin
    InflateRect(TextR, 1, 1);
    Result := TextR;
  end;
end;

function TSpTBXTextObject.GetPushed: Boolean;
begin
  Result := (inherited Pushed) and MouseInControl;
end;

procedure TSpTBXTextObject.GetSize(out TotalR, TextR, GlyphR: TRect);
var
  I: integer;
  IL: TCustomImageList;
  R: TRect;
begin
  // Size of Text + Glyph + TextMargin + Margins
  IL := FImages;
  I := FImageIndex;
  DoGetImageIndex(IL, I);

  R := ClientRect;
  if not Autosize then
    ApplyMargins(R, Margins);
  ApplyMargins(R, GetTextMargins);

  Canvas.Font.Assign(Font);
  AdjustFont(Canvas.Font);

  SpDrawXPText(Canvas, R, Caption, FALSE, clWhite, GetRealAlignment(Self), GetTextFlags,
    IL, I, ghlGlyphLeft, Enabled, DrawPushedCaption and Pushed, TextR, GlyphR, True);

  UnionRect(TotalR, TextR, GlyphR);
  if Autosize then
    with Margins do begin
      Inc(TotalR.Right, Left + Right);
      Inc(TotalR.Bottom, Top + Bottom);
    end;
end;

function TSpTBXTextObject.GetTextMargins: TRect;
begin
  Result := Rect(0, 0, 0, 0);
end;

procedure TSpTBXTextObject.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  // Fix a TBX bug, TTBXTextObject can't recieve focus if it's inside a Panel
  inherited;
  if SmartFocus and CanFocus then
    SetFocus;
end;

procedure TSpTBXTextObject.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited;
  if Operation = opRemove then
    if AComponent = Images then SetImages(nil);
end;

procedure TSpTBXTextObject.Paint;
var
  R, TextR: TRect;
begin
  R := ClientRect;
  ApplyMargins(R, Margins);

  Canvas.Font.Assign(Font);
  AdjustFont(Canvas.Font);

  if DoDrawItem(Canvas, R, pstPrePaint) then begin
    TextR := R;
    ApplyMargins(TextR, GetTextMargins);
    // Draw the Focus, Icon and Text
    DoDrawText(Canvas, TextR, GetTextFlags);
  end;

  DoDrawItem(Canvas, R, pstPostPaint);
end;

procedure TSpTBXTextObject.SetAlignment(const Value: TAlignment);
begin
  if FAlignment <> Value then begin
    FAlignment := Value;
    Invalidate;
  end;
end;

procedure TSpTBXTextObject.SetCaption(const Value: WideString);
begin
  if FCaption <> Value then begin
    FCaptionChanged := True;
    FCaption := Value;
    inherited Caption := Value;
  end;
end;

procedure TSpTBXTextObject.SetCaptionGlow(const Value: Boolean);
begin
  if FCaptionGlow <> Value then begin
    FCaptionGlow := Value;
    Invalidate;
  end;
end;

procedure TSpTBXTextObject.SetCaptionGlowColor(const Value: TColor);
begin
  if FCaptionGlowColor <> Value then begin
    FCaptionGlowColor := Value;
    Invalidate;
  end;
end;

procedure TSpTBXTextObject.SetChecked(Value: Boolean);
begin
  if Value <> FChecked then begin
    FChecked := Value;
    Invalidate;
  end;
end;

procedure TSpTBXTextObject.SetHint(const Value: WideString);
begin
  if FHint <> Value then begin
    FHint := Value;
    inherited Hint := Value;
  end;
end;

procedure TSpTBXTextObject.ImageListChange(Sender: TObject);
begin
  if Sender = Images then begin
    Invalidate;
    AdjustHeight;
  end;
end;

procedure TSpTBXTextObject.SetImageIndex(const Value: TImageIndex);
begin
  if FImageIndex <> Value then begin
    FImageIndex := Value;
    if Assigned(Images) then Invalidate;
    AdjustHeight;
  end;
end;

procedure TSpTBXTextObject.SetImages(const Value: TCustomImageList);
begin
  if FImages <> nil then FImages.UnRegisterChanges(FImageChangeLink);
  FImages := Value;
  if FImages <> nil then begin
    FImages.RegisterChanges(FImageChangeLink);
    FImages.FreeNotification(Self);
  end;
  Invalidate;
  AdjustHeight;
end;

procedure TSpTBXTextObject.SetLinkFont(const Value: TFont);
begin
  FLinkFont.Assign(Value);
end;

procedure TSpTBXTextObject.SetName(const Value: TComponentName);
var
  ChangeText: Boolean;
begin
  // Change the name to the unicode caption
  ChangeText := (csSetCaption in ControlStyle) and
    not (csLoading in ComponentState) and (Name = FCaption) and
    ((Owner = nil) or not (Owner is TControl) or
    not (csLoading in TControl(Owner).ComponentState));
  if ChangeText then FCaption := Value;
  inherited;
end;

procedure TSpTBXTextObject.SetThemeType(const Value: TSpTBXThemeType);
begin
  if Value <> FThemeType then begin
    FThemeType := Value;
    Invalidate;
  end;
end;

procedure TSpTBXTextObject.CMHintShow(var Message: TCMHintShow);
// Handle the CM_HINTSHOW message to show unicode hints using
// a custom THintWindow.
var
  HintInfo: PHintInfo;
  WideHint: Widestring;
  R, TextR: TRect;
  PaintDefault: Boolean;
begin
  HintInfo := Message.HintInfo;
  HintInfo.HintStr := '';
  HintInfo.CursorRect := ClientRect;

  // Prepare the HintInfo
  WideHint := FHint;
  HintInfo.HintWindowClass := TBitmapHint;   //custom HintWindow class
  HintInfo.HintData := SpStockHintBitmap;  //TApplication.ActivateHint will pass the data to the HintWindow
  HintInfo.HintStr := WideHint;
  //HintInfo.HideTimeout := 60000; //1 minute

  // Init HintBitmap
  SpStockHintBitmap.Canvas.Font.Assign(Self.Font);
  SpStockHintBitmap.Canvas.Font.Color := clInfoText;
  SpStockHintBitmap.Canvas.Pen.Color := clBlack;
  SpStockHintBitmap.Canvas.Brush.Color := clInfoBk;
  TextR := Rect(0, 0, 1, 1);
  SpDrawXPText(SpStockHintBitmap.Canvas, WideHint, TextR, DT_NOPREFIX or DT_CALCRECT);
  SpStockHintBitmap.Width := TextR.Right + 8;
  SpStockHintBitmap.Height := TextR.Bottom + 4;

  // Draw the hint in the HintBitmap
  PaintDefault := True;
  DoDrawHint(SpStockHintBitmap, WideHint, PaintDefault);
  if PaintDefault then begin
    R := Rect(0, 0, SpStockHintBitmap.Width, SpStockHintBitmap.Height);
    SpStockHintBitmap.Canvas.FillRect(R);
    OffsetRect(TextR, ((R.Right - TextR.Right) div 2) - 2, (R.Bottom - TextR.Bottom) div 2);
    SpDrawXPText(SpStockHintBitmap.Canvas, WideHint, TextR, DT_NOPREFIX);
  end;
end;

procedure TSpTBXTextObject.WMSetCursor(var Message: TWMSetCursor);
begin
  if not (csDesigning in ComponentState) and (Message.CursorWnd = Handle) and
    (FLinkText <> '') and MouseInControl and (Screen.Cursor = crDefault) then
  begin
    // Apply Hand cursor only if there is no other cursor assigned.
    if SpStockHandCursor <> 0 then
      Windows.SetCursor(SpStockHandCursor)
    else
      Windows.SetCursor(Screen.Cursors[Cursor]);
    Message.Result := 1;
  end
  else
    inherited;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXCustomLabel }

procedure TSpTBXCustomLabel.CMDialogChar(var Message: TCMDialogChar);
begin
  if Assigned(FFocusControl) and IsAccel(Message.CharCode, Caption) and FFocusControl.CanFocus then begin
    FFocusControl.SetFocus;
    Message.Result := 1;
  end
  else
    inherited;
end;

constructor TSpTBXCustomLabel.Create(AOwner: TComponent);
begin
  inherited;
  FUnderlineColor := clBtnShadow;
end;

function TSpTBXCustomLabel.DoDrawItem(ACanvas: TCanvas; ARect: TRect;
  const PaintStage: TSpTBXPaintStage): Boolean;
var
  C: TColor;
begin
  if (PaintStage = pstPrePaint) and FUnderline then begin
    C := ACanvas.Pen.Color;
    try
      ACanvas.Pen.Color := UnderlineColor;
      ACanvas.MoveTo(ARect.Left, ARect.Bottom - 1);
      ACanvas.LineTo(ARect.Right, ARect.Bottom - 1);
    finally
      ACanvas.Pen.Color := C;
    end;
  end;

  Result := inherited DoDrawItem(ACanvas, ARect, PaintStage);
end;

procedure TSpTBXCustomLabel.GetSize(out TotalR, TextR, GlyphR: TRect);
begin
  inherited GetSize(TotalR, TextR, GlyphR);
  if FUnderline then
    Inc(TotalR.Bottom);
end;

procedure TSpTBXCustomLabel.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if Operation = opRemove then
    if AComponent = FocusControl then SetFocusControl(nil);
end;

procedure TSpTBXCustomLabel.SetFocusControl(const Value: TWinControl);
begin
  if FFocusControl <> Value then
  begin
    FFocusControl := Value;
    if FFocusControl <> nil then FFocusControl.FreeNotification(Self);
  end;
end;

procedure TSpTBXCustomLabel.SetUnderline(const Value: Boolean);
begin
  if Value <> FUnderline then begin
    FUnderline := Value;
    Invalidate;
    AdjustHeight;
  end;
end;

procedure TSpTBXCustomLabel.SetUnderlineColor(const Value: TColor);
begin
  FUnderlineColor := Value;
  Invalidate;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXButtonControl }

function TSpTBXButtonControl.CanFocus: Boolean;
begin
  Result := True;
end;

procedure TSpTBXButtonControl.CMDialogChar(var Message: TCMDialogChar);
begin
  if Enabled and ShowAccelChar and IsAccel(Message.CharCode, Caption) and
    CanFocus and Visible then
  begin
    SetFocus;
    Click;
    Message.Result := 1;
  end
  else
    inherited;
end;

procedure TSpTBXButtonControl.CMDialogKey(var Message: TCMDialogKey);
begin
  if (Message.CharCode = VK_RETURN) and Focused and
    (KeyDataToShiftState(Message.KeyData) = []) then
  begin
    Click;
    Message.Result := 1;
  end
  else
    inherited;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXCustomCheckBox }

constructor TSpTBXCustomCheckBox.Create(AOwner: TComponent);
begin
  inherited;
  FAllowGrayed := False;
  FState := cbUnchecked;
end;

procedure TSpTBXCustomCheckBox.Click;
begin
  if FStateChanged then
    inherited
  else begin
    FStateChanged := True;
    try
      Toggle;
      inherited;
    finally
      FStateChanged := False;
    end;
  end;
end;

procedure TSpTBXCustomCheckBox.DoGetImageIndex(var AImageList: TCustomImageList;
  var AImageIndex: integer);
var
  B: TBitmap;
  R: TRect;
begin
  inherited DoGetImageIndex(AImageList, AImageIndex);
  if not Assigned(AImageList) then begin
    B := TBitmap.Create;
    try
      B.Width := 13;
      B.Height := 13;
      R := Rect(0, 0, B.Width, B.Height);
      B.Canvas.Brush.Color := clFuchsia;
      B.Canvas.FillRect(R);

      SpStockImageList.Clear;
      SpStockImageList.Width := B.Width;
      SpStockImageList.Height := B.Height;

      SpDrawXPCheckBoxGlyph(B.Canvas, R, Enabled, State, MouseInControl, Pushed, Focused, ThemeType);

      SpStockImageList.AddMasked(B, clFuchsia);
      AImageList := SpStockImageList;
      AImageIndex := 0;
    finally
      B.Free;
    end;
  end;
end;

procedure TSpTBXCustomCheckBox.SetChecked(Value: Boolean);
begin
  if Checked <> Value then begin
    inherited;
    if Checked then FState := cbChecked
    else FState := cbUnchecked;
    if not FStateChanged then begin
      FStateChanged := True;
      try
        Click;
      finally
        FStateChanged := False;
      end;
    end;
  end;
end;

procedure TSpTBXCustomCheckBox.SetState(const Value: TCheckBoxState);
begin
  if (FState <> Value) then begin
    FState := Value;
    FStateChanged := True;
    try
      SetChecked(FState = cbChecked);
      Click;
    finally
      FStateChanged := False;
    end;
  end;
end;

procedure TSpTBXCustomCheckBox.Toggle;
begin
  case State of
    cbUnchecked: if AllowGrayed then FState := cbGrayed else FState := cbChecked;
    cbChecked: FState := cbUnchecked;
    cbGrayed: FState := cbChecked;
  end;
  SetChecked(FState = cbChecked);
end;

procedure TSpTBXCustomCheckBox.GetSize(out TotalR, TextR, GlyphR: TRect);
begin
  inherited GetSize(TotalR, TextR, GlyphR);
  // Inc TotalR for the focusrect  
  if Autosize then begin
    Inc(TotalR.Right);
    Inc(TotalR.Bottom, 2);
  end;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXCustomRadioButton }

procedure TSpTBXCustomRadioButton.Click;
begin
  if FStateChanged then
    inherited
  else
    if not Checked then Toggle; // Toggle will call OnClick
end;

procedure TSpTBXCustomRadioButton.CMFocusChanged(var Message: TCMFocusChanged);
begin
  inherited;
  if Focused then
    Toggle;
end;

procedure TSpTBXCustomRadioButton.DoGetImageIndex(var AImageList: TCustomImageList;
  var AImageIndex: integer);
var
  B: TBitmap;
  R, TR, GR: TRect;
begin
  inherited DoGetImageIndex(AImageList, AImageIndex);
  if not Assigned(AImageList) then begin
    B := TBitmap.Create;
    try
      B.Width := 13;
      B.Height := 13;
      SpStockImageList.Clear;
      SpStockImageList.Width := B.Width;
      SpStockImageList.Height := B.Height;

      // The WindowsXP radiobutton glyph blends the background
      if Color = clNone then begin
        R := ClientRect;
        ApplyMargins(R, Margins);
        ApplyMargins(R, GetTextMargins);
        B.Canvas.Font.Assign(Font);
        SpDrawXPText(Canvas, R, Caption, FALSE, clWhite, GetRealAlignment(Self), GetTextFlags,
          B, ghlGlyphLeft, Enabled, DrawPushedCaption and Pushed, TR, GR, True);
        TBX.DrawParentBackground(Self, B.Canvas.Handle, GR);
        R := Rect(0, 0, B.Width, B.Height);
      end
      else begin
        R := Rect(0, 0, B.Width, B.Height);
        B.Canvas.Brush.Color := Color;
        B.Canvas.FillRect(R);
      end;
      SpDrawXPRadioButtonGlyph(B.Canvas, R, Enabled, Checked, MouseInControl,
        Pushed, Focused, ThemeType);

      SpStockImageList.Add(B, nil);
      AImageList := SpStockImageList;
      AImageIndex := 0;
    finally
      B.Free;
    end;
  end;
end;

procedure TSpTBXCustomRadioButton.GetSize(out TotalR, TextR, GlyphR: TRect);
begin
  inherited GetSize(TotalR, TextR, GlyphR);
  // Inc TotalR for the focusrect
  if Autosize then begin
    Inc(TotalR.Right);
    Inc(TotalR.Bottom, 2);
  end;
end;

procedure TSpTBXCustomRadioButton.SetChecked(Value: Boolean);
begin
  if Value and (Checked <> Value) then begin
    inherited;
    TurnSiblingsOff;
    FStateChanged := True;
    try
      Click;
    finally
      FStateChanged := False;
    end;
  end
  else
    inherited;
end;

procedure TSpTBXCustomRadioButton.Toggle;
begin
  if not Checked then Checked := True;
end;

procedure TSpTBXCustomRadioButton.TurnSiblingsOff;
var
  I: Integer;
  C: TControl;
begin
  if Assigned(Parent) then
    for I := 0 to Parent.ControlCount - 1 do begin
      C := Parent.Controls[I];
      if (C <> Self) and (C is TSpTBXCustomRadioButton) then
        TSpTBXCustomRadioButton(C).Checked := False;
    end;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXCustomButton }

constructor TSpTBXCustomButton.Create(AOwner: TComponent);
begin
  inherited;
  FBitmapTransparent := True;
  FBitmap := TBitmap.Create;
  FBitmap.OnChange := BitmapChanged;

  Alignment := taCenter;
  Autosize := False;
  DrawPushedCaption := True;
  ThemeType := thtWindows;
  Width := 75;
  Height := 25;
end;

destructor TSpTBXCustomButton.Destroy;
begin
  FBitmap.Free;
  inherited;
end;

procedure TSpTBXCustomButton.AdjustFont(AFont: TFont);
var
  ItemInfo: TTBXItemInfo;
begin
  inherited;
  if FThemeType = thtTBX then begin
    SpFillItemInfo(Enabled, Pushed, MouseInControl, Checked, ItemInfo);
    Canvas.Font.Color := CurrentTheme.GetItemTextColor(ItemInfo);
  end;
end;

procedure TSpTBXCustomButton.BitmapChanged(Sender: TObject);
begin
  Invalidate;
end;

function TSpTBXCustomButton.BitmapValid: boolean;
begin
  Result := (Bitmap <> nil) and (not Bitmap.Empty) and (Bitmap.Height mod ConstStatesCount = 0);
end;

procedure TSpTBXCustomButton.Click;
var
  Form: TCustomForm;
begin
  Form := GetParentForm(Self);
  if Assigned(Form) then Form.ModalResult := FModalResult;
  inherited;
end;

function TSpTBXCustomButton.DoDrawItem(ACanvas: TCanvas; ARect: TRect;
  const PaintStage: TSpTBXPaintStage): Boolean;
var
  ItemInfo: TTBXItemInfo;
  B: TBitmap;
  R: TRect;
begin
  Result := inherited DoDrawItem(ACanvas, ARect, PaintStage);

  if Result and (PaintStage = pstPrePaint) then begin
    // Draw the parent background, the parent must handle WM_ERASEBKGND
    if ThemeType <> thtNone then
      TBX.DrawParentBackground(Self, ACanvas.Handle, ARect);

    if BitmapValid then begin
      B := TBitmap.Create;
      try
        B.Width := ARect.Right - ARect.Left;
        B.Height := ARect.Bottom - ARect.Top;
        SetStretchBltMode(B.Canvas.Handle, COLORONCOLOR);
        B.Canvas.CopyRect(ARect, Bitmap.Canvas, GetSkinStateRect);
        if FBitmapTransparent then
          B.Transparent := True;
        ACanvas.Draw(0, 0, B);
      finally
        B.Free;
      end;
    end
    else begin
      Case ThemeType of
        thtNone:
          SpDrawXPButton(ACanvas, ARect, Enabled, Checked or Pushed,
            MouseInControl, false, FDefault or Focused, False);
        thtWindows:
          SpDrawXPButton(ACanvas, ARect, Enabled, Checked or Pushed,
            MouseInControl, false, FDefault or Focused, True);
        thtTBX:
          if MouseInControl and not Pushed then begin
            B := TBitmap.Create;
            try
              B.Width := ARect.Right - ARect.Left;
              B.Height := ARect.Bottom - ARect.Top;
              R := Rect(0, 0, B.Width, B.Height);
              B.Canvas.CopyRect(R, Canvas, ARect);

              SpFillItemInfo(Enabled, Pushed, True, Checked, ItemInfo);
              CurrentTheme.PaintButton(B.Canvas, R, ItemInfo);

              SpHighlightBitmap(ACanvas, ARect.Left, ARect.Top, B, clNone, clWhite, 180);
            finally
              B.Free;
            end;
          end
          else begin
            SpFillItemInfo(Enabled, Pushed, True, Checked, ItemInfo);
            CurrentTheme.PaintButton(ACanvas, ARect, ItemInfo);
          end;
      end;
    end;
  end;
end;

function TSpTBXCustomButton.GetFocusRect(R, TextR, GlyphR: TRect): TRect;
begin
  Result := R;
  if FThemeType = thtNone then
    InflateRect(Result, -4, -4)
  else
    InflateRect(Result, -3, -3);
end;

function TSpTBXCustomButton.GetTextMargins: TRect;
begin
  Result := Rect(8, 2, 8, 2);
end;

function TSpTBXCustomButton.GetSkinStateRect: TRect;
var
  W, H: integer;
begin
  // Finds the skin rect based on the button state
  Result := Rect(0, 0, 0, 0);

  if BitmapValid then begin
    W := Bitmap.Width;
    H := (Bitmap.Height div ConstStatesCount); // 4 states
    if not Enabled then
      Result := Bounds(0, H * 3, W, H)  // 4th state (disabled)
    else begin
      if Checked or (Pushed and MouseInControl) then
        Result := Bounds(0, H * 2, W, H)  // 3rd state (down)
      else
        if MouseInControl then
          Result := Bounds(0, H * 1, W, H)  // 2nd state (hottrack)
        else
          Result := Bounds(0, H * 0, W, H); // 1st state (up)
    end;
  end;
end;

procedure TSpTBXCustomButton.SetBitmap(const Value: TBitmap);
begin
  FBitmap.Assign(Value);
  Invalidate;
end;

procedure TSpTBXCustomButton.SetDefault(const Value: Boolean);
begin
  if Value <> FDefault then begin
    FDefault := Value;
    Invalidate;
  end;
end;

procedure TSpTBXCustomButton.CMDialogKey(var Message: TCMDialogKey);
begin
  with Message do
    if (((CharCode = VK_RETURN) and FActive) or
      ((CharCode = VK_ESCAPE) and FCancel)) and
      (KeyDataToShiftState(Message.KeyData) = []) and CanFocus then
    begin
      Click;
      Result := 1;
    end else
      inherited;
end;

procedure TSpTBXCustomButton.CMFocusChanged(var Message: TCMFocusChanged);
begin
  with Message do
    if Sender is TSpTBXCustomButton then
      FActive := Sender = Self
    else
      FActive := FDefault;
  inherited;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXCustomProgressBar }

constructor TSpTBXCustomProgressBar.Create(AOwner: TComponent);
begin
  inherited;
  FMax := 100;
  FMin := 0;
  FPosition := 0;
  FCaptionGlow := True;
  FCaptionType := pctPercentage;

  Color := clBtnFace;
  Alignment := taCenter;
  Autosize := False;
  Width := 150;
  Height := 17;
  ThemeType := thtWindows;
  Font.Style := Font.Style + [fsBold];    
end;

function TSpTBXCustomProgressBar.DoDrawItem(ACanvas: TCanvas; ARect: TRect;
  const PaintStage: TSpTBXPaintStage): Boolean;
var
  I: Integer;
begin
  Result := inherited DoDrawItem(ACanvas, ARect, PaintStage);

  if Result and (PaintStage = pstPrePaint) then begin
    // Draw the parent background, the parent must handle WM_ERASEBKGND
    if ThemeType = thtTBX then
      DrawParentBackground(Self, ACanvas.Handle, ARect);
    I := SpDrawXPProgressBar(ACanvas, ARect, FVertical, FSmooth, FMin, FMax, FPosition, ThemeType);
    Case FCaptionType of
      pctNone: Caption := '';
      pctPercentage: Caption := IntToStr(I) + '%';
      pctProgress: Caption := IntToStr(FPosition);
    end;
  end;
end;

procedure TSpTBXCustomProgressBar.DoProgressChange;
begin
  if Assigned(FOnProgressChange) then FOnProgressChange(Self, Position);
end;

function TSpTBXCustomProgressBar.GetTextMargins: TRect;
begin
  Result := Rect(8, 2, 8, 2);
end;

procedure TSpTBXCustomProgressBar.SetCaptionType(const Value: TSpTBXProgressCaption);
begin
  if FCaptionType <> Value then begin
    FCaptionType := Value;
    if Value <> pctDefault then Caption := '';
    Invalidate;
  end;
end;

procedure TSpTBXCustomProgressBar.SetMax(const Value: integer);
begin
  if FMax <> Value then begin
    FMax := Value;
    Invalidate;
  end;
end;

procedure TSpTBXCustomProgressBar.SetMin(const Value: integer);
begin
  if FMin <> Value then begin
    FMin := Value;
    Invalidate;
  end;
end;

procedure TSpTBXCustomProgressBar.SetPosition(Value: integer);
begin
  if Value > FMax then Value := FMax
  else if Value < FMin then Value := FMin;
  if FPosition <> Value then begin
    FPosition := Value;
    Invalidate;
    DoProgressChange;    
  end;
end;

procedure TSpTBXCustomProgressBar.SetSmooth(const Value: Boolean);
begin
  if FSmooth <> Value then begin
    FSmooth := Value;
    Invalidate;
  end;
end;

procedure TSpTBXCustomProgressBar.SetVertical(const Value: Boolean);
begin
  if FVertical <> Value then begin
    FVertical := Value;
    SetBounds(Left, Top, Height, Width);
    Invalidate;
  end;
end;

procedure TSpTBXCustomProgressBar.StepIt(Delta: Integer = 1);
begin
  SetPosition(FPosition + Delta);
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM

procedure InitializeStock;
begin
  SpStockImageList := TImageList.Create(nil);
end;

procedure FinalizeStock;
begin
  SpStockImageList.Free;
end;

initialization
  InitializeStock;

finalization
  FinalizeStock;

end.
