unit SpTBXItem;

{==============================================================================
Version 0.4

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License.
You may obtain a copy of the License at http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The initial developer of this code is Robert Lee.

Requirements:
  - Jordan Russell's Toolbar 2000
    http://www.jrsoftware.org
  - Alex Denisov's TBX
    http://g32.org

Wish list for TB2K/TBX:
  - It would be nice to have a way to get the TTBItemViewer width and height
    even when the Item is not visible (TTBItemViewer.BoundsRect is not valid
    when the item is not visible).
  - It would be nice to allow component writers to override
    TTBCustomToolbar.CreateWrapper so they can create custom TTBControlItem
    items descendants as the dropped control wrappers.
  - Fix this TBX bug: the Caption property of TBX items doesn't work properly
    at design time. Clicking on the 'more' button raises an AV: "Control '' has
    no parent window".
    The same happens with the ChevronHint property of the toolbars.

Development notes:
  - All the TBX theme changes and adjustments are marked with '[TBXTheme-Change]'.
  - Windows XP theme data parts that are not loaded by TBX are loaded outside
    TBXThemeManager space, they must be handled by this unit only.
  - UTF8Decode is not supported on Delphi 5, it was introduced on D6.
  - TSpTBXCompoundItemsControl is used as the base class for TB2K items enabled
    Controls, it uses the ITBItems interface and streams the items to the DFM.
  - When a control is dropped on the toolbar a TTBControlItem is created by
    TTBCustomToolbar.CreateWrapper, unfortunately it is created with the
    Toolbar.Owner instead of the Form (Owner.Owner for TSpTBXCompoundItemsControl
    like the TSpTBXTabSet or TSpTBXStatusBar). The workaround is to handle the
    CM_CONTROLCHANGE message in the compound toolbar, stream all the
    TTBControlItem.Control to the DFM (the TSpTBXCompoundItemsControl must be
    the parent in the DFM), and finally reset the TTBControlItem.Control
    parentship to the Toolbar in TSpTBXCompoundItemsControl.Loaded.

To Do:
  - Rotated caption painting.
  - Anchored items on vertical toolbars.

Known Issues:
  -

History:
21 July 2004 - version 0.4
  - Fixed TTBControlItem.Control streaming bug on TSpTBXStatusBar.
  - Fixed bad sync of the items unicode caption and hint properties
    when an Action was assigned.

12 July 2004 - version 0.3.1
  - Unchanged.

9 July 2004 - version 0.3
  - Fixed incorrect TSpTBXItem caption painting when DisplayMode
    was nbdmDefault, thanks to Cyril for reporting this.
  - Added anchors support for TTBControlItem items, if the associated
    Control is client aligned or has akRight in its Anchors property.
  - The theme items in TSpTBXThemeGroupItem are now sorted.
  - Added OnUpdate event to TSpTBXThemeGroupItem, this event is fired
    every time the theme items list is recreated, use this event
    to sort or change the items properties.

27 June 2004 - version 0.2
  - Fixed Toolbar custom painting event.
  - Fixed incorrect Shortcut painting in submenus.
  - Fixed incorrect Shortcut hint painting.
  - Removed thtBitmapSkin from TSpTBXThemeType.
  - Published more properties for TSpTBXLabelItem.
  - New Toolbar item added, TSpTBXSeparator.
  - New component added, TSpTBXStatusBar: a fully customizable
    StatusBar with Unicode text and TBX themes support.

22 June 2004 - version 0.1
  - Initial release.

==============================================================================}

interface

{$I TB2Ver.inc}

uses
  Windows, Messages, Classes, SysUtils, Controls, Graphics, ImgList, Forms,
  Menus, ActnList, TB2Common, TB2Item, TB2Dock, TB2Toolbar,
  TBX, TBXThemes;

type
  TSpTBXStatusToolbar = class;

  TSpTBXThemeType = (
    thtNone,
    thtWindows,
    thtTBX
  );

  TSpTBXPaintStage = (
    pstPrePaint,
    pstPostPaint
  );

  TSpGlyphLayout = (
    ghlGlyphLeft,
    ghlGlyphRight,
    ghlGlyphTop,
    ghlGlyphBottom
  );

  TSpTBXGetImageIndexEvent = procedure(Sender: TObject;
    var AImageList: TCustomImageList; var AItemIndex: integer) of object;

  TSpTBXDrawEvent = procedure(Sender: TObject; ACanvas: TCanvas;
    ARect: TRect; const PaintStage: TSpTBXPaintStage;
    var PaintDefault: Boolean) of object;

  TSpTBXDrawImageEvent = procedure(Sender: TObject; ACanvas: TCanvas;
    const ItemInfo: TTBXItemInfo; const PaintStage: TSpTBXPaintStage;
    AImageList: TCustomImageList; var AImageIndex: integer;
    var ARect: TRect; var PaintDefault: Boolean) of object;

  TSpTBXDrawItemEvent = procedure(Sender: TObject; ACanvas: TCanvas;
    ARect: TRect; ItemInfo: TTBXItemInfo; const PaintStage: TSpTBXPaintStage;
    var PaintDefault: Boolean) of object;

  TSpTBXDrawTextEvent = procedure(Sender: TObject; ACanvas: TCanvas;
    ClientAreaRect: TRect; var ACaption: WideString; var CaptionRect: TRect;
    IsTextRotated: Boolean; const PaintStage: TSpTBXPaintStage;
    var PaintDefault: Boolean) of object;

  TSpTBXDrawHintEvent = procedure(Sender: TObject; AHintBitmap: TBitmap;
    var AHint: WideString; var PaintDefault: Boolean) of object;

  TSpTBXItemNotificationEvent = procedure(Sender: TObject;
    Ancestor: TTBCustomItem; Relayed: Boolean; Action: TTBItemChangedAction;
    Index: Integer; Item: TTBCustomItem) of object;

  TSpTBXGetThemeCaptionEvent = procedure(Sender: TObject;
    var ThemeName: WideString) of object;

  // Method called by the GetNext routine for each item
  TSpTBXGetNextCallBack = function(AItemViewer: TTBItemViewer; Default: Boolean; var EndProcessing: Boolean): Boolean of object;

  { TSpTBXCustomItemActionLink }

  TSpTBXCustomItemActionLink = class(TTBCustomItemActionLink)
  protected
    procedure AssignClient(AClient: TObject); override;
    procedure SetCaption(const Value: String); override;
    procedure SetHint(const Value: String); override;
  end;

  { TSpTBXItem }

  TSpTBXItemViewer = class;

  TSpTBXCustomItem = class(TTBXCustomItem)
  private
    FCaption: WideString;
    FHint: WideString;
    FAlignment: TAlignment;
    FCustomWidth: Integer;
    FCustomHeight: Integer;
    FControl: TControl;
    FAnchored: Boolean;
    FOnDrawCaption: TSpTBXDrawTextEvent;
    FOnDrawHint: TSpTBXDrawHintEvent;
    FOnDrawItem: TSpTBXDrawItemEvent;
    FOnDrawImage: TSpTBXDrawImageEvent;
    procedure SetCaption(const Value: WideString);
    procedure ReadCaptionW(Reader: TReader);
    procedure WriteCaptionW(Writer: TWriter);
    procedure SetHint(const Value: WideString);
    procedure ReadHintW(Reader: TReader);
    procedure WriteHintW(Writer: TWriter);
    procedure SetAlignment(const Value: TAlignment);
    procedure SetCustomWidth(Value: Integer);
    procedure SetCustomHeight(Value: Integer);
    procedure SetControl(const Value: TControl);
    procedure SetAnchored(const Value: Boolean);
  protected
    procedure ActionChange(Sender: TObject; CheckDefaults: Boolean); override;
    function DialogChar(CharCode: Word): Boolean; virtual;
    procedure DefineProperties(Filer: TFiler); override;
    procedure DoDrawHint(AHintBitmap: TBitmap; var AHint: Widestring; var PaintDefault: Boolean); virtual;
    procedure DoDrawButton(ACanvas: TCanvas; ARect: TRect; ItemInfo: TTBXItemInfo;
      const PaintStage: TSpTBXPaintStage; var PaintDefault: Boolean); virtual;
    procedure DoDrawCaption(ACanvas: TCanvas; const ClientAreaRect: TRect;
      var ACaption: WideString; var CaptionRect: TRect; IsTextRotated: Boolean;
      const PaintStage: TSpTBXPaintStage; var PaintDefault: Boolean); virtual;
    procedure DoDrawImage(ACanvas: TCanvas; const ItemInfo: TTBXItemInfo;
      const PaintStage: TSpTBXPaintStage; AImageList: TCustomImageList;
      var AImageIndex: integer; var ARect: TRect; var PaintDefault: Boolean); virtual;
    function GetActionLinkClass: TTBCustomItemActionLinkClass; override;
    function GetItemViewerClass(AView: TTBView): TTBItemViewerClass; override;
    procedure ToggleControl; virtual;
    procedure UpdateProps; virtual;
    procedure Notification (AComponent: TComponent; Operation: TOperation); override;

    property Alignment: TAlignment read FAlignment write SetAlignment default taCenter;
    property Anchored: Boolean read FAnchored write SetAnchored default False;
    property Control: TControl read FControl write SetControl;
    property CustomWidth: Integer read FCustomWidth write SetCustomWidth default -1;
    property CustomHeight: Integer read FCustomHeight write SetCustomHeight default -1;
    property OnDrawHint: TSpTBXDrawHintEvent read FOnDrawHint write FOnDrawHint;
    property OnDrawItem: TSpTBXDrawItemEvent read FOnDrawItem write FOnDrawItem;
    property OnDrawImage: TSpTBXDrawImageEvent read FOnDrawImage write FOnDrawImage;
    property OnDrawCaption: TSpTBXDrawTextEvent read FOnDrawCaption write FOnDrawCaption;
  public
    constructor Create(AOwner: TComponent); override;
    procedure Click; override;
    procedure InitiateAction; override;
  published
    // Don't let the VCL store the WideString, it doesn't write it correctly.
    // Use DefineProperties instead
    property Caption: WideString read FCaption write SetCaption stored False; // Hides the inherited Caption
    property Hint: WideString read FHint write SetHint stored False; // Hides the inherited Hint
  end;

  TSpTBXItemViewer = class(TTBXItemViewer)
  private
    function GetItem: TSpTBXCustomItem;
  protected
    FAnchorSize: TPoint;
    FAnchorDelta: TPoint;
    procedure DoDrawButton(ACanvas: TCanvas; ARect: TRect; ItemInfo: TTBXItemInfo;
      const PaintStage: TSpTBXPaintStage; var PaintDefault: Boolean); virtual;
    procedure DoDrawCaption(ACanvas: TCanvas; const ClientAreaRect: TRect;
      var ACaption: WideString; var CaptionRect: TRect; IsTextRotated: Boolean;
      const PaintStage: TSpTBXPaintStage; var PaintDefault: Boolean); reintroduce; virtual;
    procedure DoDrawImage(ACanvas: TCanvas; const ItemInfo: TTBXItemInfo;
      const PaintStage: TSpTBXPaintStage; AImageList: TCustomImageList;
      var AImageIndex: integer; var ARect: TRect; var PaintDefault: Boolean); virtual;

    procedure DrawItemImage(Canvas: TCanvas; ARect: TRect; ItemInfo: TTBXItemInfo); override;
    procedure CalcSize(const Canvas: TCanvas; var AWidth, AHeight: Integer); override;
    function  GetTextSize(Canvas: TCanvas; const Text: string; TextFlags: Cardinal;
      Rotated: Boolean; StateFlags: Integer): TSize; override;
    procedure Paint(const Canvas: TCanvas; const ClientAreaRect: TRect;
      IsHoverItem: Boolean; IsPushed: Boolean; UseDisabledShadow: Boolean); override;
  public
    function GetHintText: Widestring;             // Hides the inherited TB2K GetHintText function
    property Item: TSpTBXCustomItem read GetItem; // Hides the inherited TB2K Item property
  end;

  TSpTBXItem = class(TSpTBXCustomItem)
  published
    property Action;
    property Checked;
    property DisplayMode default nbdmImageAndText;
    property Enabled;
    property FontSettings;
    property GroupIndex;
    property HelpContext;
    property ImageIndex;
    property Images;
    property InheritOptions;
    property Layout;
    property MaskOptions;
    property MinHeight;
    property MinWidth;
    property Options;
    property ShortCut;
    property Stretch default True;
    property Visible;
    property OnAdjustFont;
    // property OnDrawImage; use custom OnDrawImage
    property OnClick;
    property OnSelect;
    // TSpTBXCustomItem properties
    property Alignment;
    property Anchored;
    property Control;
    property CustomWidth;
    property CustomHeight;
    property OnDrawCaption;
    property OnDrawHint;
    property OnDrawImage;
    property OnDrawItem;
  end;

  { TSpTBXSubmenuItem }

  TSpTBXSubmenuItem = class(TSpTBXItem)
  private
    function GetDropdownCombo: Boolean;
    procedure SetDropdownCombo(Value: Boolean);
  public
    constructor Create(AOwner: TComponent); override;
  published
    property AlwaysSelectFirst;
    property DropdownCombo: Boolean read GetDropdownCombo write SetDropdownCombo default False;
    property LinkSubitems;
    property SubMenuImages;
    property ToolBoxPopup;
    property OnPopup;
  end;

  { TSpTBXLabelItem }

  TSpTBXCustomLabelItem = class(TSpTBXCustomItem)
  protected
    function DialogChar(CharCode: Word): Boolean; override;
    procedure DoDrawButton(ACanvas: TCanvas; ARect: TRect; ItemInfo: TTBXItemInfo;
      const PaintStage: TSpTBXPaintStage; var PaintDefault: Boolean); override;
    procedure ToggleControl; override;
    procedure UpdateProps; override;
    property Alignment default taLeftJustify;
  public
    constructor Create(AOwner: TComponent); override;
  end;

  TSpTBXLabelItem = class(TSpTBXCustomLabelItem)
  published
    property ImageIndex;
    property Images;
    property InheritOptions;
    property Layout;
    property FontSettings;
    property MaskOptions;
    property MinHeight;
    property MinWidth;
    property Options;
    property Visible;
    property OnAdjustFont;
    // TSpTBXCustomItem properties
    property Alignment;
    property Anchored;
    property Control;
    property CustomWidth;
    property CustomHeight;
    property OnDrawCaption;
    property OnDrawHint;
    property OnDrawImage;
    property OnDrawItem;
  end;

  { TSpTBXSeparatorItem }

  TSpTBXSeparatorItem = class(TTBXSeparatorItem)
  public
    function GetItemViewerClass(AView: TTBView): TTBItemViewerClass; override;
  end;

  TSpTBXSeparatorItemViewer = class(TTBXSeparatorItemViewer)
  protected
    procedure CalcSize(const Canvas: TCanvas; var AWidth, AHeight: Integer); override;
    function IsStatusBarSeparator(out T: TSpTBXStatusToolbar): Boolean;
    procedure Paint(const Canvas: TCanvas; const ClientAreaRect: TRect;
      IsHoverItem, IsPushed, UseDisabledShadow: Boolean); override;
  end;

  { TSpTBXRightAlignSpacerItem }

  TSpTBXRightAlignSpacerItem = class(TSpTBXCustomLabelItem)
  published
    property FontSettings;
    property MaskOptions;
    property Options;
    // TSpTBXCustomItem properties
    property Alignment;
    property CustomWidth;
    property CustomHeight;
    property OnDrawCaption;
  end;

  { TSpTBXThemeGroupItem }

  TSpTBXThemeGroupItem = class(TTBGroupItem)
  private
    FInsertIndex: Integer;
    FOnClick: TNotifyEvent;
    FOnGetCaption: TSpTBXGetThemeCaptionEvent;
    FOnUpdate: TNotifyEvent;
  protected
    FThemesList: TStringList;
    procedure ItemClickEvent(Sender: TObject); virtual;
    procedure DoClick; virtual;
    function DoGetItemCaption(ThemeName: WideString): WideString; virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure RecreateThemesList; virtual;
  published
    property InsertIndex: Integer read FInsertIndex write FInsertIndex default 0;
    property OnClick: TNotifyEvent read FOnClick write FOnClick;
    property OnGetCaption: TSpTBXGetThemeCaptionEvent read FOnGetCaption write FOnGetCaption;
    property OnUpdate: TNotifyEvent read FOnUpdate write FOnUpdate;
  end;

  { TSpTBXItemCacheList }

  TSpTBXItemCache = class(TCollectionItem)
  private
    FName: TComponentName;
    FItem: TTBCustomItem;
    FLeft, FTop, FWidth, FHeight: Integer;
    FParentWidth, FParentHeight: Integer;
    function GetName: TComponentName;
  public
    procedure Assign(Source: TPersistent); override;
    property Item: TTBCustomItem read FItem write FItem;
  published
    property Name: TComponentName read GetName write FName;
    property Left: Integer read FLeft write FLeft default 0;
    property Top: Integer read FTop write FTop default 0;
    property Width: Integer read FWidth write FWidth default 0;
    property Height: Integer read FHeight write FHeight default 0;
    property ParentWidth: Integer read FParentWidth write FParentWidth default 0;
    property ParentHeight: Integer read FParentHeight write FParentHeight default 0;
  end;

  TSpTBXItemCacheCollection = class(TCollection)
  private
    function GetItem(Index: Integer): TSpTBXItemCache;
    procedure SetItem(Index: Integer; const Value: TSpTBXItemCache);
  public
    function Add: TSpTBXItemCache; overload; virtual;
    function Add(Item: TTBCustomItem; RBounds: TRect): Integer; overload;
    function IndexOf(Item: TTBCustomItem): Integer;
    property Items[Index: Integer]: TSpTBXItemCache read GetItem write SetItem; default;
  end;

  { TSpTBXDock }

  TSpTBXDock = class(TTBXDock)
  private
    FOnDrawBackground: TSpTBXDrawEvent;
    FPrevWidth: Integer;
    FPrevHeight: Integer;
  protected
    function CanResize(var NewWidth: Integer; var NewHeight: Integer): Boolean; override;
    procedure DoDrawBackground(ACanvas: TCanvas; ARect: TRect;
      const PaintStage: TSpTBXPaintStage; var PaintDefault: Boolean);
    function ThemedBackground: Boolean; override;
    procedure Resize; override;
  public
    procedure DrawBackground(DC: HDC; const DrawRect: TRect); override;
    property PrevWidth: Integer read FPrevWidth;
    property PrevHeight: Integer read FPrevHeight;
  published
    property OnCanResize;
    property OnDrawBackground: TSpTBXDrawEvent read FOnDrawBackground write FOnDrawBackground;
  end;

  TSpTBXDockClass = class of TSpTBXDock;

  { TSpTBXToolbar }

  TSpTBXToolbar = class(TTBXToolbar)
  private
    FCompoundToolbar: Boolean;
    FIsResizing: Boolean;
    FOnDrawHint: TSpTBXDrawHintEvent;
    FOnDrawBackground: TSpTBXDrawEvent;
    FOnItemNotification: TSpTBXItemNotificationEvent;
    function CreateWrapper(Index: Integer; Ctl: TControl): TTBControlItem;
    function IsAnchoredControlItem(Item: TTBCustomItem): TTBControlItem;
    procedure CMControlChange(var Message: TCMControlChange); message CM_CONTROLCHANGE;
    procedure CMDialogChar(var Message: TCMDialogChar); message CM_DIALOGCHAR;
    procedure CMHintShow(var Message: TCMHintShow); message CM_HINTSHOW;
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
  protected
    FAnchoredControlItems: TSpTBXItemCacheCollection;
    procedure AnchorItems; virtual;
    procedure RightAlignItems; virtual;
    function GetRightAlignMargin: Integer; virtual;
    procedure DoDrawBackground(ACanvas: TCanvas; ARect: TRect;
      const PaintStage: TSpTBXPaintStage; var PaintDefault: Boolean); virtual;
    procedure DoDrawHint(AHintBitmap: TBitmap; var AHint: Widestring;
      var PaintDefault: Boolean); virtual;
    procedure DoItemNotification(Ancestor: TTBCustomItem; Relayed: Boolean;
      Action: TTBItemChangedAction; Index: Integer; Item: TTBCustomItem); virtual;
    procedure DrawNCArea(const DrawToDC: Boolean; const ADC: HDC; const Clip: HRGN); override;
    procedure Resize; override;
    property CompoundToolbar: Boolean read FCompoundToolbar write FCompoundToolbar;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property IsResizing: Boolean read FIsResizing;
  published
    property OnDrawBackground: TSpTBXDrawEvent read FOnDrawBackground write FOnDrawBackground;
    property OnDrawHint: TSpTBXDrawHintEvent read FOnDrawHint write FOnDrawHint;
    property OnItemNotification: TSpTBXItemNotificationEvent read FOnItemNotification write FOnItemNotification;
  end;

  TSpTBXToolbarClass = class of TSpTBXToolbar;

  { TSpTBXCompoundItemsControl }

  TSpTBXCompoundItemsControl = class(TCustomControl, ITBItems)
  private
    procedure DockRequestDock(Sender: TObject; Bar: TTBCustomDockableWindow; var Accept: Boolean);
    function GetItems: TTBCustomItem;  // For ITBItems interface
    function GetRootItems: TTBRootItem;
    function GetView: TTBToolbarView;
    function GetImages: TCustomImageList;
    procedure SetImages(const Value: TCustomImageList);
  protected
    FDock: TSpTBXDock;
    FToolbar: TSpTBXToolbar;
    function GetDockClass: TSpTBXDockClass; virtual;
    function GetToolbarClass: TSpTBXToolbarClass; virtual;
    procedure GetChildren(Proc: TGetChildProc; Root: TComponent); override;  // For ITBItems interface
    procedure CreateParams(var Params: TCreateParams); override;
    procedure Loaded; override;
    property Images: TCustomImageList read GetImages write SetImages;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property View: TTBToolbarView read GetView;
  published
    property Items: TTBRootItem read GetRootItems;
  end;

  { TSpTBXStatusBar }

  TSpTBXStatusToolbar = class(TSpTBXToolbar)
  private
    FThemeType: TSpTBXThemeType;
    procedure WMNCHitTest(var Message: TWMNCHitTest); message WM_NCHITTEST;
  public
    function GetGripRect: TRect;
    function NeedsSeparatorRepaint: Boolean;
    constructor Create(AOwner: TComponent); override;
  end;

  TSpTBXCustomStatusBar = class(TSpTBXCompoundItemsControl)
  private
    FThemeType: TSpTBXThemeType;
    FOnDrawBackground: TSpTBXDrawEvent;
    procedure DockDrawBackground(Sender: TObject; ACanvas: TCanvas; ARect: TRect;
      const PaintStage: TSpTBXPaintStage; var PaintDefault: Boolean);
    procedure DockResize(Sender: TObject);
    procedure SetThemeType(const Value: TSpTBXThemeType);
    procedure TBMThemeChange(var Message: TMessage); message TBM_THEMECHANGE;
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
  protected
    procedure DoDrawBackground(ACanvas: TCanvas; ARect: TRect;
      const PaintStage: TSpTBXPaintStage; var PaintDefault: Boolean); virtual;
    procedure DrawSeparators(ACanvas: TCanvas; ARect, AGripRect: TRect); virtual;
    function GetToolbarClass: TSpTBXToolbarClass; override;
    function ValidateToolbar: TSpTBXStatusToolbar;
    property Align default alBottom;
    property ThemeType: TSpTBXThemeType read FThemeType write SetThemeType default thtTBX;
    property OnDrawBackground: TSpTBXDrawEvent read FOnDrawBackground write FOnDrawBackground;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;

  TSpTBXStatusBar = class(TSpTBXCustomStatusBar)
  published
    property Align;
    property Anchors;
    property BiDiMode;
    property Constraints;
    property DockSite;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property ParentColor;
    property ParentBiDiMode;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Visible;
    property OnCanResize;
    property OnContextPopup;
    property OnDockDrop;
    property OnDockOver;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnGetSiteInfo;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDock;
    property OnStartDrag;
    property OnUnDock;
    // TSpTBXCustomStatusBar properties
    property Images;
    property ThemeType;
    property OnDrawBackground;
  end;

  { TBitmapHint }

  TBitmapHint = class(THintWindow)
  private
    FHintBitmap: TBitmap;
    FActivating: Boolean;
    procedure CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;
  protected
    procedure Paint; override;
  public
    property Activating: Boolean read FActivating;
    procedure ActivateHint(Rect: TRect; const AHint: string); override;
    procedure ActivateHintData(Rect: TRect; const AHint: string; AData: Pointer); override;
  end;

{ Item helpers }
function SpXPThemesAvailable: Boolean;
procedure SpFillItemInfo(AEnabled, APushed, AMouseInControl, AChecked: Boolean; var ItemInfo: TTBXItemInfo); overload;
procedure SpFillItemInfo(ItemViewer: TTBItemViewer; var ItemInfo: TTBXItemInfo); overload;

function SpGetNextItem(AItemViewer: TTBItemViewer; GoForward, Visible, Showed, SameLevel: Boolean; Inmediate: Boolean = False; ACallBack: TSpTBXGetNextCallBack = nil): TTBItemViewer;
function SpGetFirstRightAlignSpacer(View: TTBView): TSpTBXItemViewer;
function SpGetRightAlignedItems(View: TTBView; RightAlignedList: TList; out VisibleTotalWidth, RightAlignedTotalWidth: Integer): TSpTBXItemViewer;
function SpFindItemViewer(View: TTBView; Item: TTBCustomItem): TTBItemViewer;
function SpFindControlItem(Item: TTBCustomItem; Ctl: TControl): TTBControlItem;
function SpFindControl(Parent: TWinControl; Child: TControl): Integer;

{ Painting helpers }
procedure SpDrawFocusRect(Canvas: TCanvas; const R: TRect);
procedure SpHighlightBitmap(Canvas: TCanvas; X, Y: Integer; Bitmap: TBitmap; TransparentColor, HighlightColor: TColor; Amount: Byte);
function SpCalcXPText(ACanvas: TCanvas; ARect: TRect; Caption: WideString; CaptionAlignment: TAlignment; Flags: Cardinal; GlyphWidth, GlyphHeight: Integer; Layout: TSpGlyphLayout; Enabled, PushedCaption: Boolean; out ACaptionRect, AGlyphRect: TRect): Integer;
function SpDrawXPText(ACanvas: TCanvas; Caption: WideString; var ARect: TRect; Flags: Cardinal; CaptionGlow: Boolean = False; CaptionGlowColor: TColor = clYellow): Integer; overload;
function SpDrawXPText(ACanvas: TCanvas; ARect: TRect; Caption: WideString; CaptionGlow: Boolean; CaptionGlowColor: TColor; CaptionAlignment: TAlignment; Flags: Cardinal; Glyph: TBitmap; Layout: TSpGlyphLayout; Enabled, PushedCaption: Boolean; out ACaptionRect, AGlyphRect: TRect; CalcRectOnly: Boolean = False): Integer; overload;
function SpDrawXPText(ACanvas: TCanvas; ARect: TRect; Caption: WideString; CaptionGlow: Boolean; CaptionGlowColor: TColor; CaptionAlignment: TAlignment; Flags: Cardinal; IL: TCustomImageList; ImageIndex: Integer; Layout: TSpGlyphLayout; Enabled, PushedCaption: Boolean; out ACaptionRect, AGlyphRect: TRect; CalcRectOnly: Boolean = False): Integer; overload;
procedure SpDrawXPButton(ACanvas: TCanvas; ARect: TRect; Enabled, Pushed, MouseOver, Focused, Defaulted: Boolean; UseWindowTheme: Boolean = True); overload;
procedure SpDrawXPButton(ACanvas: TCanvas; ARect: TRect; const ItemInfo: TTBXItemInfo); overload;
procedure SpDrawXPStatusBar(ACanvas: TCanvas; ARect, AGripRect: TRect; ThemeType: TSpTBXThemeType);

var
  SP_TAB_THEME: THandle;
  SP_PROGRESS_THEME: THandle;
  SP_TRACKBAR_THEME: THandle;
  SpStockHintBitmap: TBitmap;
  SpStockHandCursor: HCURSOR;

implementation

uses
  TBXUxThemes, TBXUtils;

type
  TTBRootItemAccess = class(TTBRootItem);
  TTBXItemAccess = class(TTBXCustomItem);
  TTBXItemViewerAccess = class(TTBXItemViewer);
  TTBViewAccess = class(TTBView);

  // TBXThemes.TBXThemeManager loads/unloads the XP theme library and data parts.
  // The problem is that it doesn't load the Tab data parts, because TBX doesn't
  // need it. For that we need to load Tab parts (declared in TBXUxThemes unit),
  // and load/unload the theme library OUTSIDE of the TBXThemeManager, because
  // some of its properties may unload the theme library for example
  // EnableVisualStyles.
  // Fortunately we can use the TBXThemeManager notifications to load/unload
  // the Tab parts.

  { TSpTabThemeNexus }

  TSpThemeNexus = class
  private
    FUXThemeDllHandle: THandle;
    procedure TBXSysCommand(var Message: TMessage); message TBX_SYSCOMMAND;
    procedure OpenData;
    procedure CloseData;
  public
    constructor Create; virtual;
    destructor Destroy; override;
  end;

var
  SpThemeNexus: TSpThemeNexus;
  SpOpenThemeData: function(hwnd: HWND; pszClassList: LPCWSTR): HTHEME; stdcall;
  SpCloseThemeData: function(hTheme: HTHEME): HRESULT; stdcall;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTabThemeNexus }

constructor TSpThemeNexus.Create;
begin
  inherited;
  SP_TAB_THEME := 0;
  SP_PROGRESS_THEME := 0;
  FUXThemeDllHandle := LoadLibrary('uxtheme.dll');
  if FUXThemeDllHandle > 0 then begin
    SpOpenThemeData := GetProcAddress(FUXThemeDllHandle, 'OpenThemeData');
    SpCloseThemeData := GetProcAddress(FUXThemeDllHandle, 'CloseThemeData');
  end;
  OpenData;
  AddTBXSysChangeNotification(Self);
end;

destructor TSpThemeNexus.Destroy;
begin
  RemoveTBXSysChangeNotification(Self);
  CloseData;
  if FUXThemeDllHandle <> 0 then begin
    FreeLibrary(FUXThemeDllHandle);
    FUXThemeDllHandle := 0;
    SpOpenThemeData := nil;
    SpCloseThemeData := nil;
  end;
  inherited;
end;

procedure TSpThemeNexus.OpenData;
begin
  if FUXThemeDllHandle > 0 then begin
    SP_TAB_THEME := SpOpenThemeData(Application.Handle, 'TAB');
    SP_PROGRESS_THEME := SpOpenThemeData(Application.Handle, 'PROGRESS');
    SP_TRACKBAR_THEME := SpOpenThemeData(Application.Handle, 'TRACKBAR');
  end;
end;

procedure TSpThemeNexus.CloseData;

  procedure Close(var ATheme: THandle);
  begin
    if ATheme <> 0 then begin
      SpCloseThemeData(ATheme);
      ATheme := 0;
    end;
  end;

begin
  if FUXThemeDllHandle > 0 then begin
    Close(SP_TAB_THEME);
    Close(SP_PROGRESS_THEME);
    Close(SP_TRACKBAR_THEME);
  end;
end;

procedure TSpThemeNexus.TBXSysCommand(var Message: TMessage);
begin
  if Message.Msg = TBX_SYSCOMMAND then
    Case Message.WParam of
      TSC_BEFOREVIEWCHANGE:
        CloseData;
      TSC_AFTERVIEWCHANGE:
        OpenData;
    end;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ Helpers }

function SpXPThemesAvailable: Boolean;
begin
  Result := USE_THEMES and CanUseXPThemes;
end;

procedure SpFillItemInfo(AEnabled, APushed, AMouseInControl, AChecked: Boolean; var ItemInfo: TTBXItemInfo);
begin
  FillChar(ItemInfo, SizeOf(ItemInfo), 0);
  with ItemInfo do begin
    ItemInfo.ViewType := TVT_NORMALTOOLBAR or TVT_EMBEDDED;
    ItemInfo.ItemOptions := IO_TOOLBARSTYLE or IO_APPACTIVE;
    ItemInfo.Enabled := AEnabled;
    ItemInfo.Pushed := APushed and AMouseInControl;
    if AMouseInControl then
      ItemInfo.HoverKind := hkMouseHover
    else
      ItemInfo.HoverKind := hkNone;
    ItemInfo.Selected := AChecked;
    IsVertical := false;
    ItemInfo.ComboPart := cpNone;
    IsPopupParent := False;

    ItemInfo.ImageShown := false;
    ItemInfo.ImageWidth := 0;
    ItemInfo.ImageHeight := 0;
  end;
end;

procedure SpFillItemInfo(ItemViewer: TTBItemViewer; var ItemInfo: TTBXItemInfo);
const
  CToolbarStyle: array [Boolean] of Integer = (0, IO_TOOLBARSTYLE);
  CCombo: array [Boolean] of Integer = (0, IO_COMBO);
  CSubmenuItem: array [Boolean] of Integer = (0, IO_SUBMENUITEM);
  CDesigning: array [Boolean] of Integer = (0, IO_DESIGNING);
  CAppActive: array [Boolean] of Integer = (0, IO_APPACTIVE);
var
  Item: TTBXCustomItem;
  View: TTBViewAccess;

  ClientAreaRect: TRect;
  IsHoverItem, IsOpen, IsPushed: Boolean;
  ToolbarStyle: Boolean;
  HasArrow: Boolean;
  IsSplit: Boolean;
  ImageIsShown: Boolean;
  ImgSize: TSize;
  IsComboPushed: Boolean;
begin
  Item := TTBXCustomItem(ItemViewer.Item);
  View := TTBViewAccess(ItemViewer.View);

  ClientAreaRect := ItemViewer.BoundsRect;
  OffsetRect(ClientAreaRect, -ClientAreaRect.Left, -ClientAreaRect.Top);

  IsOpen := ItemViewer = View.OpenViewer;
  IsHoverItem := ItemViewer = View.Selected;
  IsPushed := IsHoverItem and (IsOpen or (View.MouseOverSelected and View.Capture));

  ToolbarStyle := ItemViewer.IsToolbarStyle;
  IsSplit := tbisCombo in TTBXItemAccess(Item).ItemStyle;
  IsComboPushed := IsSplit and IsPushed and not View.Capture;
  if IsComboPushed then IsPushed := False;

  if TTBXItemViewerAccess(ItemViewer).GetImageShown then
  begin
    ImgSize := TTBXItemViewerAccess(ItemViewer).GetImageSize;
    with ImgSize do if (CX <= 0) or (CY <= 0) then
    begin
      CX := 0;
      CY := 0;
      ImageIsShown := False;
    end
    else ImageIsShown := True;
  end
  else
  begin
    ImgSize.CX := 0;
    ImgSize.CY := 0;
    ImageIsShown := False;
  end;
  IsSplit := tbisCombo in TTBXItemAccess(Item).ItemStyle;

  FillChar(ItemInfo, SizeOf(ItemInfo), 0);
  ItemInfo.ViewType := GetViewType(View);
  ItemInfo.ItemOptions := CToolbarStyle[ToolbarStyle] or CCombo[IsSplit] or
    CDesigning[csDesigning in Item.ComponentState] or CSubmenuItem[tbisSubmenu in TTBXItemAccess(Item).ItemStyle] or
    CAppActive[Application.Active];
  ItemInfo.Enabled := Item.Enabled or View.Customizing;
  ItemInfo.Pushed := IsPushed;
  ItemInfo.Selected := Item.Checked;
  ItemInfo.ImageShown := ImageIsShown;
  ItemInfo.ImageWidth := ImgSize.CX;
  ItemInfo.ImageHeight := ImgSize.CY;
  if IsHoverItem then
  begin
    if not ItemInfo.Enabled and not View.MouseOverSelected then ItemInfo.HoverKind := hkKeyboardHover
    else if ItemInfo.Enabled then ItemInfo.HoverKind := hkMouseHover;
  end
  else ItemInfo.HoverKind := hkNone;
  ItemInfo.IsPopupParent := ToolbarStyle and
    (((vsModal in View.State) and Assigned(View.OpenViewer)) or (tbisSubmenu in TTBXItemAccess(Item).ItemStyle)) and
    ((IsSplit and IsComboPushed) or (not IsSplit and IsPushed));
  ItemInfo.IsVertical := (View.Orientation = tbvoVertical) and not IsSplit;
//  if not ToolbarStyle then ItemInfo.PopupMargin := GetPopupMargin(Self);
  ItemInfo.PopupMargin := GetPopupMargin(ItemViewer);

  HasArrow := (tbisSubmenu in TTBXItemAccess(Item).ItemStyle) and
    ((tbisCombo in TTBXItemAccess(Item).ItemStyle) or (tboDropdownArrow in Item.EffectiveOptions));

  if ToolbarStyle then
  begin
    if HasArrow then
      ItemInfo.ComboPart := cpCombo;
    if IsSplit then
      ItemInfo.ComboPart := cpSplitLeft;
  end;
end;

function SpGetNextItem(AItemViewer: TTBItemViewer; GoForward, Visible, Showed,
  SameLevel: Boolean; Inmediate: Boolean = False;
  ACallBack: TSpTBXGetNextCallBack = nil): TTBItemViewer;

  function ProcessItemViewer(IV: TTBItemViewer; var EndProcess: Boolean): Boolean;
  begin
    Result := False;
    if Visible then
      Result := IV.Item.Visible = Visible;
    if SameLevel then
      Result := Result and (IV.OffEdge = AItemViewer.OffEdge) and (IV.BoundsRect.Top = AItemViewer.BoundsRect.Top);
    if Showed then
      Result := Result and IV.Show;
    if Assigned(ACallBack) then
      Result := ACallback(IV, Result, EndProcess);
  end;

var
  V: TTBView;
  I: Integer;
  EndProcess: Boolean;
begin
  Result := nil;
  EndProcess := false;
  V := AItemViewer.View;
  V.ValidatePositions;
  I := V.IndexOf(AItemViewer);

  while not Assigned(Result) do begin
    if GoForward then
      Inc(I)
    else
      Dec(I);
    if (I > V.ViewerCount - 1) or (I < 0) then
      Break
    else
      if ProcessItemViewer(V.Viewers[I], EndProcess) then
        Result := V.Viewers[I];
    if (Inmediate and not Assigned(Result)) or EndProcess then
      Break;
  end;
end;

function SpGetFirstRightAlignSpacer(View: TTBView): TSpTBXItemViewer;
var
  I: integer;
  IV: TTBItemViewer;
begin
  Result := nil;
  for I := 0 to View.ViewerCount - 1 do begin
    IV := View.Viewers[I];
    if IV.Item.Visible and (IV.Item is TSpTBXRightAlignSpacerItem) then
      Result := IV as TSpTBXItemViewer;
  end;
end;

function SpGetRightAlignedItems(View: TTBView; RightAlignedList: TList;
  out VisibleTotalWidth, RightAlignedTotalWidth: Integer): TSpTBXItemViewer;
var
  I: integer;
  IV: TTBItemViewer;
begin
  Result := nil;

  if Assigned(RightAlignedList) then
    RightAlignedList.Clear;
  VisibleTotalWidth := 0;
  RightAlignedTotalWidth := 0;

  for I := 0 to View.ViewerCount - 1 do begin
    IV := View.Viewers[I];
    if IV.Item.Visible then
      VisibleTotalWidth := VisibleTotalWidth + IV.BoundsRect.Right - IV.BoundsRect.Left;
    if not Assigned(Result) and (IV.Item is TSpTBXRightAlignSpacerItem) then
      Result := IV as TSpTBXItemViewer;
    if Assigned(Result) then begin
      if Assigned(RightAlignedList) then
        RightAlignedList.Add(IV);
      RightAlignedTotalWidth := RightAlignedTotalWidth + IV.BoundsRect.Right - IV.BoundsRect.Left;
    end;
  end;
end;

function SpFindItemViewer(View: TTBView; Item: TTBCustomItem): TTBItemViewer;
var
  I: Integer;
begin
  Result := nil;
  for I := 0 to View.ViewerCount - 1 do
    if View.Viewers[I].Item = Item then begin
      Result := View.Viewers[I];
      Exit;
    end;
end;

function SpFindControlItem(Item: TTBCustomItem; Ctl: TControl): TTBControlItem;
var
  I: Integer;
begin
  Result := nil;
  for I := 0 to Item.Count - 1 do begin
    if (Item[I] is TTBControlItem) and (TTBControlItem(Item[I]).Control = Ctl) then begin
      Result := TTBControlItem(Item[I]);
      Break;
    end;
  end;
end;

function SpFindControl(Parent: TWinControl; Child: TControl): Integer;
var
  I: Integer;
begin
  Result := -1;
  for I := 0 to Parent.ControlCount - 1 do
    if Parent.Controls[I] = Child then begin
      Result := I;
      Break;
    end;
end;

procedure SpDrawFocusRect(Canvas: TCanvas; const R: TRect);
var
  DC: HDC;
  C1, C2: TColor;
begin
  if not IsRectEmpty(R) then begin
    DC := Canvas.Handle;
    C1 := SetTextColor(DC, clBlack);
    C2 := SetBkColor(DC, clWhite);
    Canvas.DrawFocusRect(R);
    SetTextColor(DC, C1);
    SetBkColor(DC, C2);
  end;
end;

procedure SpHighlightBitmap(Canvas: TCanvas; X, Y: Integer; Bitmap: TBitmap;
  TransparentColor, HighlightColor: TColor; Amount: Byte);
var
  IL: TImageList;
  R: TRect;
begin
  IL := TImageList.CreateSize(Bitmap.Width, Bitmap.Height);
  try
    IL.AddMasked(Bitmap, TransparentColor);
    R := Bounds(X, Y, IL.Width, IL.Height);
    TBXUtils.HighlightTBXIcon(Canvas, R, IL, 0, HighlightColor, Amount);
  finally
    IL.Free;
  end;
end;

function SpCalcXPText(ACanvas: TCanvas; ARect: TRect; Caption: WideString;
  CaptionAlignment: TAlignment; Flags: Cardinal; GlyphWidth, GlyphHeight: Integer;
  Layout: TSpGlyphLayout; Enabled, PushedCaption: Boolean; out ACaptionRect, AGlyphRect: TRect): Integer;
var
  W: integer;
  R: TRect;
  CaptionSz, TextOffset: TPoint;
const
  Space = 4;
begin
  ACaptionRect := Rect(0, 0, 0, 0);
  AGlyphRect := Rect(0, 0, 0, 0);

  // Get the caption size
  if (Flags and DT_WORDBREAK) <> 0 then
     R := Rect(0, 0, Arect.Right - ARect.Left, 1)
  else
     R := Rect(0, 0, 1, 1);

  Result := SpDrawXPText(ACanvas, Caption, R, Flags or DT_CALCRECT);
  CaptionSz := Point(R.Right, R.Bottom);

  // ACaptionRect
  R.Top := (ARect.Top + ARect.Bottom - CaptionSz.Y) div 2; // Vertically centered
  R.Bottom := R.Top + CaptionSz.Y;
  Case CaptionAlignment of
    taCenter:
      R.Left := (ARect.Right + ARect.Left - CaptionSz.X) div 2; // Horizontally centered
    taLeftJustify:
      R.Left := ARect.Left;
    taRightJustify:
      R.Left := ARect.Right - CaptionSz.X;
  end;
  R.Right := R.Left + CaptionSz.X;
  if PushedCaption then
    OffsetRect(R, 1, 1);
  ACaptionRect := R;

  // AGlyphRect
  if (GlyphWidth > 0) and (GlyphHeight > 0) then begin
    TextOffset := Point(0, 0);
    R := ARect;
    W := GlyphWidth + Space + CaptionSz.X; // Total width = Icon + Space + Text
    Case CaptionAlignment of
      taCenter:
        begin
          AGlyphRect.Left := (R.Right + R.Left - W) div 2;
          TextOffset.X := (GlyphWidth + Space) div 2;
        end;
      taLeftJustify:
        begin
          AGlyphRect.Left := R.Left;
          TextOffset.X := Space + GlyphWidth;
        end;
      taRightJustify:
        begin
          AGlyphRect.Left := R.Right - GlyphWidth;
          TextOffset.X := - Space - GlyphWidth;
        end;
    end;
    AGlyphRect.Top := (R.Top + R.Bottom + 1 - GlyphHeight) div 2;
    AGlyphRect.Right := AGlyphRect.Left + GlyphWidth;
    AGlyphRect.Bottom := AGlyphRect.Top + GlyphHeight;
    if PushedCaption then
      OffsetRect(AGlyphRect, 1, 1);

    // Move the text according to the icon position
    OffsetRect(ACaptionRect, TextOffset.X, TextOffset.Y);
  end;
end;

function SpDrawXPText(ACanvas: TCanvas; Caption: WideString; var ARect: TRect;
  Flags: Cardinal; CaptionGlow: Boolean = False; CaptionGlowColor: TColor = clYellow): Integer; overload;

  function InternalDraw(var R: TRect): Integer;
  var
    S: string;
  begin
    if Win32Platform = VER_PLATFORM_WIN32_WINDOWS then begin
      S := Caption;
      Result := Windows.DrawText(ACanvas.Handle, PChar(S), -1, R, Flags);
    end
    else
      Result := Windows.DrawTextW(ACanvas.Handle, PWideChar(Caption), -1, R, Flags);
  end;

var
  BS: TBrushStyle;
  GlowR: TRect;
  C, FC: TColor;
begin
  BS := ACanvas.Brush.Style;
  C := ACanvas.Brush.Color;
  try
    ACanvas.Brush.Style := bsClear;

    {$IFDEF JR_D5}
    // UTF8Decode was introduced in Delphi 6
    {$ELSE}
    if Pos('utf8=', Caption) = 1 then begin
      Delete(Caption, 1, 5);
      Caption := UTF8Decode(Caption);
    end;
    {$ENDIF}

    if CaptionGlow and (Flags and DT_CALCRECT = 0) then begin
      FC := ACanvas.Font.Color;
      try
        ACanvas.Font.Color := CaptionGlowColor;
        GlowR := ARect; OffsetRect(GlowR, -1, -1);
        InternalDraw(GlowR);
        GlowR := ARect; OffsetRect(GlowR, 1, -1);
        InternalDraw(GlowR);
        GlowR := ARect; OffsetRect(GlowR, 1, 1);
        InternalDraw(GlowR);
        GlowR := ARect; OffsetRect(GlowR, -1, 1);
        InternalDraw(GlowR);
      finally
        ACanvas.Font.Color := FC;
        Result := InternalDraw(ARect);
      end;
    end
    else
      Result := InternalDraw(ARect);

  finally
    ACanvas.Brush.Style := BS;
    ACanvas.Brush.Color := C;
  end;
end;

function SpDrawXPText(ACanvas: TCanvas; ARect: TRect; Caption: WideString;
  CaptionGlow: Boolean; CaptionGlowColor: TColor; CaptionAlignment: TAlignment;
  Flags: Cardinal; Glyph: TBitmap; Layout: TSpGlyphLayout; Enabled, PushedCaption: Boolean;
  out ACaptionRect, AGlyphRect: TRect; CalcRectOnly: Boolean = False): Integer; overload;
begin
  Result := SpCalcXPText(ACanvas, ARect, Caption, CaptionAlignment, Flags, Glyph.Width,
    Glyph.Height, Layout, Enabled, PushedCaption, ACaptionRect, AGlyphRect);
  // Draw the whole thing
  if not CalcRectOnly then begin
    SpDrawXPText(ACanvas, Caption, ACaptionRect, Flags and not DT_CALCRECT, CaptionGlow, CaptionGlowColor);
    if Assigned(Glyph) then
      ACanvas.Draw(AGlyphRect.Left, AGlyphRect.Top, Glyph);
  end;
end;

function SpDrawXPText(ACanvas: TCanvas; ARect: TRect; Caption: WideString;
  CaptionGlow: Boolean; CaptionGlowColor: TColor; CaptionAlignment: TAlignment;
  Flags: Cardinal; IL: TCustomImageList; ImageIndex: Integer; Layout: TSpGlyphLayout;
  Enabled, PushedCaption: Boolean; out ACaptionRect, AGlyphRect: TRect;
  CalcRectOnly: Boolean = False): Integer; overload;
var
  W, H: Integer;
begin
  if Assigned(IL) and (ImageIndex > -1) and (ImageIndex < IL.Count) then begin
    W := IL.Width;
    H := IL.Height;
  end
  else begin
    W := 0;
    H := 0;
  end;

  Result := SpCalcXPText(ACanvas, ARect, Caption, CaptionAlignment, Flags, W,
    H, Layout, Enabled, PushedCaption, ACaptionRect, AGlyphRect);
  // Draw the whole thing
  if not CalcRectOnly then begin
    SpDrawXPText(ACanvas, Caption, ACaptionRect, Flags and not DT_CALCRECT, CaptionGlow, CaptionGlowColor);
    if Assigned(IL) and (ImageIndex > -1) and (ImageIndex < IL.Count) then
      if Enabled then IL.Draw(ACanvas, AGlyphRect.Left, AGlyphRect.Top, ImageIndex)
      else DrawTBXImage(ACanvas, AGlyphRect, IL, ImageIndex, ISF_DISABLED);
  end;
end;

procedure SpDrawXPButton(ACanvas: TCanvas; ARect: TRect; Enabled, Pushed, MouseOver, Focused, Defaulted: Boolean;
  UseWindowTheme: Boolean = True);
var
  Flags: Cardinal;
  C: TColor;
begin
  if UseWindowTheme and SpXPThemesAvailable then begin
    if not Enabled then
      Flags := PBS_DISABLED
    else
      if Pushed then
        Flags := PBS_PRESSED
      else
        if MouseOver then
          Flags := PBS_HOT
        else
          if Defaulted or Focused then
            Flags := PBS_DEFAULTED
          else
            Flags := PBS_NORMAL;
    DrawThemeBackground(BUTTON_THEME, ACanvas.Handle, BP_PUSHBUTTON, Flags, ARect, nil);
  end
  else begin
    ACanvas.Brush.Color := clBtnFace;
    ACanvas.FillRect(ARect);

    if Defaulted or Focused then begin
      C := ACanvas.Brush.Color;
      ACanvas.Brush.Color := clWindowFrame;
      ACanvas.FrameRect(ARect);
      ACanvas.Brush.Color := C;
      InflateRect(ARect, -1, -1); // reduce the Rect
    end;
    if Pushed then begin
      C := ACanvas.Brush.Color;
      ACanvas.Brush.Color := clBtnShadow;
      ACanvas.FrameRect(ARect);
      ACanvas.Brush.Color := C;
    end
    else
      DrawFrameControl(ACanvas.Handle, ARect, DFC_BUTTON, DFCS_BUTTONPUSH);
  end;

  if Focused then begin
    InflateRect(ARect, -3, -3);
    SpDrawFocusRect(ACanvas, ARect);
  end;
end;

procedure SpDrawXPButton(ACanvas: TCanvas; ARect: TRect; const ItemInfo: TTBXItemInfo);
begin
  SpDrawXPButton(ACanvas, ARect, ItemInfo.Enabled, ItemInfo.Pushed, ItemInfo.HoverKind <> hkNone, false, false);
end;

procedure SpDrawXPStatusBar(ACanvas: TCanvas; ARect, AGripRect: TRect; ThemeType: TSpTBXThemeType);
begin
  if (ThemeType = thtWindows) and not SpXPThemesAvailable then
    ThemeType := thtNone;

  Case ThemeType of
    thtNone:
      begin
        TBXUtils.FillRectEx(ACanvas.Handle, ARect, clBtnFace);
        TBXUtils.Frame3D(ACanvas.Handle, ARect, clBtnShadow, clWindow, False);
        if not IsRectEmpty(AGripRect) then begin
          inc(AGripRect.Top, 4);
          dec(AGripRect.Right, 1);
          dec(AGripRect.Bottom, 1);
          DrawFrameControl(ACanvas.Handle, AGripRect, DFC_SCROLL, DFCS_SCROLLSIZEGRIP);
        end;
      end;
    thtWindows:
      begin
        DrawThemeBackground(STATUSBAR_THEME, ACanvas.Handle, 0, 0, ARect, nil);
        if not IsRectEmpty(AGripRect) then
          DrawThemeBackground(STATUSBAR_THEME, ACanvas.Handle, SP_GRIPPER, 0, AGripRect, nil)
      end;
    thtTBX:
      begin
        // [TBXTheme-Change]
        // Office themes have rectangle panels
        if Pos('Office', TBXCurrentTheme) > 0 then begin
          TBXUtils.FillRectEx(ACanvas.Handle, ARect, clBtnFace);
          TBXUtils.FrameRectEx(ACanvas.Handle, ARect, clWindow, False);
        end
        else
          CurrentTheme.PaintStatusBar(ACanvas, ARect, SBP_BODY);

        if not IsRectEmpty(AGripRect) then
          CurrentTheme.PaintStatusBar(ACanvas, AGripRect, SBP_GRIPPER);
      end;
  end;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXCustomItemActionLink }

procedure TSpTBXCustomItemActionLink.AssignClient(AClient: TObject);
begin
  FClient := AClient as TSpTBXCustomItem;
end;

procedure TSpTBXCustomItemActionLink.SetCaption(const Value: String);
begin
  if IsCaptionLinked then begin
    (FClient as TSpTBXCustomItem).Caption := Value;
  end;
end;

procedure TSpTBXCustomItemActionLink.SetHint(const Value: String);
begin
  if IsHintLinked then begin
    (FClient as TSpTBXCustomItem).Hint := Value;
  end;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXCustomItem }

constructor TSpTBXCustomItem.Create(AOwner: TComponent);
begin
  inherited;
  FCaption := '';
  FAlignment := taCenter;
  FCustomWidth := -1;
  FCustomHeight := -1;

  DisplayMode := nbdmImageAndText;
  Stretch := True;
end;

procedure TSpTBXCustomItem.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) and (AComponent = Control) then Control := nil;
end;

procedure TSpTBXCustomItem.DefineProperties(Filer: TFiler);
begin
  inherited;
  // Don't let the VCL store the WideString, it doesn't write it correctly.
  // Use DefineProperties instead, with a new name for the property
  Filer.DefineProperty('CaptionW', ReadCaptionW, WriteCaptionW, FCaption <> '');
  Filer.DefineProperty('HintW', ReadHintW, WriteHintW, FHint <> '');
end;

procedure TSpTBXCustomItem.ReadCaptionW(Reader: TReader);
begin
  case Reader.NextValue of
    vaLString, vaString:
      SetCaption(Reader.ReadString);
  else
    SetCaption(Reader.ReadWideString);
  end;
end;

procedure TSpTBXCustomItem.WriteCaptionW(Writer: TWriter);
begin
  Writer.WriteWideString(FCaption);
end;

procedure TSpTBXCustomItem.ReadHintW(Reader: TReader);
begin
  case Reader.NextValue of
    vaLString, vaString:
      SetHint(Reader.ReadString);
  else
    SetHint(Reader.ReadWideString);
  end;
end;

procedure TSpTBXCustomItem.WriteHintW(Writer: TWriter);
begin
  Writer.WriteWideString(FHint);
end;

procedure TSpTBXCustomItem.ActionChange(Sender: TObject; CheckDefaults: Boolean);
begin
  if Action is TCustomAction then
    with TCustomAction(Sender) do begin
      if not CheckDefaults or (Self.Caption = '') then
        Self.Caption := Caption;
      if not CheckDefaults or (Self.Hint = '') then
        Self.Hint := Hint;
    end;
  inherited;
end;

function TSpTBXCustomItem.DialogChar(CharCode: Word): Boolean;
begin
  Result := False;
end;

procedure TSpTBXCustomItem.DoDrawHint(AHintBitmap: TBitmap;
  var AHint: Widestring; var PaintDefault: Boolean);
begin
  if Assigned(FOnDrawHint) then FOnDrawHint(Self, AHintBitmap, AHint, PaintDefault);
end;

procedure TSpTBXCustomItem.DoDrawButton(ACanvas: TCanvas; ARect: TRect;
  ItemInfo: TTBXItemInfo; const PaintStage: TSpTBXPaintStage; var PaintDefault: Boolean);
begin
  if Assigned(FOnDrawItem) then FOnDrawItem(Self, ACanvas, ARect, ItemInfo, PaintStage, PaintDefault);
end;

procedure TSpTBXCustomItem.DoDrawCaption(ACanvas: TCanvas; const ClientAreaRect: TRect;
  var ACaption: WideString; var CaptionRect: TRect; IsTextRotated: Boolean;
  const PaintStage: TSpTBXPaintStage; var PaintDefault: Boolean);
begin
  if Assigned(FOnDrawCaption) then FOnDrawCaption(Self, ACanvas, ClientAreaRect,
    ACaption, CaptionRect, IsTextRotated, PaintStage, PaintDefault);
end;

procedure TSpTBXCustomItem.DoDrawImage(ACanvas: TCanvas;
  const ItemInfo: TTBXItemInfo; const PaintStage: TSpTBXPaintStage;
  AImageList: TCustomImageList; var AImageIndex: integer;
  var ARect: TRect; var PaintDefault: Boolean);
begin
  if Assigned(FOnDrawImage) then
    FOnDrawImage(Self, ACanvas, ItemInfo, PaintStage, AImageList, AImageIndex,
      ARect, PaintDefault);
end;

function TSpTBXCustomItem.GetActionLinkClass: TTBCustomItemActionLinkClass;
begin
  Result := TSpTBXCustomItemActionLink;
end;

function TSpTBXCustomItem.GetItemViewerClass(AView: TTBView): TTBItemViewerClass;
begin
  Result := TSpTBXItemViewer;
end;

procedure TSpTBXCustomItem.Click;
begin
  if Assigned(FControl) then ToggleControl;
  inherited;
end;

procedure TSpTBXCustomItem.InitiateAction;
begin
  inherited;
  UpdateProps;
end;

procedure TSpTBXCustomItem.SetCaption(const Value: WideString);
begin
  if FCaption <> Value then begin
    FCaption := Value;
    inherited Caption := Value;
  end;
end;

procedure TSpTBXCustomItem.SetHint(const Value: WideString);
begin
  if FHint <> Value then begin
    FHint := Value;
    inherited Hint := Value;
  end;
end;

procedure TSpTBXCustomItem.SetAlignment(const Value: TAlignment);
begin
  if FAlignment <> Value then begin
    FAlignment := Value;
    Change(False);
  end;
end;

procedure TSpTBXCustomItem.SetAnchored(const Value: Boolean);
begin
  if FAnchored <> Value then begin
    FAnchored := Value;
  end;
end;

procedure TSpTBXCustomItem.SetControl(const Value: TControl);
begin
  if FControl <> Value then
  begin
    FControl := Value;
    if Assigned(Value) then
      Value.FreeNotification(Self);
    UpdateProps;
  end;
end;

procedure TSpTBXCustomItem.SetCustomWidth(Value: Integer);
begin
  if Value < -1 then Value := -1;
  if FCustomWidth <> Value then begin
    FCustomWidth := Value;
    Change(True);
  end;
end;

procedure TSpTBXCustomItem.SetCustomHeight(Value: Integer);
begin
  if Value < -1 then Value := -1;
  if FCustomHeight <> Value then begin
    FCustomHeight := Value;
    Change(True);
  end;
end;

procedure TSpTBXCustomItem.ToggleControl;
begin
  FControl.Visible := not FControl.Visible;
end;

procedure TSpTBXCustomItem.UpdateProps;
begin
  if Assigned(Control) then
    if (ComponentState * [csDesigning, csLoading, csDestroying] = []) then
      Checked := Control.Visible;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXItemViewer }

procedure TSpTBXItemViewer.DoDrawButton(ACanvas: TCanvas; ARect: TRect;
  ItemInfo: TTBXItemInfo; const PaintStage: TSpTBXPaintStage; var PaintDefault: Boolean);
begin
  Item.DoDrawButton(ACanvas, ARect, ItemInfo, PaintStage, PaintDefault);
end;

procedure TSpTBXItemViewer.DoDrawCaption(ACanvas: TCanvas; const ClientAreaRect: TRect;
  var ACaption: WideString; var CaptionRect: TRect; IsTextRotated: Boolean;
  const PaintStage: TSpTBXPaintStage; var PaintDefault: Boolean);
begin
  Item.DoDrawCaption(ACanvas, ClientAreaRect, ACaption, CaptionRect,
    IsTextRotated, PaintStage, PaintDefault);
  if PaintDefault and (PaintStage = pstPrePaint) then begin
    PaintDefault := False;
    SpDrawXPText(ACanvas, ACaption, CaptionRect, 0);
  end;
end;

procedure TSpTBXItemViewer.DoDrawImage(ACanvas: TCanvas;
  const ItemInfo: TTBXItemInfo; const PaintStage: TSpTBXPaintStage;
  AImageList: TCustomImageList; var AImageIndex: integer;
  var ARect: TRect; var PaintDefault: Boolean);
begin
  Item.DoDrawImage(ACanvas, ItemInfo, PaintStage, AImageList, AImageIndex, ARect, PaintDefault);
end;

procedure TSpTBXItemViewer.DrawItemImage(Canvas: TCanvas; ARect: TRect;
  ItemInfo: TTBXItemInfo);
var
  PaintDefault: Boolean;
  ImgList: TCustomImageList;
  ImgIndex: integer;
begin
  ImgList := GetImageList;
  ImgIndex := Item.ImageIndex;

  PaintDefault := True;
  DoDrawImage(Canvas, ItemInfo, pstPrePaint, ImgList, ImgIndex, ARect, PaintDefault);
  if PaintDefault then begin
    if (ImgList <> nil) and (ImgIndex >= 0) and (ImgIndex < ImgList.Count) then begin
      CurrentTheme.PaintImage(Canvas, ARect, ItemInfo, ImgList, ImgIndex);
      { TODO : Pass the offset rect or leave it to the app? }
      {
      if Item is TTBXCustomItem then begin
        R := CurrentTheme.GetImageOffset(Canvas, ItemInfo, ImgList);
      end;
      }
    end;
  end;
  PaintDefault := True;
  DoDrawImage(Canvas, ItemInfo, pstPostPaint, ImgList, ImgIndex, ARect, PaintDefault);
end;

procedure TSpTBXItemViewer.CalcSize(const Canvas: TCanvas; var AWidth,
  AHeight: Integer);
begin
  inherited CalcSize(Canvas, AWidth, AHeight);
  if Item.CustomWidth > -1 then
    AWidth := Item.CustomWidth;
  if Item.CustomHeight > -1 then
    AHeight := Item.CustomHeight;

  if IsToolbarStyle then
    if Item.Anchored then begin
      AWidth := AWidth + FAnchorDelta.X;
    end;

  if AWidth < Item.MinWidth then AWidth := Item.MinWidth;
  if AHeight < Item.MinHeight then AHeight := Item.MinHeight;
end;

function TSpTBXItemViewer.GetTextSize(Canvas: TCanvas;
  const Text: string; TextFlags: Cardinal; Rotated: Boolean;
  StateFlags: Integer): TSize;
var
  R: TRect;
begin
  Result := inherited GetTextSize(Canvas, Text, TextFlags, Rotated, StateFlags);
  if (Result.cx > 0) and (Result.cy > 0) then begin
    R := Rect(0, 1, 0, 0);
    SpDrawXPText(Canvas, Item.Caption, R, DT_CALCRECT);
    Result.cx := Result.cx + (R.Right - Result.cx);
    Result.cy := Result.cy + (R.Bottom - Result.cy);
  end;
end;

procedure TSpTBXItemViewer.Paint(const Canvas: TCanvas; const ClientAreaRect: TRect;
  IsHoverItem, IsPushed, UseDisabledShadow: Boolean);
var
  View: TTBViewAccess;
  ItemInfo: TTBXItemInfo;

  R: TRect;
  ComboRect: TRect;
  CaptionRect: TRect;
  ImageRect: TRect;
  C: TColor;

  ToolbarStyle: Boolean;
  HasArrow: Boolean;
  IsSplit: Boolean;
  ImageIsShown: Boolean;
  ImageOrCheckShown: Boolean;
  ImgAndArrowWidth: Integer;
  ImgSize: TSize;
  IsComboPushed: Boolean;
  IsCaptionShown: Boolean;
  IsTextRotated: Boolean;
  ItemLayout: TTBXItemLayout;
  PaintDefault: Boolean;
  S: string;
  WS: WideString;
  StateFlags: Integer;
  IsSpecialDropDown: Boolean;
  TextFlags: Cardinal;
  TextMetrics: TTextMetric;
  TextSize: TSize;
  TextAlignment: TAlignment;
  Margins: TTBXMargins;
begin
  View := TTBViewAccess(Self.View);
  SpFillItemInfo(Self, ItemInfo);

  ToolbarStyle := IsToolbarStyle;
  IsSplit := tbisCombo in TTBXItemAccess(Item).ItemStyle;
  IsComboPushed := IsSplit and IsPushed and not View.Capture;
  if IsComboPushed then IsPushed := False;

  ItemLayout := Item.Layout;
  if ItemLayout = tbxlAuto then
  begin
    if tboImageAboveCaption in Item.EffectiveOptions then ItemLayout := tbxlGlyphTop
    else if View.Orientation <> tbvoVertical then ItemLayout := tbxlGlyphLeft
    else ItemLayout := tbxlGlyphTop;
  end;

  HasArrow := (tbisSubmenu in TTBXItemAccess(Item).ItemStyle) and
    ((tbisCombo in TTBXItemAccess(Item).ItemStyle) or (tboDropdownArrow in Item.EffectiveOptions));

  if GetImageShown then
  begin
    ImgSize := GetImageSize;
    with ImgSize do if (CX <= 0) or (CY <= 0) then
    begin
      CX := 0;
      CY := 0;
      ImageIsShown := False;
    end
    else ImageIsShown := True;
  end
  else
  begin
    ImgSize.CX := 0;
    ImgSize.CY := 0;
    ImageIsShown := False;
  end;
  ImageOrCheckShown := ImageIsShown or (not ToolbarStyle and Item.Checked);

  StateFlags := GetStateFlags(ItemInfo);

  Canvas.Font := TTBViewAccess(View).GetFont;
  Canvas.Font.Color := CurrentTheme.GetItemTextColor(ItemInfo);
  DoAdjustFont(Canvas.Font, StateFlags);
  C := Canvas.Font.Color;

  { Setup font }
  TextFlags := GetTextFlags;
  IsCaptionShown := CaptionShown;
  IsTextRotated := (View.Orientation = tbvoVertical) and ToolbarStyle;
  if IsCaptionShown then
  begin
    S := GetCaptionText;
    if (Item.Layout <> tbxlAuto) or (tboImageAboveCaption in Item.EffectiveOptions) then
      IsTextRotated := False;
    if IsTextRotated or not ToolbarStyle then TextFlags := TextFlags or DT_SINGLELINE;
    TextSize := GetTextSize(Canvas, S, TextFlags, IsTextRotated, StateFlags);
  end
  else
  begin
    StateFlags := 0;
    SetLength(S, 0);
    IsTextRotated := False;
    TextSize.CX := 0;
    TextSize.CY := 0;
  end;

  IsSpecialDropDown := HasArrow and not IsSplit and ToolbarStyle and
    ((Item.Layout = tbxlGlyphTop) or (Item.Layout = tbxlAuto) and (tboImageAboveCaption in Item.EffectiveOptions)) and
    (ImgSize.CX > 0) and not (IsTextRotated) and (TextSize.CX > 0);

  { Border & Arrows }
  R := ClientAreaRect;
  with CurrentTheme do if ToolbarStyle then
  begin
    GetMargins(MID_TOOLBARITEM, Margins);
    if HasArrow then with R do
    begin
      ItemInfo.ComboPart := cpCombo;
      if IsSplit then
      begin
        ItemInfo.ComboPart := cpSplitLeft;
        ComboRect := R;
        Dec(Right, SplitBtnArrowWidth);
        ComboRect.Left := Right;
      end
      else if not IsSpecialDropDown then
      begin
        if View.Orientation <> tbvoVertical then
          ComboRect := Rect(Right - DropdownArrowWidth - DropdownArrowMargin, 0,
            Right - DropdownArrowMargin, Bottom)
        else
          ComboRect := Rect(0, Bottom - DropdownArrowWidth - DropdownArrowMargin,
            Right, Bottom - DropdownArrowMargin);
      end
      else
      begin
        ImgAndArrowWidth := ImgSize.CX + DropdownArrowWidth + 2;
        ComboRect.Right := (R.Left + R.Right + ImgAndArrowWidth + 2) div 2;
        ComboRect.Left := ComboRect.Right - DropdownArrowWidth;
        ComboRect.Top := (R.Top + R.Bottom - ImgSize.CY - 2 - TextSize.CY) div 2;
        ComboRect.Bottom := ComboRect.Top + ImgSize.CY;
      end;
    end
    else SetRectEmpty(ComboRect);

    if not IsSplit then
    begin
      PaintDefault := True;
      DoDrawButton(Canvas, R, ItemInfo, pstPrePaint, PaintDefault);
      if PaintDefault then
        CurrentTheme.PaintButton(Canvas, R, ItemInfo);
      PaintDefault := True;
      DoDrawButton(Canvas, R, ItemInfo, pstPostPaint, PaintDefault);

      if HasArrow then
      begin
        PaintDropDownArrow(Canvas, ComboRect, ItemInfo);
        if not IsSpecialDropDown then
        begin
          if View.Orientation <> tbvoVertical then Dec(R.Right, DropdownArrowWidth)
          else Dec(R.Bottom, DropdownArrowWidth);
        end;
      end;
    end
    else // IsSplit
    begin
      PaintDefault := True;
      DoDrawButton(Canvas, R, ItemInfo, pstPrePaint, PaintDefault);
      if PaintDefault then
        CurrentTheme.PaintButton(Canvas, R, ItemInfo);
      PaintDefault := True;
      DoDrawButton(Canvas, R, ItemInfo, pstPostPaint, PaintDefault);

      ItemInfo.Pushed := IsComboPushed;
      ItemInfo.Selected := False;
      ItemInfo.ComboPart := cpSplitRight;

      PaintDefault := True;
      DoDrawButton(Canvas, ComboRect, ItemInfo, pstPrePaint, PaintDefault);
      if PaintDefault then
        CurrentTheme.PaintButton(Canvas, ComboRect, ItemInfo);
      PaintDefault := True;
      DoDrawButton(Canvas, ComboRect, ItemInfo, pstPostPaint, PaintDefault);

      ItemInfo.ComboPart := cpSplitLeft;
      ItemInfo.Pushed := IsPushed;
      ItemInfo.Selected := Item.Checked;
    end;

    InflateRect(R, -2, -2);
  end
  else
  begin
    GetMargins(MID_MENUITEM, Margins);

    PaintDefault := True;
    DoDrawButton(Canvas, R, ItemInfo, pstPrePaint, PaintDefault);
    if PaintDefault then
      CurrentTheme.PaintMenuItem(Canvas, R, ItemInfo);
    PaintDefault := True;
    DoDrawButton(Canvas, R, ItemInfo, pstPostPaint, PaintDefault);

    Inc(R.Left, Margins.LeftWidth);
    Dec(R.Right, Margins.RightWidth);
    Inc(R.Top, Margins.TopHeight);
    Dec(R.Bottom, Margins.BottomHeight);
  end;

  { Caption }
  if IsCaptionShown then
  begin
    if ToolbarStyle then
    begin
      TextFlags := TextFlags or DT_CENTER or DT_VCENTER;
      CaptionRect := R;
      TextAlignment := Self.Item.Alignment;

      if ImageIsShown then
        Case ItemLayout of
          tbxlGlyphLeft:
            begin
              Inc(CaptionRect.Left, ImgSize.CX + 3);
              TextFlags := TextFlags and not DT_CENTER;
            end;
          tbxlGlyphTop:
            begin
              Inc(CaptionRect.Top, ImgSize.CY + 1);
              if IsTextRotated then Inc(CaptionRect.Top, 3);
              TextFlags := TextFlags and not DT_VCENTER;
            end;
          end;

      Case TextAlignment of
        taCenter:
          begin
            CaptionRect.Left := (CaptionRect.Left + CaptionRect.Right - TextSize.CX) div 2;
            CaptionRect.Top := (CaptionRect.Top + CaptionRect.Bottom - TextSize.CY) div 2;
            CaptionRect.Right := CaptionRect.Left + TextSize.CX;
            CaptionRect.Bottom := CaptionRect.Top + TextSize.CY;
          end;
        taLeftJustify:
          begin
            CaptionRect.Left := CaptionRect.Left + 3;
            CaptionRect.Top := (CaptionRect.Top + CaptionRect.Bottom - TextSize.CY) div 2;
            CaptionRect.Right := CaptionRect.Left + TextSize.CX;
            CaptionRect.Bottom := CaptionRect.Top + TextSize.CY;
          end;
        taRightJustify:
          begin
            CaptionRect.Left := CaptionRect.Right - TextSize.CX - 3;
            CaptionRect.Top := (CaptionRect.Top + CaptionRect.Bottom - TextSize.CY) div 2;
            CaptionRect.Right := CaptionRect.Left + TextSize.CX;
            CaptionRect.Bottom := CaptionRect.Top + TextSize.CY;
          end;
      end;
    end
    else with CurrentTheme do
    begin
      TextFlags := DT_LEFT or DT_VCENTER or TextFlags;
      TextSize := GetTextSize(Canvas, S, TextFlags, False, StateFlags); { TODO : Check if this line is required }
      GetTextMetrics(Canvas.Handle, TextMetrics);

      CaptionRect := R;
      Inc(CaptionRect.Left, ItemInfo.PopupMargin + MenuImageTextSpace + MenuLeftCaptionMargin);
      with TextMetrics, CaptionRect do
        if (Bottom - Top) - (tmHeight + tmExternalLeading) = Margins.BottomHeight then Dec(Bottom);
      Inc(CaptionRect.Top, TextMetrics.tmExternalLeading);

      CaptionRect.Right := CaptionRect.Left + TextSize.CX;
    end;

    WS := Item.Caption;
    Canvas.Font.Color := C;
    PaintDefault := True;
    DoDrawCaption(Canvas, ClientAreaRect, WS, CaptionRect, IsTextRotated, pstPrePaint, PaintDefault);
    if PaintDefault then
      CurrentTheme.PaintCaption(Canvas, CaptionRect, ItemInfo, S, TextFlags, IsTextRotated);
    PaintDefault := True;
    DoDrawCaption(Canvas, ClientAreaRect, WS, CaptionRect, IsTextRotated, pstPostPaint, PaintDefault);
  end;

  { Shortcut and/or submenu arrow (menus only) }
  if not ToolbarStyle then
  begin
    WS := Item.GetShortCutText;
    if Length(WS) > 0 then
    begin
      CaptionRect := R;
      with CaptionRect, TextMetrics do
      begin
        Left := Right - (Bottom - Top) - GetTextWidth(Canvas.Handle, WS, True);
        if (Bottom - Top) - (tmHeight + tmExternalLeading) = Margins.BottomHeight then Dec(Bottom);
        Inc(Top, TextMetrics.tmExternalLeading);
      end;
      Canvas.Font.Color := C;
      PaintDefault := True;
      DoDrawCaption(Canvas, ClientAreaRect, WS, CaptionRect, IsTextRotated, pstPrePaint, PaintDefault);
      if PaintDefault then
        CurrentTheme.PaintCaption(Canvas, CaptionRect, ItemInfo, WS, TextFlags, False);
      PaintDefault := True;
      DoDrawCaption(Canvas, ClientAreaRect, WS, CaptionRect, IsTextRotated, pstPostPaint, PaintDefault);
    end;
  end;

  { Image, or check box }
  if ImageOrCheckShown then
  begin
    ImageRect := R;

    if ToolBarStyle then
    begin
      if IsSpecialDropDown then OffsetRect(ImageRect, (-CurrentTheme.DropdownArrowWidth + 1) div 2, 0);
      if ItemLayout = tbxlGlyphLeft then ImageRect.Right := ImageRect.Left + ImgSize.CX + 2
      else
      begin
        ImageRect.Top := (ImageRect.Top + ImageRect.Bottom - ImgSize.cy - 2 - TextSize.cy) div 2;
        ImageRect.Bottom := ImageRect.Top + ImgSize.CY;
      end;
    end
    else ImageRect.Right := ImageRect.Left + ClientAreaRect.Bottom - ClientAreaRect.Top;

    if ImageIsShown then with ImageRect, ImgSize do
    begin
      Left := Left + ((Right - Left) - CX) div 2;
      ImageRect.Top := Top + ((Bottom - Top) - CY) div 2;
      Right := Left + CX;
      Bottom := Top + CY;
      DrawItemImage(Canvas, ImageRect, ItemInfo);
    end
    else if not ToolbarStyle and Item.Checked then
      CurrentTheme.PaintCheckMark(Canvas, ImageRect, ItemInfo);
  end;
end;

function TSpTBXItemViewer.GetItem: TSpTBXCustomItem;
begin
  Result := (inherited Item) as TSpTBXCustomItem;
end;

function TSpTBXItemViewer.GetHintText: Widestring;
var
  I: Integer;
begin
  // Get the short hint
  I := Pos('|', Item.Hint);
  if I = 0 then
    Result := Item.Hint
  else
    Result := Copy(Item.Hint, 1, I - 1);
  // Use the caption if there is no hint
  if (Result = '') and
     (not(tbisSubmenu in Item.ItemStyle) or (tbisCombo in Item.ItemStyle) or
      not CaptionShown) then
    Result := Item.Caption;
  // Add shortcut text
  if (Result <> '') and Application.HintShortCuts and (Item.ShortCut <> scNone) then
    Result := Result + ' (' + ShortCutToText(Item.ShortCut) + ')';
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TTBXSubmenuItem }

constructor TSpTBXSubmenuItem.Create(AOwner: TComponent);
begin
  inherited;
  ItemStyle := ItemStyle + [tbisSubMenu, tbisSubitemsEditable];
end;

function TSpTBXSubmenuItem.GetDropdownCombo: Boolean;
begin
  Result := tbisCombo in ItemStyle;
end;

procedure TSpTBXSubmenuItem.SetDropdownCombo(Value: Boolean);
begin
  if (tbisCombo in ItemStyle) <> Value then begin
    if Value then ItemStyle := ItemStyle + [tbisCombo]
    else ItemStyle := ItemStyle - [tbisCombo];
    Change(True);
  end;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXCustomLabelItem }

function TSpTBXCustomLabelItem.DialogChar(CharCode: Word): Boolean;
begin
  Result := inherited DialogChar(CharCode);
  if Assigned(Control) and (Control is TWinControl) and
    TWinControl(Control).CanFocus and IsAccel(CharCode, Caption) then
  begin
    TWinControl(Control).SetFocus;
    Result := True;
  end;
end;

constructor TSpTBXCustomLabelItem.Create(AOwner: TComponent);
begin
  inherited;
  ItemStyle := ItemStyle - [tbisSelectable, tbisRedrawOnSelChange,
    tbisRedrawOnMouseOverChange] + [tbisClicksTransparent];
  Alignment := taLeftJustify;
  Stretch := False;
end;

procedure TSpTBXCustomLabelItem.DoDrawButton(ACanvas: TCanvas; ARect: TRect;
  ItemInfo: TTBXItemInfo; const PaintStage: TSpTBXPaintStage;
  var PaintDefault: Boolean);
begin
  PaintDefault := True;
  inherited DoDrawButton(ACanvas, ARect, ItemInfo, PaintStage, PaintDefault);
  PaintDefault := False;
end;

procedure TSpTBXCustomLabelItem.ToggleControl;
begin
  // Do nothing, the Control property is not valid
end;

procedure TSpTBXCustomLabelItem.UpdateProps;
begin
  // Do nothing, the Control property is not valid
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXSeparatorItem }

function TSpTBXSeparatorItem.GetItemViewerClass(AView: TTBView): TTBItemViewerClass;
begin
  Result := TSpTBXSeparatorItemViewer;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXSeparatorItemViewer }

procedure TSpTBXSeparatorItemViewer.CalcSize(const Canvas: TCanvas;
  var AWidth, AHeight: Integer);
begin
  inherited CalcSize(Canvas, AWidth, AHeight);
end;

function TSpTBXSeparatorItemViewer.IsStatusBarSeparator(out T: TSpTBXStatusToolbar): Boolean;
var
  C: TComponent;
begin
  Result := False;
  T := nil;
  C := Item.GetParentComponent;
  if C is TSpTBXStatusToolbar then begin
    T := C as TSpTBXStatusToolbar;
    Result := True;
  end;
end;

procedure TSpTBXSeparatorItemViewer.Paint(const Canvas: TCanvas;
  const ClientAreaRect: TRect; IsHoverItem, IsPushed,
  UseDisabledShadow: Boolean);
var
  T: TSpTBXStatusToolbar;
begin
  if not (IsStatusBarSeparator(T) and T.NeedsSeparatorRepaint) then
    inherited;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXThemeGroupItem }

constructor TSpTBXThemeGroupItem.Create(AOwner: TComponent);
begin
  inherited;
  FThemesList := TStringList.Create;
  if not (csDesigning in ComponentState) then
    RecreateThemesList;
end;

destructor TSpTBXThemeGroupItem.Destroy;
begin
  FThemesList.Free;
  inherited;
end;

procedure TSpTBXThemeGroupItem.DoClick;
begin
  if Assigned(FOnClick) then FOnClick(Self);
end;

function TSpTBXThemeGroupItem.DoGetItemCaption(ThemeName: WideString): WideString;
begin
  Result := ThemeName;
  if Assigned(FOnGetCaption) then FOnGetCaption(Self, Result);
end;

procedure TSpTBXThemeGroupItem.ItemClickEvent(Sender: TObject);
var
  Item: TSpTBXItem;
begin
  Item := Sender as TSpTBXItem;
  if not Item.Checked and (Item.Tag > -1) and (Item.Tag < FThemesList.Count) then
  begin
    Item.Checked := True;
    TBXSetTheme(FThemesList[Item.Tag]);
    DoClick;
  end;
end;

procedure TSpTBXThemeGroupItem.RecreateThemesList;
var
  I: Integer;
  A: TSpTBXItem;
  S: String;
begin
  // Delete FThemesList items
  FThemesList.Clear;
  for I := 0 to Count - 1 do
    if Items[I].GroupIndex = 888 then
      Delete(I);

  // Fill and sort the FThemesList, move the Default theme to the top
  S := TBXCurrentTheme;
  TBXThemes.GetAvailableTBXThemes(FThemesList);
  FThemesList.Sort;
  I := FThemesList.IndexOf('Default');
  if I > -1 then FThemesList.Move(I, 0);

  // Create group items
  for I := 0 to FThemesList.Count - 1 do begin
    A := TSpTBXItem.Create(Self);
    A.Caption := DoGetItemCaption(FThemesList[I]);
    A.AutoCheck := False;
    A.GroupIndex := 888;
    A.Tag := I;
    A.OnClick := ItemClickEvent;
    Insert(FInsertIndex + I, A);
    if FThemesList[I] = S then A.Click;
  end;

  if Assigned(FOnUpdate) then FOnUpdate(Self);
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXItemCache }

procedure TSpTBXItemCache.Assign(Source: TPersistent);
var
  C: TSpTBXItemCache;
begin
  if Source is TSpTBXItemCache then
  begin
    C := Source as TSpTBXItemCache;
    Item := C.Item;
    Left := C.Left;
    Top := C.Top;
    Width := C.Width;
    Height := C.Height;
    ParentWidth := C.ParentWidth;
    ParentHeight := C.ParentHeight;
  end
  else inherited Assign(Source);
end;

function TSpTBXItemCache.GetName: TComponentName;
begin
  if Assigned(FItem) then
    Result := FItem.Name
  else
    Result := FName;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXItemCacheCollection }

function TSpTBXItemCacheCollection.Add: TSpTBXItemCache;
begin
  Result := TSpTBXItemCache(inherited Add);
end;

function TSpTBXItemCacheCollection.Add(Item: TTBCustomItem; RBounds: TRect): Integer;
var
  F: TSpTBXItemCache;
begin
  F := Add;
  F.Item := Item;
  F.Left := RBounds.Left;
  F.Top := RBounds.Top;
  F.Width := RBounds.Right - RBounds.Left;
  F.Height := RBounds.Bottom - RBounds.Top;
  Result := F.Index;
end;

function TSpTBXItemCacheCollection.GetItem(Index: Integer): TSpTBXItemCache;
begin
  Result := TSpTBXItemCache(inherited Items[Index]);
end;

function TSpTBXItemCacheCollection.IndexOf(Item: TTBCustomItem): Integer;
var
  I: integer;
begin
  Result := -1;
  if Assigned(Item) then
    for I := 0 to Count - 1 do
      if Items[I].Item = Item then begin
        Result := I;
        Break;
      end;
end;

procedure TSpTBXItemCacheCollection.SetItem(Index: Integer;
  const Value: TSpTBXItemCache);
begin
  inherited Items[Index] := Value;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXDock }

function TSpTBXDock.CanResize(var NewWidth, NewHeight: Integer): Boolean;
begin
  FPrevWidth := Width;
  FPrevHeight := Height;
  Result := inherited CanResize(NewWidth, NewHeight);
end;

procedure TSpTBXDock.DoDrawBackground(ACanvas: TCanvas; ARect: TRect;
  const PaintStage: TSpTBXPaintStage; var PaintDefault: Boolean);
begin
  if Assigned(FOnDrawBackground) then FOnDrawBackground(Self, ACanvas, ARect,
   PaintStage, PaintDefault);
end;

procedure TSpTBXDock.DrawBackground(DC: HDC; const DrawRect: TRect);
var
  Canvas: TCanvas;
  PaintDefault: Boolean;
begin
  // OnDrawBackground should be used to paint all the toolbars + dock, it is
  // used by TSpTBXStatusBar and TSpTBXDockablePanel to paint the whole client
  // area with custom painting.
  // Assigning OnDrawBackground sets ThemedBackground to True.
  // OnDrawBackground is triggered by the Dock and by the docked Toolbar.
  // The Toolbar triggers it only if Dock.ThemedBackground (TTBXDock.UsingBackground)
  // is true, which depends on CurrentTheme.PaintDockBackground, this is done in
  // TTBXToolbar.WMEraseBkgnd.
  // When (ARect.Left = 0) and (ARect.Top = 0) the sender is the Dock, else
  // it's the docked Toolbar.

  Canvas := TCanvas.Create;
  try
    Canvas.Handle := DC;

    PaintDefault := True;
    DoDrawBackground(Canvas, DrawRect, pstPrePaint, PaintDefault);
    if PaintDefault then
      inherited;

    PaintDefault := True;
    DoDrawBackground(Canvas, DrawRect, pstPostPaint, PaintDefault);
  finally
    Canvas.Handle := 0;
    Canvas.Free;
  end;
end;

procedure TSpTBXDock.Resize;
var
  I: Integer;
begin
  inherited;
  if Width < FPrevWidth then
    for I := 0 to ToolbarCount - 1 do
      if Toolbars[I] is TSpTBXToolbar then
        TSpTBXToolbar(Toolbars[I]).Resize;
end;

function TSpTBXDock.ThemedBackground: Boolean;
begin
  if Assigned(FOnDrawBackground) then
    Result := True
  else
    Result := inherited ThemedBackground;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXToolbar }

constructor TSpTBXToolbar.Create(AOwner: TComponent);
begin
  inherited;
  Items.RegisterNotification(DoItemNotification);
  FAnchoredControlItems := TSpTBXItemCacheCollection.Create(TSpTBXItemCache);
end;

destructor TSpTBXToolbar.Destroy;
begin
  Items.UnRegisterNotification(DoItemNotification);
  FAnchoredControlItems.Free;  
  inherited;
end;

procedure TSpTBXToolbar.CMHintShow(var Message: TCMHintShow);
// Handle the CM_HINTSHOW message to show unicode hints using
// a custom THintWindow.
var
  HintInfo: PHintInfo;
  V: TTBItemViewer;
  WideHint: Widestring;
  R, TextR: TRect;
  PaintDefault: Boolean;
begin
  HintInfo := Message.HintInfo;
  V := FView.ViewerFromPoint(HintInfo.CursorPos);
  if Assigned(V) and not IsRectEmpty(V.BoundsRect) and (V is TSpTBXItemViewer) then
  begin
    // Prepare the HintInfo
    WideHint := TSpTBXItemViewer(V).GetHintText;
    HintInfo.HintStr := WideHint;
    HintInfo.CursorRect := V.BoundsRect;
    HintInfo.HintWindowClass := TBitmapHint;   //custom HintWindow class
    HintInfo.HintData := SpStockHintBitmap;  //TApplication.ActivateHint will pass the data to the HintWindow
    //HintInfo.HideTimeout := 60000; //1 minute

    // Prepare the HintBitmap
    SpStockHintBitmap.Canvas.Font.Assign(Self.Font);
    SpStockHintBitmap.Canvas.Font.Color := clInfoText;
    SpStockHintBitmap.Canvas.Pen.Color := clBlack;
    SpStockHintBitmap.Canvas.Brush.Color := clInfoBk;
    TextR := Rect(0, 0, 1, 1);
    SpDrawXPText(SpStockHintBitmap.Canvas, WideHint, TextR, DT_NOPREFIX or DT_CALCRECT);
    SpStockHintBitmap.Width := TextR.Right + 8;
    SpStockHintBitmap.Height := TextR.Bottom + 4;

    // Draw the hint in the HintBitmap
    PaintDefault := True;
    TSpTBXItemViewer(V).Item.DoDrawHint(SpStockHintBitmap, WideHint, PaintDefault);
    if PaintDefault then
      DoDrawHint(SpStockHintBitmap, WideHint, PaintDefault); // Inform the toolbar
    if PaintDefault then begin
      R := Rect(0, 0, SpStockHintBitmap.Width, SpStockHintBitmap.Height);
      SpStockHintBitmap.Canvas.FillRect(R);
      OffsetRect(TextR, ((R.Right - TextR.Right) div 2) - 2, (R.Bottom - TextR.Bottom) div 2);
      SpDrawXPText(SpStockHintBitmap.Canvas, WideHint, TextR, DT_NOPREFIX);
    end;
  end
  else
    inherited;
end;

procedure TSpTBXToolbar.DoDrawHint(AHintBitmap: TBitmap;
  var AHint: Widestring; var PaintDefault: Boolean);
begin
  if Assigned(FOnDrawHint) then FOnDrawHint(Self, AHintBitmap, AHint, PaintDefault);
end;

procedure TSpTBXToolbar.DoDrawBackground(ACanvas: TCanvas; ARect: TRect;
  const PaintStage: TSpTBXPaintStage; var PaintDefault: Boolean);
begin
  if Assigned(FOnDrawBackground) then FOnDrawBackground(Self, ACanvas, ARect,
    PaintStage, PaintDefault);
end;

procedure TSpTBXToolbar.DoItemNotification(Ancestor: TTBCustomItem;
  Relayed: Boolean; Action: TTBItemChangedAction; Index: Integer;
  Item: TTBCustomItem);
var
  I: Integer;
begin
  if Assigned(FOnItemNotification) then FOnItemNotification(Self, Ancestor,
    Relayed, Action, Index, Item);
  if not FIsResizing then
    Case Action of
      tbicInserted:
        RightAlignItems;
      tbicDeleting:
        begin
          I := FAnchoredControlItems.IndexOf(Item);
          if I > -1 then
            FAnchoredControlItems.Delete(I);
          RightAlignItems;
        end;
      tbicInvalidateAndResize:
        RightAlignItems;
    end;
end;

procedure TSpTBXToolbar.Resize;
begin
  FIsResizing := True;
  try
    RightAlignItems;
    AnchorItems;
  finally
    FIsResizing := False;
  end;
  inherited;
end;

procedure TSpTBXToolbar.AnchorItems;
var
  I, J: Integer;
  SpIV: TSpTBXItemViewer;
  Size: TPoint;
  CI: TTBControlItem;
begin
  if not Assigned(CurrentDock) or (CurrentDock.Position in [dpLeft, dpRight]) then
    Exit;

  View.BeginUpdate;
  try
    View.ValidatePositions;
    // Resize the anchored items
    Size := Point(CurrentDock.Width, CurrentDock.Height);
    for I := 0 to View.ViewerCount - 1 do
      if View.Viewers[I] is TSpTBXItemViewer then begin
        SpIV := TSpTBXItemViewer(View.Viewers[I]);
        if SpIV.Item.Anchored then begin
          // Revalidate FAnchorSize and set FAnchorDelta
          if (SpIV.FAnchorSize.X = 0) and (SpIV.FAnchorSize.Y = 0) then
            SpIV.FAnchorSize := Size;
          SpIV.FAnchorDelta.X := Size.X - SpIV.FAnchorSize.X;
          SpIV.FAnchorDelta.Y := Size.Y - SpIV.FAnchorSize.Y;
        end;
      end
      else begin
        // Anchor TTBControlItem items if the associated Control is client
        // aligned or has akRight in its Anchors property.
        CI := IsAnchoredControlItem(View.Viewers[I].Item);
        J := FAnchoredControlItems.IndexOf(View.Viewers[I].Item);
        if Assigned(CI) then begin
          // Add the TTBControlItem item to the list if its not there
          if J = -1 then begin
            J := FAnchoredControlItems.Add(CI, Rect(0, 0, CI.Control.Width, CI.Control.Height));
            FAnchoredControlItems[J].ParentWidth := Size.X;
            FAnchoredControlItems[J].ParentHeight := Size.Y;
          end;
          CI.Control.Width := FAnchoredControlItems[J].Width + (Size.X - FAnchoredControlItems[J].ParentWidth)
        end
        else
          // If ControlItem is not valid delete it from the list
          if J > -1 then
            FAnchoredControlItems.Delete(J);
      end;
    View.UpdatePositions;
  finally
    View.EndUpdate;
  end;
end;

function TSpTBXToolbar.IsAnchoredControlItem(Item: TTBCustomItem): TTBControlItem;
var
  CI: TTBControlItem;
begin
  Result := nil;
  if Assigned(CurrentDock) and (Item is TTBControlItem) then begin
    CI := Item as TTBControlItem;
    if Assigned(CI.Control) and
      ((CI.Control.Align = alClient) or (akRight in CI.Control.Anchors)) then
    begin
      Result := CI;
    end
    else
      Result := nil;
  end;
end;

function TSpTBXToolbar.GetRightAlignMargin: Integer;
begin
  Result := 10;
end;

procedure TSpTBXToolbar.RightAlignItems;
var
  I, VisibleWidth, RightAlignedWidth: Integer;
  FirstIV: TTBItemViewer;
  Spacer: TSpTBXItemViewer;
begin
  if not Assigned(CurrentDock) or (CurrentDock.Width = 0) or
    (ShrinkMode <> tbsmNone) or (CurrentDock.Position in [dpLeft, dpRight]) then
      Exit;

  FirstIV := View.NextSelectable(nil, true);
  if Assigned(FirstIV) then begin
    View.ValidatePositions;
    View.BeginUpdate;
    try
      // Find the spacer and the right aligned items
      Spacer := SpGetRightAlignedItems(View, nil, VisibleWidth, RightAlignedWidth);
      if Assigned(Spacer) then begin
        // Resize the spacer
        I := CurrentDock.Width - GetRightAlignMargin - (VisibleWidth - Spacer.Item.CustomWidth);
        if I < 0 then I := 0;
        Spacer.Item.CustomWidth := I;
      end;
      View.UpdatePositions;
    finally
      View.EndUpdate;
    end;
  end;
end;

procedure TSpTBXToolbar.DrawNCArea(const DrawToDC: Boolean; const ADC: HDC;
  const Clip: HRGN);
var
  DC: HDC;
  R, CR: TRect;
  TD: TSpTBXDock;
  ToolbarInfo: TTBXToolbarInfo;
begin
  inherited;

  if HandleAllocated and Docked and Assigned(CurrentDock) and
    (CurrentDock is TSpTBXDock) and Assigned(TSpTBXDock(CurrentDock).OnDrawBackground) then
  begin
    TD := CurrentDock as TSpTBXDock;
    if not DrawToDC then DC := GetWindowDC(Handle)
    else DC := ADC;
    try
      GetToolbarInfo(ToolbarInfo);
      GetWindowRect(Handle, R);
      OffsetRect(R, -R.Left, -R.Top);
      if not DrawToDC then
      begin
        SelectNCUpdateRgn(Handle, DC, Clip);
        CR := R;
        with ToolbarInfo.BorderSize, CR do
        begin
          InflateRect(CR, -X, -Y);
          if ToolbarInfo.IsVertical then Inc(Top, GetTBXDragHandleSize(ToolbarInfo))
          else Inc(Left, GetTBXDragHandleSize(ToolbarInfo));
          ExcludeClipRect(DC, Left, Top, Right, Bottom);
        end;
      end;
      TD.DrawBackground(DC, R);
    finally
      if not DrawToDC then ReleaseDC(Handle, DC);
    end;
  end
  else
    inherited;
end;

procedure TSpTBXToolbar.CMControlChange(var Message: TCMControlChange);
begin
  // When a control is dropped on the toolbar a TTBControlItem is created by
  // TTBCustomToolbar.CreateWrapper, unfortunately it is created with the
  // Self.Owner instead of the Form (Owner.Owner for CompoundToolbars like
  // the TabToolbar or StatusToolbar).

  if CompoundToolbar and Message.Inserting and not(csLoading in ComponentState) and
    not(csUpdating in ComponentState) and
    (SpFindControlItem(Items, Message.Control) = nil) then
  begin
    CreateWrapper(Items.Count, Message.Control);
  end
  else
    inherited;
end;

function TSpTBXToolbar.CreateWrapper(Index: Integer; Ctl: TControl): TTBControlItem;
var
  I: Integer;
  S: String;
  C: TComponent;
begin
  C := Owner.Owner;
  Result := TTBControlItem.CreateControlItem(C, Ctl);
  if (csDesigning in ComponentState) and Assigned(C) then begin
    { Needs a name for compatibility with form inheritance }
    I := 1;
    while True do begin
      S := Format('TBControlItem%d', [I]);
      if C.FindComponent(S) = nil then
        Break;
      Inc(I);
    end;
    Result.Name := S;
  end;
  Items.Insert(Index, Result);
end;

procedure TSpTBXToolbar.CMDialogChar(var Message: TCMDialogChar);
var
  I: Integer;
begin
  for I := 0 to Items.Count - 1 do
    if Items[I] is TSpTBXCustomItem then
      if TSpTBXCustomItem(Items[I]).DialogChar(Message.CharCode) then begin
        Message.Result := 1;
        Exit;
      end;
end;

procedure TSpTBXToolbar.WMEraseBkgnd(var Message: TWMEraseBkgnd);
var
  Canvas: TCanvas;
  PaintDefault: Boolean;
  R: TRect;
  TD: TSpTBXDock;
begin
  if Assigned(CurrentDock) and (CurrentDock is TSpTBXDock) then begin
    TD := CurrentDock as TSpTBXDock;
    if Assigned(TD.OnDrawBackground) then begin
      R := TD.ClientRect;
      R.TopLeft := ScreenToClient(TD.ClientToScreen(R.TopLeft));
      R.BottomRight := ScreenToClient(TD.ClientToScreen(R.BottomRight));
      TD.DrawBackground(Message.DC, R);
      Message.Result := 1;
      Exit;  // The background painting is done by the Dock
    end;
  end;

  if Assigned(FOnDrawBackground) then begin
    Message.Result := 1;
    Canvas := TCanvas.Create;
    Canvas.Handle := Message.DC;
    try
      if Docked then
      begin
        R := CurrentDock.ClientRect;
        R.TopLeft := ScreenToClient(CurrentDock.ClientToScreen(R.TopLeft));
        R.BottomRight := ScreenToClient(CurrentDock.ClientToScreen(R.BottomRight));
      end
      else
        R := ClientRect;

      PaintDefault := True;
      DoDrawBackground(Canvas, R, pstPrePaint, PaintDefault);
      if PaintDefault then
        inherited;

      PaintDefault := True;
      DoDrawBackground(Canvas, R, pstPostPaint, PaintDefault);
    finally
      Canvas.Handle := 0;
      Canvas.Free;
    end;
  end
  else
    inherited;
end;

procedure TSpTBXToolbar.WMSize(var Message: TWMSize);
begin
  inherited;
  Update;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXCompoundItemsControl }

constructor TSpTBXCompoundItemsControl.Create(AOwner: TComponent);
begin
  inherited;

  FDock := GetDockClass.Create(Self);
  FDock.Parent := Self;
  FDock.OnRequestDock := DockRequestDock;

  FToolbar := GetToolbarClass.Create(Self);
  FToolbar.CompoundToolbar := True;
  FToolbar.Parent := FDock;
  FToolbar.CurrentDock := FDock;
  FToolbar.BorderStyle := bsNone;
  FToolbar.DockMode := dmCannotFloatOrChangeDocks;
  FToolbar.DragHandleStyle := dhNone;
  FToolbar.Stretch := True;
  FToolbar.ShrinkMode := tbsmNone;
  FToolbar.ShowCaption := False;
end;

procedure TSpTBXCompoundItemsControl.CreateParams(var Params: TCreateParams);
begin
  // Disable complete redraws when size changes. CS_HREDRAW and CS_VREDRAW
  // cause flicker and are not necessary for this control at run time
  // Invalidate in WMWindowPosChanged message instead.
  inherited CreateParams(Params);
  if not (csDesigning in ComponentState) then
    with Params do
      WindowClass.Style := WindowClass.Style and not (CS_HREDRAW or CS_VREDRAW);
end;

destructor TSpTBXCompoundItemsControl.Destroy;
begin
  FToolbar.Free;
  FDock.Free;
  inherited;
end;

procedure TSpTBXCompoundItemsControl.Loaded;
var
  I: Integer;
  C: TControl;
begin
  inherited;

  // The parent of TTBControlItem.Control should be the toolbar, not Self
  // (as setted in GetChildren for dfm streaming).
  for I := 0 to Items.Count - 1 do
    if Items[I] is TTBControlItem then begin
      C := TTBControlItem(Items[I]).Control;
      if Assigned(C) and (C.Parent <> FToolbar) then
        C.Parent := FToolbar;
    end;
end;

procedure TSpTBXCompoundItemsControl.DockRequestDock(Sender: TObject;
  Bar: TTBCustomDockableWindow; var Accept: Boolean);
begin
  if Assigned(FToolbar) then Accept := Bar = FToolbar;
end;

procedure TSpTBXCompoundItemsControl.GetChildren(Proc: TGetChildProc;
  Root: TComponent);
var
  I: Integer;
  C: TControl;
begin
  // Needed to fake the DFM streaming system because the owner of the items
  // is the Form and not the Toolbar nor Self.
  // But the parent must be the Toolbar.
  // GetChildren is used to pass the children components of Self to the DFM
  // streaming system.
  // We also need to do the same with the controls of TTBControlItems.
  // More info on the Delphi help or Classes.TWriter.WriteData

  TTBRootItemAccess(Items).GetChildren(Proc, Root);
  for I := 0 to Items.Count - 1 do
    if (Items[I] is TTBControlItem) then
      if Assigned(TTBControlItem(Items[I]).Control) then begin
        C := TTBControlItem(Items[I]).Control;
        if SpFindControl(Self, C) = -1 then Proc(C);
      end;
  inherited;
end;

function TSpTBXCompoundItemsControl.GetItems: TTBCustomItem;
begin
  Result := FToolbar.Items;
end;

function TSpTBXCompoundItemsControl.GetRootItems: TTBRootItem;
begin
  Result := FToolbar.Items;
end;

function TSpTBXCompoundItemsControl.GetDockClass: TSpTBXDockClass;
begin
  Result := TSpTBXDock;
end;

function TSpTBXCompoundItemsControl.GetToolbarClass: TSpTBXToolbarClass;
begin
  Result := TSpTBXToolbar;
end;

function TSpTBXCompoundItemsControl.GetView: TTBToolbarView;
begin
  Result := FToolbar.View;
end;

function TSpTBXCompoundItemsControl.GetImages: TCustomImageList;
begin
  if Assigned(FToolbar) then
    Result := FToolbar.Images
  else
    Result := nil;
end;

procedure TSpTBXCompoundItemsControl.SetImages(const Value: TCustomImageList);
begin
  if Assigned(FToolbar) then FToolbar.Images := Value;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXStatusToolbar }

constructor TSpTBXStatusToolbar.Create(AOwner: TComponent);
begin
  inherited;
  FThemeType := thtTBX;
end;

function TSpTBXStatusToolbar.GetGripRect: TRect;
var
  ParentForm: TCustomForm;
begin
  Result := Rect(0, 0, 0, 0);
  if Assigned(CurrentDock) then begin
    ParentForm := GetParentForm(Self);
    if Assigned(ParentForm) and (ParentForm.WindowState = wsNormal) then begin
      Result := CurrentDock.ClientRect;
      with Result do begin
        Top := -3;
        Left := Right - GetSystemMetrics(SM_CXVSCROLL);
      end;
    end;
  end;
end;

function TSpTBXStatusToolbar.NeedsSeparatorRepaint: Boolean;
begin
  // [TBXTheme-Change]
  // Office themes have rectangle panels, the separator needs
  // to be painted by the Toolbar.
  Result := (FThemeType = thtNone) or
    ((FThemeType = thtWindows) and not SpXPThemesAvailable) or
    (Pos('Office', TBXCurrentTheme) > 0);
end;

procedure TSpTBXStatusToolbar.WMNCHitTest(var Message: TWMNCHitTest);
var
  Pt: TPoint;
  GR: TRect;
begin
  inherited;

  // Allow statusbar resize when the Grip is being clicked
  if (Message.Result = HTCLIENT) then begin
    GR := GetGripRect;
    if not IsRectEmpty(GR) then begin
      Pt := ScreenToClient(SmallPointToPoint(Message.Pos));
      if PtInRect(GR, Pt) then
        Message.Result := HTBOTTOMRIGHT;
    end;
  end;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXCustomStatusBar }

constructor TSpTBXCustomStatusBar.Create(AOwner: TComponent);
begin
  inherited;

  FDock.OnDrawBackground := DockDrawBackground;
  FDock.OnResize := DockResize;

  FThemeType := thtTBX;

  Align := alBottom;
  Height := FDock.Height;
  AddThemeNotification(Self);
end;

destructor TSpTBXCustomStatusBar.Destroy;
begin
  RemoveThemeNotification(Self);
  inherited;
end;

procedure TSpTBXCustomStatusBar.DockDrawBackground(Sender: TObject;
  ACanvas: TCanvas; ARect: TRect; const PaintStage: TSpTBXPaintStage;
  var PaintDefault: Boolean);
var
  G: TRect;
  InternalPaintDefault: Boolean;
begin
  if PaintStage = pstPrePaint then begin
    PaintDefault := False; // Don't let the Dock draw itself

    // OnDrawBackground is triggered by the Dock and by the docked Toolbar.
    // The Toolbar triggers it only if Dock.ThemedBackground is true, which depends
    // on CurrentTheme.PaintDockBackground, this is done in
    // TTBXToolbar.WMEraseBkgnd.

    G := ValidateToolbar.GetGripRect;
    if not IsRectEmpty(G) then begin
      // When it's called by the Toolbar the Gripper position should be corrected
      if (ARect.Left = -2) and (ARect.Top = -2) then
        OffsetRect(G, -2, -2);
    end;

    ACanvas.Brush.Color := clBtnFace;
    InternalPaintDefault := True;
    DoDrawBackground(ACanvas, ARect, pstPrePaint, InternalPaintDefault);
    if InternalPaintDefault then begin
      SpDrawXPStatusBar(ACanvas, ARect, G, FThemeType);
      DrawSeparators(ACanvas, ARect, G);
    end;

    InternalPaintDefault := True;
    DoDrawBackground(ACanvas, ARect, pstPostPaint, InternalPaintDefault);
  end;
end;

procedure TSpTBXCustomStatusBar.DockResize(Sender: TObject);
begin
  if Assigned(FDock) then
    if Height <> FDock.Height then
      Height := FDock.Height;
end;

procedure TSpTBXCustomStatusBar.DoDrawBackground(ACanvas: TCanvas;
  ARect: TRect; const PaintStage: TSpTBXPaintStage;
  var PaintDefault: Boolean);
begin
  if Assigned(FOnDrawBackground) then FOnDrawBackground(Self, ACanvas, ARect, PaintStage, PaintDefault);
end;

procedure TSpTBXCustomStatusBar.DrawSeparators(ACanvas: TCanvas; ARect, AGripRect: TRect);
var
  I: Integer;
  IV: TTBItemViewer;
  R: TRect;
  C: TColor;
  NotThemed: Boolean;
begin
  if ValidateToolbar.NeedsSeparatorRepaint then begin
    C := ACanvas.Pen.Color;
    try
      for I := 0 to FToolbar.View.ViewerCount - 1 do begin
        IV := FToolbar.View.Viewers[I];
        if (IV is TSpTBXSeparatorItemViewer) and (not TSpTBXSeparatorItem(IV.Item).Blank) then
        begin
          R := IV.BoundsRect;
          if (ARect.Left = -2) and (ARect.Top = -2) then
            OffsetRect(R, 0, 0)
          else
            OffsetRect(R, 2, 0);
          R.Top := 0;
          R.Bottom := Height;
          R.Left := ((R.Right + R.Left) div 2) - 2;
          R.Right := R.Left + 3;

          NotThemed := (FThemeType = thtNone) or ((FThemeType = thtWindows) and not SpXPThemesAvailable);
          ACanvas.FillRect(R);

          ACanvas.Pen.Color := clWindow;
          ACanvas.PenPos := Point(R.Left, 0);
          ACanvas.LineTo(R.Left, Height);

          if NotThemed then
            ACanvas.Pen.Color := clBtnShadow
          else
            ACanvas.Pen.Color := clWindow; // for OfficeXP theme
          ACanvas.PenPos := Point(R.Right, 0);
          ACanvas.LineTo(R.Right, Height);
        end;
      end;
    finally
      ACanvas.Pen.Color := C;
    end;
  end;
end;

function TSpTBXCustomStatusBar.GetToolbarClass: TSpTBXToolbarClass;
begin
  Result := TSpTBXStatusToolbar;
end;

function TSpTBXCustomStatusBar.ValidateToolbar: TSpTBXStatusToolbar;
begin
  Result := FToolbar as TSpTBXStatusToolbar;
end;

procedure TSpTBXCustomStatusBar.SetThemeType(const Value: TSpTBXThemeType);
begin
  if FThemeType <> Value then begin
    FThemeType := Value;
    if HandleAllocated then
      RedrawWindow(Handle, nil, 0, RDW_ERASE or RDW_INVALIDATE or RDW_ALLCHILDREN);
  end;
end;

procedure TSpTBXCustomStatusBar.TBMThemeChange(var Message: TMessage);
begin
  inherited;
  if Message.WParam = TSC_AFTERVIEWCHANGE then
    if HandleAllocated then
      RedrawWindow(Handle, nil, 0, RDW_ERASE or RDW_INVALIDATE or RDW_ALLCHILDREN);
end;

procedure TSpTBXCustomStatusBar.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
  Message.Result := 1;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TBitmapHint }

procedure TBitmapHint.ActivateHint(Rect: TRect; const AHint: string);
begin
  FActivating := True;
  try
    inherited ActivateHint(Rect, AHint);
  finally
    FActivating := False;
  end;
end;

procedure TBitmapHint.ActivateHintData(Rect: TRect; const AHint: string; AData: Pointer);
begin
  //The AData parameter is a bitmap
  FHintBitmap := TBitmap(AData);
  Rect.Right := Rect.Left + FHintBitmap.Width - 2;
  Rect.Bottom := Rect.Top + FHintBitmap.Height - 2;
  inherited ActivateHintData(Rect, AHint, AData);
end;

procedure TBitmapHint.CMTextChanged(var Message: TMessage);
begin
  Message.Result := 1;
end;

procedure TBitmapHint.Paint;
begin
  if Assigned(FHintBitmap) then
    Canvas.Draw(0, 0, FHintBitmap);
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM

procedure InitializeStock;
begin
  SpStockHintBitmap := TBitmap.Create;
  SpStockHandCursor := LoadCursor(0, IDC_HAND);
  SpThemeNexus := TSpThemeNexus.Create;
end;

procedure FinalizeStock;
begin
  SpStockHintBitmap.Free;
  SpThemeNexus.Free;
end;

initialization
  InitializeStock;

finalization
  FinalizeStock;

end.
