unit SpTBXTabs;

{==============================================================================
Version 0.4

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License.
You may obtain a copy of the License at http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.

The initial developer of this code is Robert Lee.

Requirements:
  - Jordan Russell's Toolbar 2000
    http://www.jrsoftware.org
  - Alex Denisov's TBX
    http://g32.org

Development Notes:
  - All the TBX theme changes and adjustments are marked with '[TBXTheme-Change]'.
  - When an item is hidden the ItemViewer.BoundsRect property is invalid.
  - tbicDeleting item notification is fired after the ItemViewer of the
    Item is destroyed by TTBView, but the Items array still has the Item.

TODO:
  - Make the NC area of the tabset doublebuffered.

History:
21 July 2004 - version 0.4
  - Fixed TSpTBXTabControl design time bug, it was allowing to drop
    components when ActiveTabIndex = -1
  - Fixed TSpTBXTabSet and TSpTBXTabControl design time bug, the
    hidden items were not streamed to the DFM.
  - Changed TSpTBXTabControl.OnTabClick event for OnActiveTabChange.
  - Added GetPage method to TSpTBXTabControl to get the TSpTBXTabSheet
    linked to a TSpTBXTabItem.

12 July 2004 - version 0.3.1
  - Fixed nasty AV when setting TBXSwitcher.EnableXPStyles to false,
    thanks to Alfred for reporting this.
    Note: TBXThemeManager unloads the theme library and the theme
    parts when some conditions are met, we must handle extra theme
    parts outside TBXThemeManager space.
  - Fixed incorrect TSpTBXTabSet.ActiveTabIndex property update at
    design time.
  - Fixed incorrect TSpTBXTabSet painting on some TBX themes, thanks
    to Tim for reporting this.

9 July 2004 - version 0.3
  - Fixed design time AVs when moving or deleting TabSheets.
  - Published ThemeType and TabPosition properties for TSpTBXTabItem.
  - New component added, TSpTBXTabSet, a fully customizable TabSet
    with unicode and toolbar items support.

28 June 2004 - version 0.2
  - No changes.

22 June 2004 - version 0.1
  - Initial release.

==============================================================================}

interface

uses
  Windows, Messages, Classes, SysUtils, Controls, Graphics, ImgList, Forms,
  ExtCtrls, TB2Item, TB2Dock, TB2Toolbar, TB2Common, TBX, TBXDkPanels,
  TBXThemes, SpTBXItem;

const
  C_SpTBXTabGroupIndex = 7777;
  WM_INVALIDATETABBACKGROUND = WM_USER + 7777;

type
  TSpTBXTabThemeType = (
    tttNone,         // No themes, no hovered tabs
    tttFlat,         // Flat active tab and TBX themes on hovered tabs
    tttWindows,      // Windows XP theme if available, otherwise don't use themes
    tttTBX           // TBX themes for active and hovered tabs
  );

  TSpTBXTabEdge = (
    tedNone,         // No edge needed
    tedLeft,         // Left edge of the tab
    tedRight         // Right edge of the tab
  );

  TSpTBXTabPosition = (
    ttpTop,          // Top aligned tabset
    ttpBottom        // Bottom aligned tabset
  );

  { TSpTBXTabItem }

  TSpTBXTabToolbar = class;
  TSpTBXCustomTabSet = class;
  TSpTBXCustomTabControl = class;

  TSpTBXTabItem = class(TSpTBXCustomItem)
  private
    FTabPosition: TSpTBXTabPosition;
    FThemeType: TSpTBXTabThemeType;
    function GetTabColor: TColor;
    function GetTabPosition: TSpTBXTabPosition;
    function GetThemeType: TSpTBXTabThemeType;
    procedure SetTabPosition(const Value: TSpTBXTabPosition);
    procedure SetThemeType(const Value: TSpTBXTabThemeType);
  protected
    procedure DrawBottomBorder(ACanvas: TCanvas; ARect: TRect);
    procedure DoDrawButton(ACanvas: TCanvas; ARect: TRect; ItemInfo: TTBXItemInfo;
      const PaintStage: TSpTBXPaintStage; var PaintDefault: Boolean); override;
    procedure DoDrawCaption(ACanvas: TCanvas; const ClientAreaRect: TRect;
      var ACaption: WideString; var CaptionRect: TRect; IsTextRotated: Boolean;
      const PaintStage: TSpTBXPaintStage; var PaintDefault: Boolean); override;
    procedure DoDrawImage(ACanvas: TCanvas; const ItemInfo: TTBXItemInfo;
      const PaintStage: TSpTBXPaintStage; AImageList: TCustomImageList;
      var AImageIndex: Integer; var ARect: TRect;
      var PaintDefault: Boolean); override;
    procedure DoDrawTab(ACanvas: TCanvas; ARect: TRect;
      AChecked, AHoverItem: Boolean; Position: TSpTBXTabPosition;
      ASeparator: Boolean = False; AEdge: TSpTBXTabEdge = tedNone); virtual;
    function GetTabToolbar(out TabToolbar: TSpTBXTabToolbar): Boolean;
    procedure ToggleControl; override;
    property Control;  // TabSheet
  public
    constructor Create(AOwner: TComponent); override;
    procedure Click; override;
    function GetNextTab(GoForward: Boolean; Inmediate: Boolean = True): TSpTBXTabItem;
    function IsFirstTab: Boolean;
    function IsLastTab: Boolean;
  published
    property Action;
    property Checked;
    // Hide DisplayMode
    // property DisplayMode default nbdmImageAndText;
    property Enabled;
    property FontSettings;
    // Hide GroupIndex, all the TabItems must have the same GroupIndex
    // property GroupIndex;
    property HelpContext;
    property ImageIndex;
    property Images;
    property InheritOptions;
    property Layout;
    property MaskOptions;
    property MinHeight;
    property MinWidth;
    property Options;
    property ShortCut;
    // Hide Stretch, all the TabItems must have Stretch setted to True
    // property Stretch default True;
    property Visible;
    property OnAdjustFont;
    property OnClick;
    property OnSelect;
    // TSpTBXCustomItem properties
    property Alignment;
    property CustomWidth;
    property CustomHeight;
    property OnDrawImage;
    property OnDrawItem;
    property OnDrawHint;
    property OnDrawCaption;
    // TSpTBXTabItem properties
    property TabColor: TColor read GetTabColor;
    property TabPosition: TSpTBXTabPosition read GetTabPosition write SetTabPosition default ttpTop;
    property ThemeType: TSpTBXTabThemeType read GetThemeType write SetThemeType default tttWindows;
  end;

  { TSpTBXTabToolbar }
  TSpTBXActiveTabChangeEvent = procedure(Sender: TObject; ItemIndex: Integer) of object;

  TSpTBXTabToolbar = class(TSpTBXToolbar)
  private
    FTabColor: TColor;
    FThemeType: TSpTBXTabThemeType;
    FUpdatingHidden: Boolean;
    FTabPosition: TSpTBXTabPosition;
    FOnActiveTabChange: TSpTBXActiveTabChangeEvent;
    function GetActiveTab: TSpTBXTabItem;
    procedure SetActiveTabIndex(Value: integer);
    procedure SetThemeType(const Value: TSpTBXTabThemeType);
    procedure SetTabColor(const Value: TColor);
    procedure SetTabPosition(const Value: TSpTBXTabPosition);
    procedure CMDesignHitTest(var Message: TCMDesignHitTest); message CM_DESIGNHITTEST;
  protected
    FHiddenTabs: TSpTBXItemCacheCollection;
    FActiveTabIndex: integer;
    FParentTabControl: TSpTBXCustomTabSet;
    procedure DrawNCArea(const DrawToDC: Boolean; const ADC: HDC; const Clip: HRGN); override;
    procedure DoItemNotification(Ancestor: TTBCustomItem; Relayed: Boolean;
      Action: TTBItemChangedAction; Index: Integer; Item: TTBCustomItem); override;
    procedure DoActiveTabChange(const ItemIndex: Integer); virtual;
    procedure RightAlignItems; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure InvalidateActiveTab;
    procedure InvalidateLeftAndRightTabs(ATab: TSpTBXTabItem);
    procedure InvalidateNC;
    procedure ScrollLeft;
    procedure ScrollRight;
    procedure ScrollState(out CanScrollToLeft, CanScrollToRight: Boolean);
  published
    property ActiveTab: TSpTBXTabItem read GetActiveTab;
    property ActiveTabIndex: integer read FActiveTabIndex write SetActiveTabIndex default 0;
    property TabColor: TColor read FTabColor write SetTabColor default clBtnFace;
    property TabPosition: TSpTBXTabPosition read FTabPosition write SetTabPosition default ttpTop;
    property ThemeType: TSpTBXTabThemeType read FThemeType write SetThemeType default tttWindows;
    property OnActiveTabChange: TSpTBXActiveTabChangeEvent read FOnActiveTabChange write FOnActiveTabChange;
  end;

  { TSpTBXTabSheet }

  TSpTBXTabSheet = class(TCustomControl)
  private
    FTabControl: TSpTBXCustomTabSet;
    FItem: TSpTBXTabItem;
    FItemName: String;
    procedure ReadItemName(Reader: TReader);
    procedure WriteItemName(Writer: TWriter);
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
  protected
    procedure AdjustClientRect(var Rect: TRect); override;
    procedure DefineProperties(Filer: TFiler); override;
    procedure ReadState(Reader: TReader); override;
    property Align default alClient;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Item: TSpTBXTabItem read FItem write FItem;
    property TabControl: TSpTBXCustomTabSet read FTabControl write FTabControl;
  end;

  { TSpTBXTabControl }

  TSpTBXCustomTabSet = class(TSpTBXCompoundItemsControl)
  private
    FTabEmpty: Boolean;
    FLoadingActiveIndex: Integer;
    FUpdatingIndex: Boolean;
    FOnDrawBackground: TSpTBXDrawEvent;
    FOnActiveTabChange: TSpTBXActiveTabChangeEvent;
    procedure ReadHiddenItems(Reader: TReader);
    procedure WriteHiddenItems(Writer: TWriter);
    function GetActiveTabIndex: integer;
    procedure SetActiveTabIndex(Value: integer);
    function GetThemeType: TSpTBXTabThemeType;
    procedure SetThemeType(const Value: TSpTBXTabThemeType);
    function GetTabBackgroundColor: TColor;
    procedure SetTabBackgroundColor(const Value: TColor);
    function GetTabPosition: TSpTBXTabPosition;
    procedure SetTabPosition(const Value: TSpTBXTabPosition);
    procedure CMColorchanged(var Message: TMessage); message CM_COLORCHANGED;
    procedure TBMThemeChange(var Message: TMessage); message TBM_THEMECHANGE;
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure WMInvalidateTabBackground(var Message: TMessage); message WM_INVALIDATETABBACKGROUND;
    procedure WMWindowPosChanged(var Message: TWMWindowPosChanged); message WM_WINDOWPOSCHANGED;
  protected
    FBackground: TBitmap;
    // Painting
    procedure DoDrawBackground(ACanvas: TCanvas; ARect: TRect;
      const PaintStage: TSpTBXPaintStage; var PaintDefault: Boolean); virtual;
    function GetFullRepaint: Boolean; virtual;
    procedure InvalidateBackground(Resizing: Boolean = False); virtual;

    // Tabs
    procedure DoActiveTabChange(const ItemIndex: Integer); virtual;
    procedure ItemNotification(Ancestor: TTBCustomItem; Relayed: Boolean;
      Action: TTBItemChangedAction; Index: Integer; Item: TTBCustomItem); virtual;  // Items change notification
    procedure TabInserted(Item: TSpTBXTabItem); virtual;
    procedure TabDeleting(Item: TSpTBXTabItem; FreeTabSheet: Boolean = True); virtual;

    // Component
    procedure DefineProperties(Filer: TFiler); override;
    function GetToolbarClass: TSpTBXToolbarClass; override;
    procedure Loaded; override;
    function ValidateToolbar: TSpTBXTabToolbar;

    property Color default clBtnFace;
    property ParentColor default False;
    property ActiveTabIndex: Integer read GetActiveTabIndex write SetActiveTabIndex default 0;
    property TabBackgroundColor: TColor read GetTabBackgroundColor write SetTabBackgroundColor default clNone;
    property TabPosition: TSpTBXTabPosition read GetTabPosition write SetTabPosition default ttpTop;
    property ThemeType: TSpTBXTabThemeType read GetThemeType write SetThemeType default tttWindows;
    property OnActiveTabChange: TSpTBXActiveTabChangeEvent read FOnActiveTabChange write FOnActiveTabChange;
    property OnDrawBackground: TSpTBXDrawEvent read FOnDrawBackground write FOnDrawBackground;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Add(ACaption: WideString): TSpTBXTabItem;
    function Insert(NewIndex: Integer; ACaption: WideString): TSpTBXTabItem;
    function DrawBackground(DC: HDC; ARect: TRect): Boolean;
    function GetTabSetHeight: Integer;
    procedure ScrollLeft;
    procedure ScrollRight;
    procedure ScrollState(out Left, Right: Boolean);
    procedure TabClick(const ATab: TSpTBXTabItem); virtual;
    property Canvas;
  end;

  TSpTBXTabSet = class(TSpTBXCustomTabSet)
  published
    property Align;
    property Anchors;
    property BiDiMode;
    property Color;
    property Constraints;
    property DockSite;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property ParentColor;
    property ParentBiDiMode;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Visible;
    property OnCanResize;
    property OnContextPopup;
    property OnDockDrop;
    property OnDockOver;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnGetSiteInfo;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDock;
    property OnStartDrag;
    property OnUnDock;
    // TSpTBXCustomTabSet properties
    property ActiveTabIndex;
    property Images;
    property TabBackgroundColor;
    property TabPosition;
    property ThemeType;
    property OnActiveTabChange;
    property OnDrawBackground;
  end;

  { TSpTBXTabControl }

  TSpTBXCustomTabControl = class(TSpTBXCustomTabSet)
  private
    FEmptyTabSheet: TSpTBXTabSheet;
    function GetPages(Index: Integer): TSpTBXTabSheet;
  protected
    FPages: TList;
    procedure DoActiveTabChange(const ItemIndex: Integer); override;
    function GetFullRepaint: Boolean; override;
    procedure TabInserted(Item: TSpTBXTabItem); override;
    procedure TabDeleting(Item: TSpTBXTabItem; FreeTabSheet: Boolean = True); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function GetPage(Item: TSpTBXTabItem): TSpTBXTabSheet;
    property Pages[Index: Integer]: TSpTBXTabSheet read GetPages;
  end;

  TSpTBXTabControl = class(TSpTBXCustomTabControl)
  published
    property Align;
    property Anchors;
    property BiDiMode;
    property Color;
    property Constraints;
    property DockSite;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property ParentColor;
    property ParentBiDiMode;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Visible;
    property OnCanResize;
    property OnContextPopup;
    property OnDockDrop;
    property OnDockOver;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnGetSiteInfo;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDock;
    property OnStartDrag;
    property OnUnDock;
    // TSpTBXCustomTabControl properties
    property ActiveTabIndex;
    property Images;
    property TabBackgroundColor;
    property TabPosition;
    property ThemeType;
    property OnActiveTabChange;
    property OnDrawBackground;
  end;

procedure SpDrawXPTab(ACanvas: TCanvas; ARect: TRect; Enabled, Selected, MouseOver, Focused: Boolean; Position: TSpTBXTabPosition; ThemeType: TSpTBXTabThemeType = tttWindows; Edge: TSpTBXTabEdge = tedNone);
procedure SpDrawXPTabControlBackground(ACanvas: TCanvas; ARect: TRect; AColor: TColor; Position: TSpTBXTabPosition; ThemeType: TSpTBXTabThemeType = tttWindows);

implementation

uses
  TBXUxThemes, TBXUtils;

type
  TTBRootItemAccess = class(TTBRootItem);
  TSpTBXCustomItemAccess = class(TSpTBXCustomItem);

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ Helpers }

procedure DrawFocusRect2(Canvas: TCanvas; const R: TRect);
var
  DC: HDC;
  C1, C2: TColor;
begin
  DC := Canvas.Handle;
  C1 := SetTextColor(DC, clBlack);
  C2 := SetBkColor(DC, clWhite);
  Canvas.DrawFocusRect(R);
  SetTextColor(DC, C1);
  SetBkColor(DC, C2);
end;

procedure SpDrawXPTab(ACanvas: TCanvas; ARect: TRect;
  Enabled, Selected, MouseOver, Focused: Boolean; Position: TSpTBXTabPosition;
  ThemeType: TSpTBXTabThemeType = tttWindows; Edge: TSpTBXTabEdge = tedNone);
var
  Part, Flags: Cardinal;
  B: TBitmap;
  R: TRect;
  ItemInfo: TTBXItemInfo;
begin
  if (ThemeType = tttWindows) and not SpXPThemesAvailable then
    ThemeType := tttNone;
  if (ThemeType = tttNone) and not Selected then
    Exit;

  B := TBitmap.Create;
  try
    B.Width := ARect.Right - ARect.Left;
    B.Height := ARect.Bottom - ARect.Top;
    R := Rect(0, 0, B.Width, B.Height);

    if ThemeType = tttWindows then begin
      Case Edge of
        tedLeft: Part := TABP_TABITEMLEFTEDGE;
        tedRight: Part := TABP_TABITEMRIGHTEDGE;
      else
        Part := TABP_TABITEM;
      end;

      if not Enabled then
        Flags := TIS_DISABLED
      else
        if Selected then
          Flags := TIS_SELECTED
        else
          if MouseOver then
            Flags := TIS_HOT
          else
            Flags := TIS_NORMAL;
      DrawThemeBackground(SP_TAB_THEME, B.Canvas.Handle, Part, Flags, R, nil);
    end
    else begin
      if ThemeType = tttTBX then begin
        SpFillItemInfo(Enabled, False, MouseOver, Selected, ItemInfo);
        CurrentTheme.PaintButton(B.Canvas, R, ItemInfo);
      end
      else
        if Selected then begin
          Position := ttpTop; // Don't flip
          B.Canvas.Brush.Color := ACanvas.Brush.Color;
          B.Canvas.FillRect(R);

          if ThemeType = tttFlat then
            ExtCtrls.Frame3D(B.Canvas, R, clWindow, clBtnShadow, 1)
          else begin
            ExtCtrls.Frame3D(B.Canvas, R, clWindow, clWindowFrame, 1);
            ExtCtrls.Frame3D(B.Canvas, R, B.Canvas.Brush.Color, clBtnShadow, 1);
          end;
        end;
    end;

    R := Rect(0, 0, B.Width, B.Height);
    if Selected and Focused then begin
      InflateRect(R, -3, -3);
      DrawFocusRect2(B.Canvas, R);
    end;

    // Flip top to bottom
    if Position = ttpBottom then begin
      // Unclear why extra "-1" is needed here.
      R.Top := B.Height - 1;
      R.Bottom := -1;
    end;

    ACanvas.CopyRect(ARect, B.Canvas, R);
  finally
    B.Free;
  end;
end;

procedure SpDrawXPTabControlBackground(ACanvas: TCanvas; ARect: TRect; AColor: TColor;
  Position: TSpTBXTabPosition; ThemeType: TSpTBXTabThemeType = tttWindows);
var
  C: TColor;
  ItemInfo: TTBXItemInfo;
  B: TBitmap;
  R: TRect;
begin
  B := TBitmap.Create;
  try
    B.Width := ARect.Right - ARect.Left;
    B.Height := ARect.Bottom - ARect.Top;
    R := Rect(0, 0, B.Width, B.Height);

    if (ThemeType = tttWindows) and SpXPThemesAvailable then
      DrawThemeBackground(SP_TAB_THEME, B.Canvas.Handle, TABP_PANE, 0, R, nil)
    else begin
       // Draw the top/bottom border
      if ThemeType = tttTBX then begin
        B.Canvas.Brush.Color := clWhite;
        B.Canvas.FillRect(R);
        SpFillItemInfo(True, False, False, True, ItemInfo);
        CurrentTheme.PaintButton(B.Canvas, R, ItemInfo);
      end
      else begin
        Position := ttpTop; // Don't flip
        C := B.Canvas.Brush.Color;
        B.Canvas.Brush.Color := AColor;
        B.Canvas.FillRect(R);

        if ThemeType = tttFlat then
          ExtCtrls.Frame3D(B.Canvas, R, clWindow, clBtnShadow, 1)
        else begin
          ExtCtrls.Frame3D(B.Canvas, R, clWindow, clWindowFrame, 1);
          ExtCtrls.Frame3D(B.Canvas, R, AColor, clBtnShadow, 1);
        end;

        B.Canvas.Brush.Color := C;
      end;
    end;

    R := Rect(0, 0, B.Width, B.Height);

    Case Position of
      ttpBottom:
        begin
          // Flip top to bottom
          // Unclear why extra "-1" is needed here.
          R.Top := B.Height - 1;
          R.Bottom := -1
        end;
    end;

    ACanvas.CopyRect(ARect, B.Canvas, R);
  finally
    B.Free;
  end;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXTabItem }

constructor TSpTBXTabItem.Create(AOwner: TComponent);
begin
  inherited;
  GroupIndex := C_SpTBXTabGroupIndex;
  FTabPosition := ttpTop;
  FThemeType := tttWindows;
end;

procedure TSpTBXTabItem.Click;
begin
  if not Checked and Enabled then begin
    // Set the Checked property to True, Autocheck is False
    // Checked will call Item.Invalidate, the TabSet.ItemNotification will
    // handle the invalidation and set the ActiveTabIndex.
    Checked := True;
    inherited;
  end;
end;

procedure TSpTBXTabItem.DoDrawTab(ACanvas: TCanvas; ARect: TRect;
  AChecked, AHoverItem: Boolean; Position: TSpTBXTabPosition;
  ASeparator: Boolean = False; AEdge: TSpTBXTabEdge = tedNone);
var
  C: TColor;
  ItemInfo: TTBXItemInfo;
  TT: TSpTBXTabThemeType;
begin
  TT := ThemeType;

  if ASeparator then begin
    // TBX Bug!!! PaintSeparator treats the Rect as zero based
    // It assumes Rect.Left = 0
    // I had to multiply the R.Right by 2 to solve this issue,
    // but in the Default theme it won't work because it uses DrawEdge
    if (TT <> tttTBX) or (CurrentTheme.Name = 'Default') then begin
      C := ACanvas.Pen.Color;
      ACanvas.Pen.Color := clBtnShadow;
      ACanvas.MoveTo(ARect.Right - 2, ARect.Top + 3);
      ACanvas.LineTo(ARect.Right - 2, ARect.Bottom - 3);
      ACanvas.Pen.Color := clBtnHighlight;
      ACanvas.MoveTo(ARect.Right - 1, ARect.Top + 3);
      ACanvas.LineTo(ARect.Right - 1, ARect.Bottom - 3);
      ACanvas.Pen.Color := C;
    end
    else begin
      FillChar(ItemInfo, SizeOf(TTBXItemInfo), 0);
      with ItemInfo do begin
        ViewType    := TVT_MENUBAR;
        ItemOptions := IO_APPACTIVE or IO_TOOLBARSTYLE or IO_DESIGNING;
        Enabled     := True;
        IsVertical  := True;
      end;
      ARect := Rect(0, ARect.Top + 3, (ARect.Right - 2)*2, ARect.Bottom - 3);
      CurrentTheme.PaintSeparator(ACanvas, ARect, ItemInfo, false, true);
    end;
  end
  else begin
    if not Checked and AHoverItem and (TT = tttFlat) then
      TT := tttTBX; // Hover using TBX style
    C := ACanvas.Brush.Color;
    ACanvas.Brush.Color := TabColor;
    SpDrawXPTab(ACanvas, ARect, Enabled, AChecked, AHoverItem, False, Position, TT, AEdge);
    ACanvas.Brush.Color := C;
  end;
end;

procedure TSpTBXTabItem.DrawBottomBorder(ACanvas: TCanvas; ARect: TRect);
var
  CR, R: TRect;
  Edge: TSpTBXTabEdge;
  LeftT, RightT: Boolean;
  Position: TSpTBXTabPosition;
  T: TSpTBXTabToolbar;
  B: TBitmap;
  BottomOffset, BitmapBottomOffset: Integer;
begin
  Position := GetTabPosition;
  Edge := tedNone;
  CR := ARect;

  // [TBXTheme-Change]
  // Add the bottom borders
  // Stripes and Xito theme buttons are smaller, we need to inc the Height by 1
  if (TBXCurrentTheme = 'Stripes') or (TBXCurrentTheme = 'Xito') then
    BottomOffset := 2
  else
    BottomOffset := 1;
  BitmapBottomOffset := 4;
  Case Position of
    ttpTop:
      Inc(CR.Bottom, BottomOffset);
    ttpBottom:
      begin
        // When tttNone the bottom border size is 2
        if ThemeType = tttNone then
          Dec(CR.Top, 2)
        else
          Dec(CR.Top, BottomOffset);
      end;
  end;

  if GetTabToolbar(T) and (ThemeType = tttWindows) and SpXPThemesAvailable then begin
    LeftT := Assigned(GetNextTab(false, true));
    RightT := Assigned(GetNextTab(true, true));
    if IsFirstTab then
      Edge := tedLeft;

    if Edge = tedLeft then begin
      CR.Left := CR.Left - 2;
      if RightT then
        CR.Right := CR.Right + 2;
    end
    else begin
      if LeftT then
        CR.Left := CR.Left - 2;
      if RightT then
        CR.Right := CR.Right + 2;
    end;
  end;

  B := TBitmap.Create;
  try
    B.Width := CR.Right - CR.Left;
    B.Height := CR.Bottom - CR.Top + BitmapBottomOffset; // Larger than CR
    R := Rect(0, 0, B.Width, B.Height);
    DoDrawTab(B.Canvas, R, True, False, Position, False, Edge);

    Case Position of
      ttpTop:
        R  := Bounds(0, 0, CR.Right - CR.Left, CR.Bottom - CR.Top); // Copy from Y = 0
      ttpBottom:
        R  := Bounds(0, 2, CR.Right - CR.Left, CR.Bottom - CR.Top + 2); // Copy from Y = 2
    end;

    ACanvas.CopyRect(CR, B.Canvas, R);
  finally
    B.Free;
  end;
end;

procedure TSpTBXTabItem.DoDrawButton(ACanvas: TCanvas; ARect: TRect;
  ItemInfo: TTBXItemInfo; const PaintStage: TSpTBXPaintStage;
  var PaintDefault: Boolean);
var
  LeftT, RightT: TSpTBXTabItem;
  IsHoverItem: Boolean;
  R: TRect;
  Position: TSpTBXTabPosition;
  TabThemeType: TSpTBXTabThemeType;
  T: TSpTBXTabToolbar;
begin
  inherited;

  if (PaintStage = pstPrePaint) and PaintDefault then begin
    PaintDefault := false;
    IsHoverItem := ItemInfo.HoverKind <> hkNone;
    Position := GetTabPosition;

    // Match the bottom of the Tab with the bottom of the TabSet
    Case Position of
      ttpTop:    ARect.Bottom := ARect.Bottom + 1;
      ttpBottom: ARect.Top := ARect.Top - 1;
    end;
    R := ARect;

    TabThemeType := ThemeType;
    if (TabThemeType = tttWindows) and not SpXPThemesAvailable then
      TabThemeType := tttNone;

    if TabThemeType = tttWindows then begin
      if GetTabToolbar(T) then begin
        LeftT := GetNextTab(false, true);
        RightT := GetNextTab(true, true);
      end
      else begin
        LeftT := nil;
        RightT := nil;
      end;

      if Checked then begin
        //The left border of the Tab will be painted by the Left tab
        if Assigned(LeftT) or IsFirstTab then
          R.Left := R.Left - 2;
        //The right border of the Tab will be painted by the Right tab
        if Assigned(RightT) then
          R.Right := R.Right + 2;
      end
      else begin
        // Non checked tabs should be smaller
        Case Position of
          ttpTop:    Inc(R.Top, 2);
          ttpBottom: Dec(R.Bottom, 2);
        end;
      end;

      // Draw the Tab
      DoDrawTab(ACanvas, R, Checked, IsHoverItem, Position);

      // If the Tab is not checked then it should paint the active tab borders
      if not Checked then begin
        R := ARect;
        // Draw the left border
        if Assigned(LeftT) and LeftT.Checked then begin
          R.Right := R.Left + 2;
          R.Left := R.Right - 10;
          DoDrawTab(ACanvas, R, True, IsHoverItem, Position);
        end
        else
          // Draw the right border
          if Assigned(RightT) and RightT.Checked then begin
            R.Left := R.Right - 2;
            R.Right := R.Left + 10;
            DoDrawTab(ACanvas, R, True, IsHoverItem, Position);
          end;
      end;
    end
    else begin
      // Custom or TBX theme
      if Checked or (IsHoverItem and (TabThemeType <> tttNone)) then begin
        Case Position of
          ttpTop:    Inc(R.Bottom, 5);
          ttpBottom: Dec(R.Top, 5);
        end;
        DoDrawTab(ACanvas, R, Checked, IsHoverItem, Position);
      end
      else begin
        // Draw the separators
        RightT := GetNextTab(true, true);
        if Assigned(RightT) and not RightT.Checked then
          DoDrawTab(ACanvas, R, Checked, IsHoverItem, Position, True);
      end;
    end;
  end;
end;

procedure TSpTBXTabItem.DoDrawCaption(ACanvas: TCanvas; const ClientAreaRect: TRect;
  var ACaption: WideString; var CaptionRect: TRect; IsTextRotated: Boolean;
  const PaintStage: TSpTBXPaintStage; var PaintDefault: Boolean);
begin
  if not Checked then
    Case GetTabPosition of
      ttpTop:    OffsetRect(CaptionRect, 0, 2);
      ttpBottom: OffsetRect(CaptionRect, 0, -2);
    end;
  inherited DoDrawCaption(ACanvas, ClientAreaRect, ACaption, CaptionRect,
    IsTextRotated, PaintStage, PaintDefault);
end;

procedure TSpTBXTabItem.DoDrawImage(ACanvas: TCanvas;
  const ItemInfo: TTBXItemInfo; const PaintStage: TSpTBXPaintStage;
  AImageList: TCustomImageList; var AImageIndex: Integer; var ARect: TRect;
  var PaintDefault: Boolean);
begin
  if not Checked then
    OffsetRect(ARect, 0, 2);
  inherited DoDrawImage(ACanvas, ItemInfo, PaintStage, AImageList, AImageIndex,
    ARect, PaintDefault);
end;

function TSpTBXTabItem.GetNextTab(GoForward: Boolean; Inmediate: Boolean = True): TSpTBXTabItem;
// Returns the left or right Tab item depending on GoForward
// When Inmediate is false it will iterate until a tab is found
var
  T: TSpTBXTabToolbar;
  DefIV, IV: TTBItemViewer;
begin
  Result := nil;
  if GetTabToolbar(T) then begin
    DefIV := SpFindItemViewer(T.View, Self);
    if Assigned(DefIV) then begin
      IV := DefIV;
      while Result = nil do begin
        IV := SpGetNextItem(IV, GoForward, True, False, True);
        if not Assigned(IV) then
          Break
        else
          if (IV.Item is TSpTBXTabItem) then begin
            Result := IV.Item as TSpTBXTabItem;
            Break;
          end
          else
            if Inmediate then
              Break;
      end;
    end;
  end;
end;

function TSpTBXTabItem.GetTabColor: TColor;
var
  T: TSpTBXTabToolbar;
begin
  Result := clBtnFace;
  if GetTabToolbar(T) then
    Result := T.TabColor;
end;

function TSpTBXTabItem.GetTabPosition: TSpTBXTabPosition;
var
  T: TSpTBXTabToolbar;
begin
  if GetTabToolbar(T) then
    FTabPosition := T.TabPosition;
  Result := FTabPosition;
end;

procedure TSpTBXTabItem.SetTabPosition(const Value: TSpTBXTabPosition);
var
  T: TSpTBXTabToolbar;
begin
  // Don't change the TabPosition if the item is inside a TabToolbar
  if not GetTabToolbar(T) then begin
    FTabPosition := Value;
    Invalidate;
  end;
end;

function TSpTBXTabItem.GetThemeType: TSpTBXTabThemeType;
var
  T: TSpTBXTabToolbar;
begin
  if GetTabToolbar(T) then
    FThemeType := T.ThemeType;
  Result := FThemeType;
end;

procedure TSpTBXTabItem.SetThemeType(const Value: TSpTBXTabThemeType);
var
  T: TSpTBXTabToolbar;
begin
  // Don't change the ThemeType if the item is inside a TabToolbar
  if not GetTabToolbar(T) then begin
    FThemeType := Value;
    Invalidate;
  end;
end;

function TSpTBXTabItem.IsFirstTab: Boolean;
var
  T: TSpTBXTabToolbar;
begin
  Result := False;
  if GetTabToolbar(T) then
    if T.View.ViewerCount > 0 then
      Result := T.View.Viewers[0].Item = Self;
end;

function TSpTBXTabItem.IsLastTab: Boolean;
var
  T: TSpTBXTabToolbar;
begin
  if GetTabToolbar(T) then
    Result := not Assigned(GetNextTab(true, FALSE))
  else
    Result := False;
end;

function TSpTBXTabItem.GetTabToolbar(out TabToolbar: TSpTBXTabToolbar): Boolean;
var
  C: TComponent;
begin
  C := GetParentComponent;
  if Assigned(C) and (C is TSpTBXTabToolbar) then
    TabToolbar := C as TSpTBXTabToolbar
  else
    TabToolbar := nil;
  Result := Assigned(TabToolbar);
end;

procedure TSpTBXTabItem.ToggleControl;
begin
  // Do nothing, the Control property is the Tabsheet, and its visibility
  // is setted by TabSet.ActiveTabIndex
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXTabToolbar }

constructor TSpTBXTabToolbar.Create(AOwner: TComponent);
begin
  inherited;
  FHiddenTabs := TSpTBXItemCacheCollection.Create(TSpTBXItemCache);
  if Owner is TSpTBXCustomTabSet then
    FParentTabControl := Owner as TSpTBXCustomTabSet
  else
    FParentTabControl := nil;
  FActiveTabIndex := -1;
  FThemeType := tttWindows;
  FTabColor := clBtnFace;
  FTabPosition := ttpTop;
end;

destructor TSpTBXTabToolbar.Destroy;
begin
  FHiddenTabs.Free;
  inherited;
end;

procedure TSpTBXTabToolbar.DoItemNotification(Ancestor: TTBCustomItem;
  Relayed: Boolean; Action: TTBItemChangedAction; Index: Integer;
  Item: TTBCustomItem);
begin
  inherited;
  if not FUpdatingHidden then
    if Action = tbicInvalidateAndResize then
      InvalidateNC;
end;

procedure TSpTBXTabToolbar.DoActiveTabChange(const ItemIndex: Integer);
begin
  if Assigned(FOnActiveTabChange) then FOnActiveTabChange(Self, ItemIndex);
  if Assigned(FParentTabControl) then
    FParentTabControl.DoActiveTabChange(ItemIndex);
end;

procedure TSpTBXTabToolbar.DrawNCArea(const DrawToDC: Boolean;
  const ADC: HDC; const Clip: HRGN);
var
  DC: HDC;
  R, BR: TRect;
  PaintDefault: Boolean;
  ACanvas: TCanvas;
  ToolbarInfo: TTBXToolbarInfo;
  Tab: TSpTBXTabItem;
  IV: TTBItemViewer;
begin
  inherited;

  if (csDestroying in ComponentState) or not Docked or not HandleAllocated or
    not Assigned(FParentTabControl) then Exit;

  if not DrawToDC then DC := GetWindowDC(Handle)
  else DC := ADC;
  try
    GetWindowRect(Handle, R);
    OffsetRect(R, -R.Left, -R.Top);
    GetToolbarInfo(ToolbarInfo);
    if not DrawToDC then
    begin
      SelectNCUpdateRgn(Handle, DC, Clip);
      BR := R;
      with ToolbarInfo.BorderSize, BR do
      begin
        InflateRect(BR, -X, -Y);
        if ToolbarInfo.IsVertical then Inc(Top, GetTBXDragHandleSize(ToolbarInfo))
        else Inc(Left, GetTBXDragHandleSize(ToolbarInfo));
        ExcludeClipRect(DC, Left, Top, Right, Bottom);
      end;
    end;

    ACanvas := TCanvas.Create;
    try
      ACanvas.Handle := DC;

      // Draw the bottom border of the tabs pane
      Case FTabPosition of
        ttpTop:
          BR := Rect(R.Left, R.Bottom - 2, R.Right, R.Bottom + 4);
        ttpBottom:
          BR := Rect(R.Left, R.Top - 4, R.Right, R.Top + 2);
      end;
      PaintDefault := True;
      FParentTabControl.DoDrawBackground(ACanvas, BR, pstPrePaint, PaintDefault);
      if PaintDefault then
        SpDrawXPTabControlBackground(ACanvas, BR, FTabColor, FTabPosition, FThemeType);
      PaintDefault := True;
      FParentTabControl.DoDrawBackground(ACanvas, BR, pstPostPaint, PaintDefault);

      // Draw the bottom border of the active tab
      Tab := ActiveTab;
      if Assigned(Tab) and Tab.Visible then begin
        IV := SpFindItemViewer(View, Tab);
        if Assigned(IV) then begin
          BR := IV.BoundsRect;
          OffsetRect(BR, 2, 2);  // Add the toolbar margins
          Tab.DrawBottomBorder(ACanvas, BR);
        end;
      end;

    finally
      ACanvas.Handle := 0;
      ACanvas.Free;
    end;
  finally
    if not DrawToDC then ReleaseDC(Handle, DC);
  end;
end;

procedure TSpTBXTabToolbar.InvalidateActiveTab;
var
  Tab: TSpTBXTabItem;
  IV: TTBItemViewer;
begin
  Tab := ActiveTab;
  if Assigned(Tab) then begin
    IV := SpFindItemViewer(View, Tab);
    if Assigned(IV) then
      View.Invalidate(IV);
  end;
end;

procedure TSpTBXTabToolbar.InvalidateLeftAndRightTabs(ATab: TSpTBXTabItem);
var
  T: TSpTBXTabItem;
begin
  // Invalidate left and right tabs
  if Assigned(ATab) and not (csDestroying in ATab.ComponentState) then begin
    T := ATab.GetNextTab(false, true);
    if Assigned(T) then
      T.Invalidate;
    T := ATab.GetNextTab(true, true);
    if Assigned(T) then
      T.Invalidate;
  end;
end;

procedure TSpTBXTabToolbar.InvalidateNC;
begin
  if HandleAllocated then
    RedrawWindow(Handle, nil, 0, RDW_ERASE or RDW_INVALIDATE or RDW_FRAME);
end;

function TSpTBXTabToolbar.GetActiveTab: TSpTBXTabItem;
var
  Item: TTBCustomItem;
begin
  Result := nil;
  if Assigned(Items) and (FActiveTabIndex > -1) and (Items.Count > 0) and (FActiveTabIndex < Items.Count) then
  begin
    Item := Items[FActiveTabIndex];
    if Assigned(Item.Parent) and not (csDestroying in Item.ComponentState) and
      (Item is TSpTBXTabItem) then
    begin
      Result := Items[FActiveTabIndex] as TSpTBXTabItem;
    end;
  end;
end;

procedure TSpTBXTabToolbar.RightAlignItems;
// Hide the items on resizing
var
  I, J, VisibleWidth, RightAlignedWidth, SpacerW, RightAlignedBorder: integer;
  FirstIV, LastIV, IV: TTBItemViewer;
  Spacer: TSpTBXItemViewer;
  RightAlignedList: TList;
begin
  if FUpdatingHidden or not Assigned(CurrentDock) or (CurrentDock.Width = 0) or
    (CurrentDock.Position in [dpLeft, dpRight]) then
  begin
    Exit;
  end;

  FirstIV := View.NextSelectable(nil, true);
  if Assigned(FirstIV) then begin
    View.ValidatePositions;
    View.BeginUpdate;
    RightAlignedList := TList.Create;
    try
      // Find the spacer and the right aligned items
      Spacer := SpGetRightAlignedItems(View, RightAlignedList, VisibleWidth, RightAlignedWidth);
      if Assigned(Spacer) then begin
        SpacerW := TSpTBXCustomItemAccess(Spacer.Item).CustomWidth;
        RightAlignedBorder := CurrentDock.Width - 8 - RightAlignedWidth + SpacerW;
        VisibleWidth := VisibleWidth - SpacerW;
        SpacerW := CurrentDock.Width - VisibleWidth - 4;
      end
      else begin
        SpacerW := 0;
        RightAlignedBorder := CurrentDock.Width - 8;
      end;

      LastIV := View.NextSelectable(Spacer, False);

      // Show items
      for I := LastIV.Index to View.ViewerCount - 1 do begin
        IV := View.Viewers[I];
        if (IV <> FirstIV) and not IV.Item.Visible and (RightAlignedList.IndexOf(IV) = -1) then begin
          // If the item was hidden and can be showed remove it from the HiddenList
          J := FHiddenTabs.IndexOf(IV.Item);
          if J > -1 then begin
            VisibleWidth := VisibleWidth + FHiddenTabs[J].Width;
            if (VisibleWidth < CurrentDock.Width - 8) then begin
              SpacerW := SpacerW - FHiddenTabs[J].Width;
              FHiddenTabs.Delete(J);
              IV.Item.Visible := True;
            end
            else
              Break;
          end;
        end;
      end;

      // Hide items
      for I := View.ViewerCount - 1 downto 0 do begin
        IV := View.Viewers[I];
        if (IV <> FirstIV) and IV.Item.Visible and (IV.BoundsRect.Right > RightAlignedBorder) and (RightAlignedList.IndexOf(IV) = -1) then begin
          // If the item can't be showed add it to the HiddenList
          SpacerW := SpacerW + (IV.BoundsRect.Right - IV.BoundsRect.Left);
          FHiddenTabs.Add(IV.Item, IV.BoundsRect); // Add it before it's hidden, otherwise IV.BoundsRect is invalid
          IV.Item.Visible := False;
        end;
      end;

      // Resize the spacer
      if Assigned(Spacer) then
        TSpTBXCustomItemAccess(Spacer.Item).CustomWidth := SpacerW;

      View.UpdatePositions;
    finally
      RightAlignedList.Free;
      View.EndUpdate;
    end;
  end;
end;

procedure TSpTBXTabToolbar.ScrollLeft;
var
  FirstIV, LastIV: TTBItemViewer;
  I, ClippedIndex: integer;
  Spacer: TSpTBXItemViewer;
begin
  if Items.Count > 1 then begin
    FirstIV := View.NextSelectable(nil, True);

    // LastIV minus the right aligned items
    Spacer := SpGetFirstRightAlignSpacer(View);
    if Assigned(Spacer) then
      LastIV := View.NextSelectable(Spacer, False)
    else
      LastIV := View.NextSelectable(nil, False);

    if Assigned(FirstIV) and Assigned(LastIV) and (FirstIV <> LastIV) then begin
      // Find the first clipped tab from the left side of the tabset
      for I := FirstIV.Index - 1 downto 0 do begin
        ClippedIndex := FHiddenTabs.IndexOf(View.Viewers[I].Item);
        if ClippedIndex > -1 then begin
          View.BeginUpdate;
          FUpdatingHidden := True;
          try
            // Hide the first visible tab from the right side of the tabset
            FHiddenTabs.Add(LastIV.Item, LastIV.BoundsRect);
            LastIV.Item.Visible := False;
            // Show the clipped tab
            FHiddenTabs[ClippedIndex].Item.Visible := True;
            FHiddenTabs.Delete(ClippedIndex);
            Break;
          finally
            FUpdatingHidden := False;
            if Assigned(Spacer) then
              RightAlignItems;
            View.EndUpdate;
          end;
        end;
      end;
    end;
  end;
end;

procedure TSpTBXTabToolbar.ScrollRight;
var
  FirstIV, LastIV: TTBItemViewer;
  I, ClippedIndex: integer;
  Spacer: TSpTBXItemViewer;
begin
  if Items.Count > 1 then begin
    FirstIV := View.NextSelectable(nil, True);

    // LastIV minus the right aligned items
    Spacer := SpGetFirstRightAlignSpacer(View);
    if Assigned(Spacer) then
      LastIV := View.NextSelectable(Spacer, False)
    else
      LastIV := View.NextSelectable(nil, False);

    if Assigned(FirstIV) and Assigned(LastIV) and (FirstIV <> LastIV) then begin
      // Find the first clipped tab from the right side of the tabset
      for I := LastIV.Index + 1 to View.ViewerCount - 1 do begin
        ClippedIndex := FHiddenTabs.IndexOf(View.Viewers[I].Item);
        if ClippedIndex > -1 then begin
          // Hide the first visible tab from the left side of the tabset
          View.BeginUpdate;
          FUpdatingHidden := True;
          try
            FHiddenTabs.Add(FirstIV.Item, FirstIV.BoundsRect);
            FirstIV.Item.Visible := False;
            // Show first clipped tab
            FHiddenTabs[ClippedIndex].Item.Visible := True;
            FHiddenTabs.Delete(ClippedIndex);
            Break;
          finally
            FUpdatingHidden := False;
            if Assigned(Spacer) then
              RightAlignItems;
            View.EndUpdate;
          end;
        end;
      end;
    end;
  end;
end;

procedure TSpTBXTabToolbar.ScrollState(out CanScrollToLeft, CanScrollToRight: Boolean);
var
  FirstIV, LastIV: TTBItemViewer;
  I, ClippedIndex: integer;
  Spacer: TSpTBXItemViewer;
begin
  CanScrollToLeft := False;
  CanScrollToRight := False;
  if FHiddenTabs.Count > 0 then begin
    FirstIV := View.NextSelectable(nil, True);
    if Assigned(FirstIV) then begin
      // Find the first clipped tab from the left side of the tabset
      for I := FirstIV.Index - 1 downto 0 do begin
        ClippedIndex := FHiddenTabs.IndexOf(View.Viewers[I].Item);
        if ClippedIndex > -1 then
          CanScrollToLeft := True;
      end;
    end;

    // LastIV minus the right aligned items
    Spacer := SpGetFirstRightAlignSpacer(View);
    if Assigned(Spacer) then
      LastIV := View.NextSelectable(Spacer, False)
    else
      LastIV := View.NextSelectable(nil, False);
    // Find the first clipped tab from the right side of the tabset
    if Assigned(LastIV) then begin
      for I := LastIV.Index + 1 to View.ViewerCount - 1 do begin
        ClippedIndex := FHiddenTabs.IndexOf(View.Viewers[I].Item);
        if ClippedIndex > -1 then
          CanScrollToRight := True;
      end;
    end;
  end;
end;

procedure TSpTBXTabToolbar.SetActiveTabIndex(Value: integer);
var
  ATab, APrevTab: TSpTBXTabItem;
  I: integer;
begin
  if (Value > -1) and (Value < Items.Count) and not (csDestroying in Items[Value].ComponentState) then begin
    if not (Items[Value] is TSpTBXTabItem) then
      Value := FActiveTabIndex;
  end
  else
    Value := -1;

  if Value <> FActiveTabIndex then
  begin
    I := FActiveTabIndex;
    FActiveTabIndex := Value;

    // Hide the previous TabSheet
    if (I > -1)  and (I < Items.Count) and not (csDestroying in Items[I].ComponentState) and
      (Items[I] is TSpTBXTabItem) then
    begin
      APrevTab := Items[I] as TSpTBXTabItem;
      APrevTab.Checked := False;
      if Assigned(APrevTab.Control) then
        APrevTab.Control.Visible := False;
    end
    else
      APrevTab := nil;

    // Check the item and invalidate NC
    if FActiveTabIndex > -1 then begin
      // Show the TabSheet
      ATab := Items[FActiveTabIndex] as TSpTBXTabItem;
      ATab.Checked := True;
      if Assigned(ATab.Control) then begin
        ATab.Control.Visible := True;
        ATab.Control.BringToFront;
      end;
    end
    else
      ATab := nil;

    DoActiveTabChange(FActiveTabIndex);
    InvalidateNC;

    // Invalidate left and right tabs of the previous active tab
    if Assigned(APrevTab) then
      InvalidateLeftAndRightTabs(APrevTab);
    // Invalidate left and right tabs of the active tab
    if Assigned(ATab) then
      InvalidateLeftAndRightTabs(ATab);
  end;
end;

procedure TSpTBXTabToolbar.SetTabColor(const Value: TColor);
begin
  if (FTabColor <> Value) then begin
    FTabColor := Value;
    if FThemeType <> tttTBX then begin
      Invalidate;
      InvalidateNC;
    end;
  end;
end;

procedure TSpTBXTabToolbar.SetTabPosition(const Value: TSpTBXTabPosition);
begin
  if FTabPosition <> Value then
    FTabPosition := Value;
end;

procedure TSpTBXTabToolbar.SetThemeType(const Value: TSpTBXTabThemeType);
begin
  if FThemeType <> Value then begin
    FThemeType := Value;
    Invalidate;
    InvalidateNC;
  end;
end;

procedure TSpTBXTabToolbar.CMDesignHitTest(var Message: TCMDesignHitTest);
var
  P: TPoint;
  V: TTBItemViewer;
  Shift: TShiftState;
begin
  // Allow left-clicks on TabItems at design time
  Shift := KeysToShiftState(Message.Keys);
  if (csDesigning in ComponentState) and (ssLeft in Shift) and Assigned(View) then begin
    P := SmallPointToPoint(Message.Pos);
    V := View.ViewerFromPoint(P);
    if Assigned(V) and Assigned(V.Item) and (V.Item is TSpTBXTabItem) then
      V.Item.Click;
  end;

  inherited;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXTabSheet }

procedure TSpTBXTabSheet.AdjustClientRect(var Rect: TRect);
begin
  if Assigned(FTabControl) then begin
    inc(Rect.Left, 2);
    dec(Rect.Right, 2);

    Case FTabControl.TabPosition of
      ttpTop:    dec(Rect.Bottom, 2);
      ttpBottom: inc(Rect.Top, 2);
    end;
  end;

  inherited AdjustClientRect(Rect);
end;

constructor TSpTBXTabSheet.Create(AOwner: TComponent);
begin
  inherited;
  ControlStyle := ControlStyle + [csAcceptsControls, csOpaque, csSetCaption];
  Align := alClient;
  Visible := False;
end;

procedure TSpTBXTabSheet.DefineProperties(Filer: TFiler);
begin
  inherited;
  Filer.DefineProperty('TabItem', ReadItemName, WriteItemName, True);
end;

destructor TSpTBXTabSheet.Destroy;
begin
  FTabControl.TabDeleting(Item, False);
  FTabControl := nil;
  FItem := nil;
  inherited;
end;

procedure TSpTBXTabSheet.ReadItemName(Reader: TReader);
begin
  case Reader.NextValue of
    vaLString, vaString:
     FItemName := Reader.ReadString;
  else
    FItemName := Reader.ReadWideString;
  end;
end;

procedure TSpTBXTabSheet.WriteItemName(Writer: TWriter);
begin
  if Assigned(Item) then
    FItemName := Item.Name;
  Writer.WriteWideString(FItemName);
end;

procedure TSpTBXTabSheet.ReadState(Reader: TReader);
var
  C: TComponent;
  TC: TSpTBXCustomTabControl;
begin
  // The TabSheet is being created from the DFM stream
  // We must set the initial values of TabControl, Item and add itself to
  // the Pages list of the parent TabControl.

  inherited ReadState(Reader);
  if Reader.Parent is TSpTBXCustomTabControl then begin
    // Set TabControl
    TC := TSpTBXCustomTabControl(Reader.Parent);
    TabControl := TC;
    // Set Item and add Self to TabControl.Pages
    if not Assigned(FItem) and (FItemName <> '') then begin
      C := Owner.FindComponent(FItemName);
      if Assigned(C) and (C is TSpTBXTabItem) then begin
        FItem := C as TSpTBXTabItem;
        FItem.Control := Self;
        if TC.FPages.IndexOf(Self) = -1 then
          TC.FPages.Add(Self);
      end;
    end;
  end;
end;

procedure TSpTBXTabSheet.WMEraseBkgnd(var Message: TWMEraseBkgnd);
var
  R: TRect;
begin
  if Assigned(FTabControl) then begin
    R := ClientRect;
    Case FTabControl.TabPosition of
      ttpTop:    dec(R.Top, 4);
      ttpBottom: inc(R.Bottom, 4);
    end;

    if FTabControl.DrawBackground(Message.DC, R) then
      Message.Result := 1
    else
      inherited;
  end;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXCustomTabSet }

constructor TSpTBXCustomTabSet.Create(AOwner: TComponent);
begin
  inherited;
  ControlStyle := ControlStyle + [csAcceptsControls, csOpaque];

  FTabEmpty := True;
  FBackground := TBitmap.Create;

  Width := 289;
  Height := FDock.Height + 2;
  ParentColor := False;
  Color := clBtnFace;

  FToolbar.Items.RegisterNotification(ItemNotification);
  AddThemeNotification(Self);
end;

destructor TSpTBXCustomTabSet.Destroy;
begin
  FToolbar.Items.UnRegisterNotification(ItemNotification);
  RemoveThemeNotification(Self);
  FBackground.Free;
  FBackground := nil;
  inherited;
end;

procedure TSpTBXCustomTabSet.DefineProperties(Filer: TFiler);
begin
  inherited;
  Filer.DefineProperty('HiddenItems', ReadHiddenItems, WriteHiddenItems, True);
end;

procedure TSpTBXCustomTabSet.ReadHiddenItems(Reader: TReader);
begin
  if Reader.ReadValue = vaCollection then
    Reader.ReadCollection(ValidateToolbar.FHiddenTabs);
end;

procedure TSpTBXCustomTabSet.WriteHiddenItems(Writer: TWriter);
begin
  Writer.WriteCollection(ValidateToolbar.FHiddenTabs);
end;

procedure TSpTBXCustomTabSet.Loaded;
var
  I: Integer;
  CacheCollection: TSpTBXItemCacheCollection;
  Cache: TSpTBXItemCache;
  C: TComponent;
begin
  inherited;
  // Read the HiddenTabs collection, and fill the Item property of the
  // collection items reading the Name from the DFM
  CacheCollection := ValidateToolbar.FHiddenTabs;
  if Assigned(CacheCollection) then
    for I := CacheCollection.Count - 1 downto 0 do begin
      Cache := CacheCollection[I];
      if not Assigned(Cache.Item) then begin
        if Cache.Name = '' then
          CacheCollection.Delete(I)
        else begin
          C := Owner.FindComponent(Cache.Name);
          if Assigned(C) and (C is TTBCustomItem) then
            Cache.Item := C as TTBCustomItem;
        end;
      end;
    end;

  ActiveTabIndex := FLoadingActiveIndex;
end;

function TSpTBXCustomTabSet.GetToolbarClass: TSpTBXToolbarClass;
begin
  Result := TSpTBXTabToolbar;
end;

function TSpTBXCustomTabSet.ValidateToolbar: TSpTBXTabToolbar;
begin
  Result := FToolbar as TSpTBXTabToolbar;
end;

function TSpTBXCustomTabSet.GetFullRepaint: Boolean;
begin
  Result := True;
end;

function TSpTBXCustomTabSet.Add(ACaption: WideString): TSpTBXTabItem;
var
  I: Integer;
  SpacerIV: TSpTBXItemViewer;
begin
  Result := TSpTBXTabItem.Create(nil);
  try
    Result.Caption := ACaption;
    SpacerIV := SpGetFirstRightAlignSpacer(View);
    if Assigned(SpacerIV) then begin
      I := Items.IndexOf(SpacerIV.Item);
      if I > -1 then
        Items.Insert(I, Result);
    end
    else
      Items.Add(Result);
  except
    Result.Free;
    Result := nil;
  end;
end;

function TSpTBXCustomTabSet.Insert(NewIndex: Integer; ACaption: WideString): TSpTBXTabItem;
begin
  Result := TSpTBXTabItem.Create(nil);
  try
    Result.Caption := ACaption;
    Items.Insert(NewIndex, Result);
  except
    Result.Free;
    Result := nil;
  end;
end;

procedure TSpTBXCustomTabSet.TabClick(const ATab: TSpTBXTabItem);
begin
  ATab.Click; // calls TabToolbar.DoTabClick and Self.DoTabClick
end;

procedure TSpTBXCustomTabSet.DoActiveTabChange(const ItemIndex: Integer);
begin
  if Assigned(FOnActiveTabChange) then FOnActiveTabChange(Self, ItemIndex);
end;

procedure TSpTBXCustomTabSet.ScrollLeft;
begin
  if Assigned(FToolbar) then ValidateToolbar.ScrollLeft;
end;

procedure TSpTBXCustomTabSet.ScrollRight;
begin
  if Assigned(FToolbar) then ValidateToolbar.ScrollRight;
end;

procedure TSpTBXCustomTabSet.ScrollState(out Left, Right: Boolean);
begin
  if Assigned(FToolbar) then ValidateToolbar.ScrollState(Left, Right);
end;

function TSpTBXCustomTabSet.GetActiveTabIndex: integer;
begin
  if Assigned(FToolbar) then
    Result := ValidateToolbar.ActiveTabIndex
  else
    Result := -1;
end;

procedure TSpTBXCustomTabSet.SetActiveTabIndex(Value: integer);
begin
  // When the component is reading from the DFM the Items are not created.
  // We must save the value setted at design time and use it when the
  // form is finally loaded.
  if csLoading in ComponentState then
    FLoadingActiveIndex := Value
  else
    if Assigned(FToolbar) then
      ValidateToolbar.ActiveTabIndex := Value;
end;

function TSpTBXCustomTabSet.GetTabBackgroundColor: TColor;
begin
  if Assigned(FToolbar) then
    Result := FToolbar.Color
  else
    Result := clNone;
end;

procedure TSpTBXCustomTabSet.SetTabBackgroundColor(const Value: TColor);
begin
  if Assigned(FToolbar) then FToolbar.Color := Value
end;

function TSpTBXCustomTabSet.GetThemeType: TSpTBXTabThemeType;
begin
  if Assigned(FToolbar) then
    Result := ValidateToolbar.ThemeType
  else
    Result := tttNone;
end;

procedure TSpTBXCustomTabSet.SetThemeType(const Value: TSpTBXTabThemeType);
begin
  if Assigned(FToolbar) then begin
    ValidateToolbar.ThemeType := Value;
    InvalidateBackground;
  end;
end;

function TSpTBXCustomTabSet.GetTabPosition: TSpTBXTabPosition;
begin
  if Assigned(FToolbar) then
    Result := ValidateToolbar.TabPosition
  else
    Result := ttpTop;
end;

procedure TSpTBXCustomTabSet.SetTabPosition(const Value: TSpTBXTabPosition);
var
  T: TSpTBXTabToolbar;
begin
  if Assigned(FToolbar) and Assigned(FDock) then begin
    T := ValidateToolbar;
    if T.TabPosition <> Value then begin
      T.Visible := False;
      T.Floating := True;
      T.TabPosition := Value;
      Case Value of
        ttpTop:    FDock.Position := dpTop;
        ttpBottom: FDock.Position := dpBottom;
      end;
      T.CurrentDock := FDock;
      T.Visible := True;

      InvalidateBackground;
    end;
  end;
end;

function TSpTBXCustomTabSet.GetTabSetHeight: Integer;
begin
  if Assigned(FDock) then
    Result := FDock.Height
  else
    Result := 0;
end;

procedure TSpTBXCustomTabSet.DoDrawBackground(ACanvas: TCanvas;
  ARect: TRect; const PaintStage: TSpTBXPaintStage;
  var PaintDefault: Boolean);
begin
  if Assigned(FOnDrawBackground) then FOnDrawBackground(Self, ACanvas, ARect, PaintStage, PaintDefault);
end;

function TSpTBXCustomTabSet.DrawBackground(DC: HDC; ARect: TRect): Boolean;
var
  ACanvas: TCanvas;
  PaintDefault: Boolean;
  R: TRect;
begin
  Result := False;
  if (csDestroying in ComponentState) or not Assigned(FDock) or
    not Assigned(FBackground) or IsRectEmpty(ARect) then Exit;

  ACanvas := TCanvas.Create;
  try
    ACanvas.Handle := DC;
    R := Rect(0, 0, ARect.Right - ARect.Left, ARect.Bottom - ARect.Top);

    if (FBackground.Width = R.Right) and (FBackground.Height = R.Bottom) then
      ACanvas.Draw(ARect.Left, ARect.Top, FBackground)
    else begin
      FBackground.Width := R.Right;
      FBackground.Height := R.Bottom;

      PaintDefault := true;
      DoDrawBackground(FBackground.Canvas, R, pstPrePaint, PaintDefault);
      if PaintDefault then begin
        SpDrawXPTabControlBackground(FBackground.Canvas, R, Color, TabPosition, ThemeType);
        ACanvas.Draw(ARect.Left, ARect.Top, FBackground);
      end;

      PaintDefault := true;
      DoDrawBackground(FBackground.Canvas, R, pstPostPaint, PaintDefault);
    end;
    Result := True;
  finally
    ACanvas.Handle := 0;
    ACanvas.Free;
  end;
end;

procedure TSpTBXCustomTabSet.InvalidateBackground(Resizing: Boolean = False);
var
  I: Integer;
  ChildW: TWinControl;
begin
  // Force background repaint
  if (csDestroying in ComponentState) or not Assigned(FToolbar) then Exit;
  if Assigned(FBackground) then FBackground.Width := 1;
  // Invalidate will not fire WM_ERASEBKGND, because csOpaque is setted
  if HandleAllocated then
    if Resizing then begin
      RedrawWindow(Handle, nil, 0, RDW_ERASE or RDW_INVALIDATE);
        // We need to invalidate all children minus the Dock and Toolbar
      for I := 0 to ControlCount - 1 do
        if Assigned(Controls[I]) and (Controls[I] is TWinControl) then begin
          ChildW := Controls[I] as TWinControl;
          if (ChildW <> FDock) and (ChildW <> FToolbar) then
            RedrawWindow(ChildW.Handle, nil, 0, RDW_ERASE or RDW_INVALIDATE);
        end;
    end
    else
      RedrawWindow(Handle, nil, 0, RDW_ERASE or RDW_INVALIDATE or RDW_ALLCHILDREN);
end;

procedure TSpTBXCustomTabSet.ItemNotification(Ancestor: TTBCustomItem;
  Relayed: Boolean; Action: TTBItemChangedAction; Index: Integer;
  Item: TTBCustomItem);
var
  I: Integer;
  Tab: TSpTBXTabItem;
begin
  inherited;

  if Assigned(FToolbar) and Assigned(Item) and not Relayed then
    Case Action of
      tbicInserted:
        begin
          // Update the index if a new item is inserted before the ActiveTabIndex
          I := Items.IndexOf(Item);
          if (I > -1) and (I <= ActiveTabIndex) then begin
            FUpdatingIndex := True;
            try
              // Don't change the ActiveTabIndex, just set the internal value
              // because the page didn't change
              ValidateToolbar.FActiveTabIndex := ValidateToolbar.FActiveTabIndex + 1;
            finally
              FUpdatingIndex := False;
            end;
          end;
          if (Item is TSpTBXTabItem) then
            TabInserted(Item as TSpTBXTabItem);
          InvalidateBackground;
        end;
      tbicDeleting:
        // The ItemViewer of the Item is not valid, it was destroyed by TTBView
        // The Items array still has the Item.
        if not (csDestroying in ComponentState) then begin
          FUpdatingIndex := True;
          try
            Tab := nil;
            I := Items.IndexOf(Item);
            if I > -1 then begin
              if I < ActiveTabIndex then
                // Don't change the ActiveTabIndex, just set the internal value
                // because the page didn't change
                ValidateToolbar.FActiveTabIndex := ValidateToolbar.FActiveTabIndex - 1
              else
                if I = ActiveTabIndex then
                  if I = 0 then begin
                    if (Items.Count > 1) and (Items[1] is TSpTBXTabItem) then begin
                      // The first tab was deleted, change the internal value of
                      // ActiveTabIndex to -1 and click the next tab.
                      // Update the checked tab on WM_INVALIDATETABBACKGROUND
                      Tab := Items[1] as TSpTBXTabItem;
                      Tab.Click;
                    end
                    else
                      ValidateToolbar.FActiveTabIndex := -1;
                  end
                  else
                    SetActiveTabIndex(I - 1); // Prev tab
            end;
            if (Item is TSpTBXTabItem) then
              TabDeleting(Item as TSpTBXTabItem);
            if (csDesigning in ComponentState) or Assigned(Tab) then
              PostMessage(Handle, WM_INVALIDATETABBACKGROUND, 0, 0)
            else
              InvalidateBackground;
          finally
            FUpdatingIndex := False;
          end;
        end;
      tbicInvalidate:
        // When the Item.Checked property changes we must reset the ActiveTabIndex
        if not FUpdatingIndex and (Item is TSpTBXTabItem) and Item.Checked and Item.Enabled then
        begin
          I := Items.IndexOf(Item);
          if I <> ActiveTabIndex then begin
            FUpdatingIndex := True;
            try
              SetActiveTabIndex(I);
            finally
              FUpdatingIndex := False;
            end;
          end;
        end;
    end;
end;

procedure TSpTBXCustomTabSet.TabDeleting(Item: TSpTBXTabItem;
  FreeTabSheet: Boolean);
begin
  if not (csDestroying in ComponentState) then
    FTabEmpty := Items.Count = 0;
end;

procedure TSpTBXCustomTabSet.TabInserted(Item: TSpTBXTabItem);
begin
  if not (csLoading in ComponentState) then
    if FTabEmpty then
      Item.Click // Select the first Tab
    else
      FTabEmpty := False;
end;

procedure TSpTBXCustomTabSet.CMColorchanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FToolbar) then begin
    ValidateToolbar.TabColor := Color;
    InvalidateBackground;
  end;
end;

procedure TSpTBXCustomTabSet.TBMThemeChange(var Message: TMessage);
begin
  inherited;
  if Message.WParam = TSC_AFTERVIEWCHANGE then
    InvalidateBackground;
end;

procedure TSpTBXCustomTabSet.WMEraseBkgnd(var Message: TWMEraseBkgnd);
var
  R: TRect;
begin
  if not (csDestroying in ComponentState) and GetFullRepaint then begin
    R := ClientRect;
    Case TabPosition of
      ttpTop:    inc(R.Top, GetTabSetHeight - 2);
      ttpBottom: dec(R.Bottom, GetTabSetHeight - 2);
    end;
    if DrawBackground(Message.DC, R) then
      Message.Result := 1
    else
      inherited;
  end
  else
    Message.Result := 1;
end;

procedure TSpTBXCustomTabSet.WMInvalidateTabBackground(var Message: TMessage);
var
  Tab: TSpTBXTabItem;
  I: Integer;
begin
  if Assigned(FToolbar) then begin
    Tab := ValidateToolbar.ActiveTab;
    if Assigned(Tab) then begin
      I := Items.IndexOf(Tab);
      FUpdatingIndex := True;
      try
        ValidateToolbar.FActiveTabIndex := -1;
        SetActiveTabIndex(I);
      finally
        FUpdatingIndex := False;
      end;
    end;
    if HandleAllocated then
      RedrawWindow(Handle, nil, 0, RDW_ERASE or RDW_INVALIDATE or RDW_ALLCHILDREN);
  end;
end;

procedure TSpTBXCustomTabSet.WMWindowPosChanged(var Message: TWMWindowPosChanged);
begin
  if GetFullRepaint then
    InvalidateBackground(True);
  inherited;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM
{ TSpTBXCustomTabControl }

constructor TSpTBXCustomTabControl.Create(AOwner: TComponent);
begin
  inherited;
  ControlStyle := ControlStyle - [csAcceptsControls];
  FPages := TList.Create;

  // FEmptyTabSheet is used to hide the rest of the TabSheets
  // when ActiveTabIndex = -1 at design time.
  FEmptyTabSheet := TSpTBXTabSheet.Create(Self);
  FEmptyTabSheet.Parent := Self;
  FEmptyTabSheet.TabControl := Self;
  FEmptyTabSheet.Item := nil;
  FEmptyTabSheet.Visible := True;
  FEmptyTabSheet.BringToFront;
  FEmptyTabSheet.ControlStyle := FEmptyTabSheet.ControlStyle - [csAcceptsControls];

  Width := 289;
  Height := 193;
end;

destructor TSpTBXCustomTabControl.Destroy;
begin
  FPages.Free;
  inherited;
end;

procedure TSpTBXCustomTabControl.DoActiveTabChange(const ItemIndex: Integer);
begin
  if ItemIndex = -1 then begin
    FEmptyTabSheet.Visible := True;
    FEmptyTabSheet.BringToFront;
  end
  else
    FEmptyTabSheet.Visible := False;
  inherited;
end;

function TSpTBXCustomTabControl.GetFullRepaint: Boolean;
begin
  Result := not Assigned(FPages) or (FPages.Count = 0) or not Assigned(FToolbar) or
    not Assigned(ValidateToolbar.ActiveTab) or not ValidateToolbar.ActiveTab.Checked;
end;

function TSpTBXCustomTabControl.GetPage(Item: TSpTBXTabItem): TSpTBXTabSheet;
var
  I: Integer;
begin
  Result := nil;
  I := Items.IndexOf(Item);
  if (I > - 1) and Assigned(Item.Control) and (Item.Control is TSpTBXTabSheet) then
    Result := Item.Control as TSpTBXTabSheet;
end;

function TSpTBXCustomTabControl.GetPages(Index: Integer): TSpTBXTabSheet;
begin
  Result := TSpTBXTabSheet(FPages[Index]);
end;

procedure TSpTBXCustomTabControl.TabInserted(Item: TSpTBXTabItem);
var
  T: TSpTBXTabSheet;
  I: Integer;
begin
  // Create a TabSheet and Link it to the TabItem, only if the Item is created
  // at DesignTime.
  // If the Item is created from the DFM stream then the TabSheet will be
  // automatically created, because it will also be streamed, but it won't be
  // linked to the Item, this is done in TabSheet.ReadState

  if (csLoading in ComponentState) or not Assigned(Item) then Exit;

  for I := 0 to FPages.Count - 1 do begin
    T := TSpTBXTabSheet(FPages[I]);
    if T.Item = Item then begin
      Exit;
      raise Exception.Create('TabSheet Already Exists');
    end;
  end;

  // Find unique name
  I := 1;
  while Owner.FindComponent('SpTBXTabSheet' + IntToStr(I)) <> nil do
    inc(I);

  // The Form will be the owner, it will stream the tabsheet to the DFM
  T := TSpTBXTabSheet.Create(Owner);
  T.Name := 'SpTBXTabSheet' + IntToStr(I);
  T.Parent := Self;
  T.TabControl := Self;
  T.Item := Item;
  Item.Control := T;
  T.SendToBack;
  FPages.Add(T);

  if FPages.Count = 1 then
    Item.Click; // Select the first Tab
end;

procedure TSpTBXCustomTabControl.TabDeleting(Item: TSpTBXTabItem;
  FreeTabSheet: Boolean);
var
  I: Integer;
  T: TSpTBXTabSheet;
begin
  // The Toolbar will free the Items, and the Form will free the TabSheets
  if (csDestroying in ComponentState) or not Assigned(Item) then Exit;

  for I := 0 to FPages.Count - 1 do begin
    T := TSpTBXTabSheet(FPages[I]);
    if Assigned(T) and Assigned(T.Item) and (T.Item = Item) then begin
      FPages[I] := nil;
      FPages.Delete(I);
      if FreeTabSheet then
        T.Free
      else begin
        // TabSheet deleted at design time, free the linked Item
        T.Item.Free;
        T.Item := nil;
      end;
      Break;
    end;
  end;
end;

//WMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWMWM

end.
